@@ -1,5 +1,7 @@
 /* Determine whether two file names refer to the same file.
-   Copyright (C) 1997-2000 Free Software Foundation, Inc.
+
+   Copyright (C) 1997, 1998, 1999, 2000, 2003, 2004 Free Software
+   Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -18,19 +20,8 @@
 #ifndef SAME_H_
 # define SAME_H_ 1
 
-# if HAVE_CONFIG_H
-#  include <config.h>
-# endif
+# include <stdbool.h>
 
-# ifndef PARAMS
-#  if defined PROTOTYPES || (defined __STDC__ && __STDC__)
-#   define PARAMS(Args) Args
-#  else
-#   define PARAMS(Args) ()
-#  endif
-# endif
-
-int
-same_name PARAMS ((const char *source, const char *dest));
+bool same_name (const char *source, const char *dest);
 
 #endif /* SAME_H_ */
