@@ -1,5 +1,5 @@
 /* basename -- strip directory and suffix from filenames
-   Copyright (C) 1990-1997, 1999-2003 Free Software Foundation, Inc.
+   Copyright (C) 1990-1997, 1999-2004 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -26,6 +26,7 @@
    => functions.lis */
 
 #include <config.h>
+#include <getopt.h>
 #include <stdio.h>
 #include <sys/types.h>
 
@@ -33,6 +34,7 @@
 #include "long-options.h"
 #include "dirname.h"
 #include "error.h"
+#include "quote.h"
 
 /* The official name of this program (e.g., no `g' prefix).  */
 #define PROGRAM_NAME "basename"
@@ -45,7 +47,7 @@ char *program_name;
 void
 usage (int status)
 {
-  if (status != 0)
+  if (status != EXIT_SUCCESS)
     fprintf (stderr, _("Try `%s --help' for more information.\n"),
 	     program_name);
   else
@@ -100,27 +102,27 @@ main (int argc, char **argv)
   atexit (close_stdout);
 
   parse_long_options (argc, argv, PROGRAM_NAME, GNU_PACKAGE, VERSION,
-		      AUTHORS, usage);
-  /* The above handles --help and --version.
-     Since there is no other invocation of getopt, handle `--' here.  */
-  if (argc > 1 && STREQ (argv[1], "--"))
-    {
-      --argc;
-      ++argv;
-    }
+		      usage, AUTHORS, (char const *) NULL);
+  if (getopt_long (argc, argv, "+", NULL, NULL) != -1)
+    usage (EXIT_FAILURE);
 
-  if (argc <= 1 || argc > 3)
+  if (argc < optind + 1)
     {
-      error (0, 0, (argc <= 1 ? _("too few arguments")
-		    : _("too many arguments")));
+      error (0, 0, _("missing operand"));
       usage (EXIT_FAILURE);
     }
 
-  name = base_name (argv[1]);
+  if (optind + 2 < argc)
+    {
+      error (0, 0, _("extra operand %s"), quote (argv[optind + 2]));
+      usage (EXIT_FAILURE);
+    }
+
+  name = base_name (argv[optind]);
   name[base_len (name)] = '\0';
 
-  if (argc == 3)
-    remove_suffix (name, argv[2]);
+  if (argc == optind + 2)
+    remove_suffix (name, argv[optind + 1]);
 
   puts (name);
 
