@@ -1,5 +1,5 @@
 /* cat -- concatenate files and print on the standard output.
-   Copyright (C) 88, 90, 91, 1995-2002 Free Software Foundation, Inc.
+   Copyright (C) 88, 90, 91, 1995-2004 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -20,7 +20,7 @@
    * Usually much faster than other versions of cat, the difference
    is especially apparent when using the -v option.
 
-   By tege@sics.se, Torbjorn Granlund, advised by rms, Richard Stallman. */
+   By tege@sics.se, Torbjorn Granlund, advised by rms, Richard Stallman.  */
 
 #include <config.h>
 
@@ -31,15 +31,15 @@
 # include <sys/ioctl.h>
 #endif
 #include "system.h"
-#include "closeout.h"
 #include "error.h"
 #include "full-write.h"
+#include "getpagesize.h"
 #include "safe-read.h"
 
 /* The official name of this program (e.g., no `g' prefix).  */
 #define PROGRAM_NAME "cat"
 
-#define AUTHORS N_ ("Torbjorn Granlund and Richard M. Stallman")
+#define AUTHORS "Torbjorn Granlund", "Richard M. Stallman"
 
 /* Undefine, to avoid warning about redefinition on some systems.  */
 #undef max
@@ -78,13 +78,10 @@ static char *line_num_end = line_buf + LINE_COUNTER_BUF_LEN - 3;
 /* Preserves the `cat' function's local `newlines' between invocations.  */
 static int newlines2 = 0;
 
-/* Count of non-fatal error conditions.  */
-static int exit_status = 0;
-
 void
 usage (int status)
 {
-  if (status != 0)
+  if (status != EXIT_SUCCESS)
     fprintf (stderr, _("Try `%s --help' for more information.\n"),
 	     program_name);
   else
@@ -123,7 +120,7 @@ With no FILE, or when FILE is -, read standard input.\n\
 #endif
       printf (_("\nReport bugs to <%s>.\n"), PACKAGE_BUGREPORT);
     }
-  exit (status == 0 ? EXIT_SUCCESS : EXIT_FAILURE);
+  exit (status);
 }
 
 /* Compute the next line number.  */
@@ -147,19 +144,20 @@ next_line_num (void)
     line_num_print--;
 }
 
-/* Plain cat.  Copies the file behind `input_desc' to STDOUT_FILENO.  */
+/* Plain cat.  Copies the file behind `input_desc' to STDOUT_FILENO.
+   Return true if successful.  */
 
-static void
+static bool
 simple_cat (
      /* Pointer to the buffer, used by reads and writes.  */
      char *buf,
 
      /* Number of characters preferably read or written by each read and write
         call.  */
-     int bufsize)
+     size_t bufsize)
 {
   /* Actual number of characters read, and therefore written.  */
-  ssize_t n_read;
+  size_t n_read;
 
   /* Loop until the end of the file.  */
 
@@ -168,52 +166,56 @@ simple_cat (
       /* Read a block of input.  */
 
       n_read = safe_read (input_desc, buf, bufsize);
-      if (n_read < 0)
+      if (n_read == SAFE_READ_ERROR)
 	{
 	  error (0, errno, "%s", infile);
-	  exit_status = 1;
-	  return;
+	  return false;
 	}
 
       /* End of this file?  */
 
       if (n_read == 0)
-	break;
+	return true;
 
       /* Write this block out.  */
 
-      if (full_write (STDOUT_FILENO, buf, n_read) != n_read)
-	error (EXIT_FAILURE, errno, _("write error"));
+      {
+	/* The following is ok, since we know that 0 < n_read.  */
+	size_t n = n_read;
+	if (full_write (STDOUT_FILENO, buf, n) != n)
+	  error (EXIT_FAILURE, errno, _("write error"));
+      }
     }
 }
 
 /* Cat the file behind INPUT_DESC to the file behind OUTPUT_DESC.
+   Return true if successful.
    Called if any option more than -u was specified.
 
    A newline character is always put at the end of the buffer, to make
    an explicit test for buffer end unnecessary.  */
 
-static void
+static bool
 cat (
      /* Pointer to the beginning of the input buffer.  */
      char *inbuf,
 
      /* Number of characters read in each read call.  */
-     int insize,
+     size_t insize,
 
      /* Pointer to the beginning of the output buffer.  */
      char *outbuf,
 
      /* Number of characters written by each write call.  */
-     int outsize,
+     size_t outsize,
 
      /* Variables that have values according to the specified options.  */
-     int quote,
-     int output_tabs,
-     int numbers,
-     int numbers_at_empty_lines,
-     int mark_line_ends,
-     int squeeze_empty_lines)
+     bool show_nonprinting,
+     bool show_tabs,
+     bool number,
+     bool number_nonblank,
+     bool show_ends,
+     bool squeeze_blank)
 {
   /* Last character read from the input buffer.  */
   unsigned char ch;
@@ -229,7 +231,7 @@ cat (
   char *bpout;
 
   /* Number of characters read by the last read call.  */
-  ssize_t n_read;
+  size_t n_read;
 
   /* Determines how many consecutive newlines there have been in the
      input.  0 newlines makes NEWLINES -1, 1 newline makes NEWLINES 1,
@@ -240,8 +242,8 @@ cat (
 
 #ifdef FIONREAD
   /* If nonzero, use the FIONREAD ioctl, as an optimization.
-     (On Ultrix, it is not supported on NFS filesystems.)  */
-  int use_fionread = 1;
+     (On Ultrix, it is not supported on NFS file systems.)  */
+  bool use_fionread = true;
 #endif
 
   /* The inbuf pointers are initialized so that BPIN > EOB, and thereby input
@@ -258,22 +260,24 @@ cat (
 	{
 	  /* Write if there are at least OUTSIZE bytes in OUTBUF.  */
 
-	  if (bpout - outbuf >= outsize)
+	  if (outbuf + outsize <= bpout)
 	    {
 	      char *wp = outbuf;
+	      size_t remaining_bytes;
 	      do
 		{
 		  if (full_write (STDOUT_FILENO, wp, outsize) != outsize)
 		    error (EXIT_FAILURE, errno, _("write error"));
 		  wp += outsize;
+		  remaining_bytes = bpout - wp;
 		}
-	      while (bpout - wp >= outsize);
+	      while (outsize <= remaining_bytes);
 
 	      /* Move the remaining bytes to the beginning of the
 		 buffer.  */
 
-	      memmove (outbuf, wp, bpout - wp);
-	      bpout = outbuf + (bpout - wp);
+	      memmove (outbuf, wp, remaining_bytes);
+	      bpout = outbuf + remaining_bytes;
 	    }
 
 	  /* Is INBUF empty?  */
@@ -298,17 +302,13 @@ cat (
 		     Irix-5 returns ENOSYS on pipes.  */
 		  if (errno == EOPNOTSUPP || errno == ENOTTY
 		      || errno == EINVAL || errno == ENODEV
-# ifdef ENOSYS
-		      || errno == ENOSYS
-# endif
-		      )
-		    use_fionread = 0;
+		      || errno == ENOSYS)
+		    use_fionread = false;
 		  else
 		    {
 		      error (0, errno, _("cannot do ioctl on `%s'"), infile);
-		      exit_status = 1;
 		      newlines2 = newlines;
-		      return;
+		      return false;
 		    }
 		}
 	      if (n_to_read == 0)
@@ -324,17 +324,16 @@ cat (
 	      /* Read more input into INBUF.  */
 
 	      n_read = safe_read (input_desc, inbuf, insize);
-	      if (n_read < 0)
+	      if (n_read == SAFE_READ_ERROR)
 		{
 		  error (0, errno, "%s", infile);
-		  exit_status = 1;
 		  newlines2 = newlines;
-		  return;
+		  return false;
 		}
 	      if (n_read == 0)
 		{
 		  newlines2 = newlines;
-		  return;
+		  return true;
 		}
 
 	      /* Update the pointers and insert a sentinel at the buffer
@@ -363,7 +362,7 @@ cat (
 		      /* Are multiple adjacent empty lines to be substituted
 			 by single ditto (-s), and this was the second empty
 			 line?  */
-		      if (squeeze_empty_lines)
+		      if (squeeze_blank)
 			{
 			  ch = *bpin++;
 			  continue;
@@ -372,7 +371,7 @@ cat (
 
 		  /* Are line numbers to be written at empty lines (-n)?  */
 
-		  if (numbers && numbers_at_empty_lines)
+		  if (number & !number_nonblank)
 		    {
 		      next_line_num ();
 		      bpout = stpcpy (bpout, line_num_print);
@@ -381,7 +380,7 @@ cat (
 
 	      /* Output a currency symbol if requested (-e).  */
 
-	      if (mark_line_ends)
+	      if (show_ends)
 		*bpout++ = '$';
 
 	      /* Output the newline.  */
@@ -394,7 +393,7 @@ cat (
 
       /* Are we at the beginning of a line, and line numbers are requested?  */
 
-      if (newlines >= 0 && numbers)
+      if (newlines >= 0 && number)
 	{
 	  next_line_num ();
 	  bpout = stpcpy (bpout, line_num_print);
@@ -408,7 +407,7 @@ cat (
 
       /* If quoting, i.e. at least one of -v, -e, or -t specified,
 	 scan for chars that need conversion.  */
-      if (quote)
+      if (show_nonprinting)
 	{
 	  for (;;)
 	    {
@@ -442,7 +441,7 @@ cat (
 			}
 		    }
 		}
-	      else if (ch == '\t' && output_tabs)
+	      else if (ch == '\t' && !show_tabs)
 		*bpout++ = '\t';
 	      else if (ch == '\n')
 		{
@@ -463,7 +462,7 @@ cat (
 	  /* Not quoting, neither of -v, -e, or -t specified.  */
 	  for (;;)
 	    {
-	      if (ch == '\t' && !output_tabs)
+	      if (ch == '\t' && show_tabs)
 		{
 		  *bpout++ = '^';
 		  *bpout++ = ch + 64;
@@ -497,10 +496,12 @@ int
 main (int argc, char **argv)
 {
   /* Optimal size of i/o operations of output.  */
-  int outsize;
+  size_t outsize;
 
   /* Optimal size of i/o operations of input.  */
-  int insize;
+  size_t insize;
+
+  size_t page_size = getpagesize ();
 
   /* Pointer to the input buffer.  */
   char *inbuf;
@@ -508,6 +509,7 @@ main (int argc, char **argv)
   /* Pointer to the output buffer.  */
   char *outbuf;
 
+  bool ok = true;
   int c;
 
   /* Index in argv to processed argument.  */
@@ -519,30 +521,27 @@ main (int argc, char **argv)
   /* I-node number of the output.  */
   ino_t out_ino;
 
-  /* Nonzero if the output file should not be the same as any input file. */
-  int check_redirection = 1;
+  /* True if the output file should not be the same as any input file.  */
+  bool check_redirection = true;
 
-  /* Nonzero if we have ever read standard input. */
-  int have_read_stdin = 0;
+  /* Nonzero if we have ever read standard input.  */
+  bool have_read_stdin = false;
 
   struct stat stat_buf;
 
   /* Variables that are set according to the specified options.  */
-  int numbers = 0;
-  int numbers_at_empty_lines = 1;
-  int squeeze_empty_lines = 0;
-  int mark_line_ends = 0;
-  int quote = 0;
-  int output_tabs = 1;
+  bool number = false;
+  bool number_nonblank = false;
+  bool squeeze_blank = false;
+  bool show_ends = false;
+  bool show_nonprinting = false;
+  bool show_tabs = false;
 #if O_BINARY
-  int binary_files  = 0;
-  int binary_output = 0;
+  bool binary = false;
+  bool binary_output = false;
 #endif
   int file_open_mode = O_RDONLY;
 
-/* If nonzero, call cat, otherwise call simple_cat to do the actual work. */
-  int options = 0;
-
   static struct option const long_options[] =
   {
     {"number-nonblank", no_argument, NULL, 'b'},
@@ -560,6 +559,7 @@ main (int argc, char **argv)
     {NULL, 0, NULL, 0}
   };
 
+  initialize_main (&argc, &argv);
   program_name = argv[0];
   setlocale (LC_ALL, "");
   bindtextdomain (PACKAGE, LOCALEDIR);
@@ -581,35 +581,27 @@ main (int argc, char **argv)
     {
       switch (c)
 	{
-	case 0:
-	  break;
-
 	case 'b':
-	  ++options;
-	  numbers = 1;
-	  numbers_at_empty_lines = 0;
+	  number = true;
+	  number_nonblank = true;
 	  break;
 
 	case 'e':
-	  ++options;
-	  mark_line_ends = 1;
-	  quote = 1;
+	  show_ends = true;
+	  show_nonprinting = true;
 	  break;
 
 	case 'n':
-	  ++options;
-	  numbers = 1;
+	  number = true;
 	  break;
 
 	case 's':
-	  ++options;
-	  squeeze_empty_lines = 1;
+	  squeeze_blank = true;
 	  break;
 
 	case 't':
-	  ++options;
-	  output_tabs = 0;
-	  quote = 1;
+	  show_tabs = true;
+	  show_nonprinting = true;
 	  break;
 
 	case 'u':
@@ -617,32 +609,27 @@ main (int argc, char **argv)
 	  break;
 
 	case 'v':
-	  ++options;
-	  quote = 1;
+	  show_nonprinting = true;
 	  break;
 
 	case 'A':
-	  ++options;
-	  quote = 1;
-	  mark_line_ends = 1;
-	  output_tabs = 0;
+	  show_nonprinting = true;
+	  show_ends = true;
+	  show_tabs = true;
 	  break;
 
 #if O_BINARY
 	case 'B':
-	  ++options;
-	  binary_files = 1;
+	  binary = true;
 	  break;
 #endif
 
 	case 'E':
-	  ++options;
-	  mark_line_ends = 1;
+	  show_ends = true;
 	  break;
 
 	case 'T':
-	  ++options;
-	  output_tabs = 0;
+	  show_tabs = true;
 	  break;
 
 	case_GETOPT_HELP_CHAR;
@@ -667,7 +654,7 @@ main (int argc, char **argv)
      fstat on pipes returns S_IFSOCK on some systems, S_IFIFO
      on others, so the checking should not be done for those types,
      and to allow things like cat < /dev/tty > /dev/tty, checking
-     is not done for device files either. */
+     is not done for device files either.  */
 
   if (S_ISREG (stat_buf.st_mode))
     {
@@ -676,7 +663,7 @@ main (int argc, char **argv)
     }
   else
     {
-      check_redirection = 0;
+      check_redirection = false;
 #ifdef lint  /* Suppress `used before initialized' warning.  */
       out_dev = 0;
       out_ino = 0;
@@ -692,18 +679,18 @@ main (int argc, char **argv)
      with only CR-LF is an empty line.  (Besides, if they ask for
      one of these options, they don't care much about the original
      file contents anyway).  */
-  if ((!isatty (STDOUT_FILENO)
-       && !(numbers || squeeze_empty_lines || mark_line_ends))
-      || binary_files)
+  if (binary
+      || ! ((number | squeeze_blank | show_ends)
+	    || isatty (STDOUT_FILENO)))
     {
       /* Switch stdout to BINARY mode.  */
-      binary_output = 1;
+      binary_output = true;
       SET_BINARY (STDOUT_FILENO);
       /* When stdout is in binary mode, make sure all input files are
 	 also read in binary mode.  */
       file_open_mode |= O_BINARY;
     }
-  else if (quote)
+  else if (show_nonprinting)
     {
       /* If they want to see the non-printables, let's show them
 	 those CR characters as well, so make the input binary.
@@ -733,14 +720,14 @@ main (int argc, char **argv)
 
       if (infile[0] == '-' && infile[1] == 0)
 	{
-	  have_read_stdin = 1;
+	  have_read_stdin = true;
 	  input_desc = 0;
 
 #if O_BINARY
 	  /* Switch stdin to BINARY mode if needed.  */
 	  if (binary_output)
 	    {
-	      int tty_in = isatty (input_desc);
+	      bool tty_in = isatty (input_desc);
 
 	      /* If stdin is a terminal device, and it is the ONLY
 		 input file (i.e. we didn't write anything to the
@@ -769,7 +756,7 @@ main (int argc, char **argv)
 	  if (input_desc < 0)
 	    {
 	      error (0, errno, "%s", infile);
-	      exit_status = 1;
+	      ok = false;
 	      continue;
 	    }
 	}
@@ -777,7 +764,7 @@ main (int argc, char **argv)
       if (fstat (input_desc, &stat_buf) < 0)
 	{
 	  error (0, errno, "%s", infile);
-	  exit_status = 1;
+	  ok = false;
 	  goto contin;
 	}
       insize = ST_BLKSIZE (stat_buf);
@@ -792,26 +779,27 @@ main (int argc, char **argv)
 	  && (input_desc != STDIN_FILENO))
 	{
 	  error (0, 0, _("%s: input file is output file"), infile);
-	  exit_status = 1;
+	  ok = false;
 	  goto contin;
 	}
 
-      /* Select which version of `cat' to use. If any options (more than -u,
-	 --version, or --help) were specified, use `cat', otherwise use
-	 `simple_cat'.  */
+      /* Select which version of `cat' to use.  If any format-oriented
+	 options were given use `cat'; otherwise use `simple_cat'.  */
 
-      if (options == 0)
+      if (! (number | show_ends | show_nonprinting
+	     | show_tabs | squeeze_blank))
 	{
 	  insize = max (insize, outsize);
-	  inbuf = xmalloc (insize);
+	  inbuf = xmalloc (insize + page_size - 1);
 
-	  simple_cat (inbuf, insize);
+	  ok &= simple_cat (ptr_align (inbuf, page_size), insize);
 	}
       else
 	{
-	  inbuf = xmalloc (insize + 1);
+	  inbuf = xmalloc (insize + 1 + page_size - 1);
 
-	  /* Why are (OUTSIZE - 1 + INSIZE * 4 + LINE_COUNTER_BUF_LEN)
+	  /* Why are
+	     (OUTSIZE - 1 + INSIZE * 4 + LINE_COUNTER_BUF_LEN + PAGE_SIZE - 1)
 	     bytes allocated for the output buffer?
 
 	     A test whether output needs to be written is done when the input
@@ -826,13 +814,19 @@ main (int argc, char **argv)
 	     options) as the first thing in the output buffer. (Done after the
 	     new input is read, but before processing of the input begins.)
 	     A line number requires seldom more than LINE_COUNTER_BUF_LEN
-	     positions.  */
+	     positions.
 
-	  outbuf = xmalloc (outsize - 1 + insize * 4 + LINE_COUNTER_BUF_LEN);
+	     Align the output buffer to a page size boundary, for efficency on
+	     some paging implementations, so add PAGE_SIZE - 1 bytes to the
+	     request to make room for the alignment.  */
 
-	  cat (inbuf, insize, outbuf, outsize, quote,
-	       output_tabs, numbers, numbers_at_empty_lines, mark_line_ends,
-	       squeeze_empty_lines);
+	  outbuf = xmalloc (outsize - 1 + insize * 4 + LINE_COUNTER_BUF_LEN
+			    + page_size - 1);
+
+	  ok &= cat (ptr_align (inbuf, page_size), insize,
+		     ptr_align (outbuf, page_size), outsize, show_nonprinting,
+		     show_tabs, number, number_nonblank, show_ends,
+		     squeeze_blank);
 
 	  free (outbuf);
 	}
@@ -843,7 +837,7 @@ main (int argc, char **argv)
       if (!STREQ (infile, "-") && close (input_desc) < 0)
 	{
 	  error (0, errno, "%s", infile);
-	  exit_status = 1;
+	  ok = false;
 	}
     }
   while (++argind < argc);
@@ -854,5 +848,5 @@ main (int argc, char **argv)
   if (close (STDOUT_FILENO) < 0)
     error (EXIT_FAILURE, errno, _("closing standard output"));
 
-  exit (exit_status == 0 ? EXIT_SUCCESS : EXIT_FAILURE);
+  exit (ok ? EXIT_SUCCESS : EXIT_FAILURE);
 }
