@@ -1,5 +1,5 @@
 /* csplit - split a file into sections determined by context lines
-   Copyright (C) 91, 1995-2003 Free Software Foundation, Inc.
+   Copyright (C) 91, 1995-2004 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -32,17 +32,18 @@
 #include "error.h"
 #include "inttostr.h"
 #include "safe-read.h"
+#include "quote.h"
 #include "xstrtol.h"
 
+#ifndef SA_NOCLDSTOP
+# define sigprocmask(How, Set, Oset) /* empty */
+# define sigset_t int
+#endif
+
 /* The official name of this program (e.g., no `g' prefix).  */
 #define PROGRAM_NAME "csplit"
 
-#define AUTHORS N_ ("Stuart Kemp and David MacKenzie")
-
-#ifndef TRUE
-# define FALSE 0
-# define TRUE 1
-#endif
+#define AUTHORS "Stuart Kemp", "David MacKenzie"
 
 /* Increment size of area for control records. */
 #define ALLOC_SIZE 20
@@ -50,19 +51,17 @@
 /* The default prefix for output file names. */
 #define DEFAULT_PREFIX	"xx"
 
-typedef int boolean;
-
 /* A compiled pattern arg. */
 struct control
 {
   char *regexpr;		/* Non-compiled regular expression. */
   struct re_pattern_buffer re_compiled;	/* Compiled regular expression. */
-  int offset;			/* Offset from regexp to split at. */
+  intmax_t offset;		/* Offset from regexp to split at. */
   uintmax_t lines_required;	/* Number of lines required. */
   uintmax_t repeat;		/* Repeat count. */
-  int repeat_forever;		/* Non-zero if `*' used as a repeat count. */
   int argnum;			/* ARGV index. */
-  boolean ignore;		/* If true, produce no output (for regexp). */
+  bool repeat_forever;		/* True if `*' used as a repeat count. */
+  bool ignore;			/* If true, produce no output (for regexp). */
 };
 
 /* Initial size of data area in buffers. */
@@ -84,7 +83,7 @@ struct control
 /* A string with a length count. */
 struct cstring
 {
-  unsigned int len;
+  size_t len;
   char *str;
 };
 
@@ -92,9 +91,9 @@ struct cstring
    These structures are linked together if needed. */
 struct line
 {
-  unsigned used;		/* Number of offsets used in this struct. */
-  unsigned insert_index;	/* Next offset to use when inserting line. */
-  unsigned retrieve_index;	/* Next index to use when retrieving line. */
+  size_t used;			/* Number of offsets used in this struct. */
+  size_t insert_index;		/* Next offset to use when inserting line. */
+  size_t retrieve_index;	/* Next index to use when retrieving line. */
   struct cstring starts[CTRL_SIZE]; /* Lines in the data area. */
   struct line *next;		/* Next in linked list. */
 };
@@ -104,11 +103,11 @@ struct line
    pointers to the individual lines. */
 struct buffer_record
 {
-  unsigned bytes_alloc;		/* Size of the buffer area. */
-  unsigned bytes_used;		/* Bytes used in the buffer area. */
-  unsigned start_line;		/* First line number in this buffer. */
-  unsigned first_available;	/* First line that can be retrieved. */
-  unsigned num_lines;		/* Number of complete lines in this buffer. */
+  size_t bytes_alloc;		/* Size of the buffer area. */
+  size_t bytes_used;		/* Bytes used in the buffer area. */
+  uintmax_t start_line;		/* First line number in this buffer. */
+  uintmax_t first_available;	/* First line that can be retrieved. */
+  size_t num_lines;		/* Number of complete lines in this buffer. */
   char *buffer;			/* Data area. */
   struct line *line_start;	/* Head of list of pointers to lines. */
   struct line *curr_line;	/* The line start record currently in use. */
@@ -124,54 +123,44 @@ void usage (int status);
 /* The name this program was run with. */
 char *program_name;
 
-/* Convert the number of 8-bit bytes of a binary representation to
-   the number of characters required to represent the same quantity
-   as an unsigned octal.  For example, a 32-bit (4-byte) quantity may
-   require a field width as wide as 11 characters.  */
-static const unsigned int bytes_to_octal_digits[] =
-{0, 3, 6, 8, 11, 14, 16, 19, 22, 25, 27, 30, 32, 35, 38, 41, 43};
-
 /* Input file descriptor. */
 static int input_desc = 0;
 
-/* List of available buffers. */
-static struct buffer_record *free_list = NULL;
-
 /* Start of buffer list. */
 static struct buffer_record *head = NULL;
 
 /* Partially read line. */
 static char *hold_area = NULL;
 
-/* Number of chars in `hold_area'. */
-static unsigned hold_count = 0;
+/* Number of bytes in `hold_area'. */
+static size_t hold_count = 0;
 
 /* Number of the last line in the buffers. */
-static unsigned last_line_number = 0;
+static uintmax_t last_line_number = 0;
 
 /* Number of the line currently being examined. */
-static unsigned current_line = 0;
+static uintmax_t current_line = 0;
 
-/* If TRUE, we have read EOF. */
-static boolean have_read_eof = FALSE;
+/* If true, we have read EOF. */
+static bool have_read_eof = false;
 
 /* Name of output files. */
-static char *filename_space = NULL;
+static char * volatile filename_space = NULL;
 
 /* Prefix part of output file names. */
-static char *prefix = NULL;
+static char * volatile prefix = NULL;
 
 /* Suffix part of output file names. */
-static char *suffix = NULL;
+static char * volatile suffix = NULL;
 
 /* Number of digits to use in output file names. */
-static int digits = 2;
+static int volatile digits = 2;
 
 /* Number of files created so far. */
-static unsigned int files_created = 0;
+static unsigned int volatile files_created = 0;
 
 /* Number of bytes written to current file. */
-static unsigned int bytes_written;
+static uintmax_t bytes_written;
 
 /* Output file pointer. */
 static FILE *output_stream = NULL;
@@ -182,21 +171,24 @@ static char *output_filename = NULL;
 /* Perhaps it would be cleaner to pass arg values instead of indexes. */
 static char **global_argv;
 
-/* If TRUE, do not print the count of bytes in each output file. */
-static boolean suppress_count;
+/* If true, do not print the count of bytes in each output file. */
+static bool suppress_count;
 
-/* If TRUE, remove output files on error. */
-static boolean remove_files;
+/* If true, remove output files on error. */
+static bool volatile remove_files;
 
-/* If TRUE, remove all output files which have a zero length. */
-static boolean elide_empty_files;
+/* If true, remove all output files which have a zero length. */
+static bool elide_empty_files;
 
 /* The compiled pattern arguments, which determine how to split
    the input file. */
 static struct control *controls;
 
 /* Number of elements in `controls'. */
-static unsigned int control_used;
+static size_t control_used;
+
+/* The set of signals that are caught.  */
+static sigset_t caught_signals;
 
 static struct option const longopts[] =
 {
@@ -218,13 +210,16 @@ static struct option const longopts[] =
 static void
 cleanup (void)
 {
-  if (output_stream)
-    close_output_file ();
+  sigset_t oldset;
 
-  if (remove_files)
-    delete_all_files ();
+  close_output_file ();
+
+  sigprocmask (SIG_BLOCK, &caught_signals, &oldset);
+  delete_all_files ();
+  sigprocmask (SIG_SETMASK, &oldset, NULL);
 }
 
+static void cleanup_fatal (void) ATTRIBUTE_NORETURN;
 static void
 cleanup_fatal (void)
 {
@@ -232,46 +227,46 @@ cleanup_fatal (void)
   exit (EXIT_FAILURE);
 }
 
-static RETSIGTYPE
+extern void
+xalloc_die (void)
+{
+  error (0, 0, "%s", _("memory exhausted"));
+  cleanup_fatal ();
+}
+
+static void
 interrupt_handler (int sig)
 {
-#ifdef SA_NOCLDSTOP
-  struct sigaction sigact;
-
-  sigact.sa_handler = SIG_DFL;
-  sigemptyset (&sigact.sa_mask);
-  sigact.sa_flags = 0;
-  sigaction (sig, &sigact, NULL);
-#else
-  signal (sig, SIG_DFL);
+#ifndef SA_NOCLDSTOP
+  signal (sig, SIG_IGN);
 #endif
-  cleanup ();
+
+  delete_all_files ();
+
+  signal (sig, SIG_DFL);
   raise (sig);
 }
 
-/* Keep track of NUM chars of a partial line in buffer START.
-   These chars will be retrieved later when another large buffer is read.
-   It is not necessary to create a new buffer for these chars; instead,
+/* Keep track of NUM bytes of a partial line in buffer START.
+   These bytes will be retrieved later when another large buffer is read.
+   It is not necessary to create a new buffer for these bytes; instead,
    we keep a pointer to the existing buffer.  This buffer *is* on the
    free list, and when the next buffer is obtained from this list
-   (even if it is this one), these chars will be placed at the
+   (even if it is this one), these bytes will be placed at the
    start of the new buffer. */
 
 static void
-save_to_hold_area (char *start, unsigned int num)
+save_to_hold_area (char *start, size_t num)
 {
   hold_area = start;
   hold_count = num;
 }
 
-/* Read up to MAX_N_BYTES chars from the input stream into DEST.
-   Return the number of chars read. */
-/* FIXME: MAX_N_BYTES should be of type size_t, but if you pull
-   that thread, you'll find there are many other `unsigned' types
-   in this file that should also be changed.  */
+/* Read up to MAX_N_BYTES bytes from the input stream into DEST.
+   Return the number of bytes read. */
 
 static size_t
-read_input (char *dest, int max_n_bytes)
+read_input (char *dest, size_t max_n_bytes)
 {
   size_t bytes_read;
 
@@ -281,7 +276,7 @@ read_input (char *dest, int max_n_bytes)
   bytes_read = safe_read (input_desc, dest, max_n_bytes);
 
   if (bytes_read == 0)
-    have_read_eof = TRUE;
+    have_read_eof = true;
 
   if (bytes_read == SAFE_READ_ERROR)
     {
@@ -302,25 +297,12 @@ clear_line_control (struct line *p)
   p->retrieve_index = 0;
 }
 
-/* Initialize all line records in B. */
-
-static void
-clear_all_line_control (struct buffer_record *b)
-{
-  struct line *l;
-
-  for (l = b->line_start; l; l = l->next)
-    clear_line_control (l);
-}
-
 /* Return a new, initialized line record. */
 
 static struct line *
 new_line_control (void)
 {
-  struct line *p;
-
-  p = xmalloc (sizeof (struct line));
+  struct line *p = xmalloc (sizeof *p);
 
   p->next = NULL;
   clear_line_control (p);
@@ -332,7 +314,7 @@ new_line_control (void)
    of length LINE_LEN in the large buffer, in the lines buffer of B. */
 
 static void
-keep_new_line (struct buffer_record *b, char *line_start, int line_len)
+keep_new_line (struct buffer_record *b, char *line_start, size_t line_len)
 {
   struct line *l;
 
@@ -364,14 +346,14 @@ keep_new_line (struct buffer_record *b, char *line_start, int line_len)
    a pointer is kept to this area, which will be used when
    the next buffer is filled. */
 
-static unsigned int
+static size_t
 record_line_starts (struct buffer_record *b)
 {
   char *line_start;		/* Start of current line. */
   char *line_end;		/* End of each line found. */
-  unsigned int bytes_left;	/* Length of incomplete last line. */
-  unsigned int lines;		/* Number of lines found. */
-  unsigned int line_length;	/* Length of each line found. */
+  size_t bytes_left;		/* Length of incomplete last line. */
+  size_t lines;			/* Number of lines found. */
+  size_t line_length;		/* Length of each line found. */
 
   if (b->bytes_used == 0)
     return 0;
@@ -415,12 +397,9 @@ record_line_starts (struct buffer_record *b)
    an extra byte for safety. */
 
 static struct buffer_record *
-create_new_buffer (unsigned int size)
+create_new_buffer (size_t size)
 {
-  struct buffer_record *new_buffer;
-
-  new_buffer = (struct buffer_record *)
-    xmalloc (sizeof (struct buffer_record));
+  struct buffer_record *new_buffer = xmalloc (sizeof *new_buffer);
 
   new_buffer->buffer = xmalloc (size + 1);
 
@@ -434,45 +413,20 @@ create_new_buffer (unsigned int size)
    least that size is currently free, use it, otherwise create a new one. */
 
 static struct buffer_record *
-get_new_buffer (unsigned int min_size)
+get_new_buffer (size_t min_size)
 {
-  struct buffer_record *p, *q;
   struct buffer_record *new_buffer; /* Buffer to return. */
-  unsigned int alloc_size;	/* Actual size that will be requested. */
+  size_t alloc_size;	/* Actual size that will be requested. */
 
   alloc_size = START_SIZE;
-  while (min_size > alloc_size)
-    alloc_size += INCR_SIZE;
-
-  if (free_list == NULL)
-    new_buffer = create_new_buffer (alloc_size);
-  else
+  if (alloc_size < min_size)
     {
-      /* Use first-fit to find a buffer. */
-      p = new_buffer = NULL;
-      q = free_list;
-
-      do
-	{
-	  if (q->bytes_alloc >= min_size)
-	    {
-	      if (p == NULL)
-		free_list = q->next;
-	      else
-		p->next = q->next;
-	      break;
-	    }
-	  p = q;
-	  q = q->next;
-	}
-      while (q);
-
-      new_buffer = (q ? q : create_new_buffer (alloc_size));
-
-      new_buffer->curr_line = new_buffer->line_start;
-      clear_all_line_control (new_buffer);
+      size_t s = min_size - alloc_size + INCR_SIZE - 1;
+      alloc_size += s - s % INCR_SIZE;
     }
 
+  new_buffer = create_new_buffer (alloc_size);
+
   new_buffer->num_lines = 0;
   new_buffer->bytes_used = 0;
   new_buffer->start_line = new_buffer->first_available = last_line_number + 1;
@@ -481,13 +435,10 @@ get_new_buffer (unsigned int min_size)
   return new_buffer;
 }
 
-/* Add buffer BUF to the list of free buffers. */
-
 static void
 free_buffer (struct buffer_record *buf)
 {
-  buf->next = free_list;
-  free_list = buf;
+  free (buf->buffer);
 }
 
 /* Append buffer BUF to the linked list of buffers that contain
@@ -521,20 +472,20 @@ save_buffer (struct buffer_record *buf)
    hold area) and repeat the process with another large buffer until at least
    one entire line has been read.
 
-   Return TRUE if a new buffer was obtained, otherwise false
+   Return true if a new buffer was obtained, otherwise false
    (in which case end-of-file must have been encountered). */
 
-static boolean
+static bool
 load_buffer (void)
 {
   struct buffer_record *b;
-  unsigned int bytes_wanted = START_SIZE; /* Minimum buffer size. */
-  unsigned int bytes_avail;		/* Size of new buffer created. */
-  unsigned int lines_found;		/* Number of lines in this new buffer. */
+  size_t bytes_wanted = START_SIZE; /* Minimum buffer size. */
+  size_t bytes_avail;		/* Size of new buffer created. */
+  size_t lines_found;		/* Number of lines in this new buffer. */
   char *p;			/* Place to load into buffer. */
 
   if (have_read_eof)
-    return FALSE;
+    return false;
 
   /* We must make the buffer at least as large as the amount of data
      in the partial line left over from the last call. */
@@ -558,7 +509,7 @@ load_buffer (void)
 	  hold_count = 0;
 	}
 
-      b->bytes_used += (unsigned int) read_input (p, bytes_avail);
+      b->bytes_used += read_input (p, bytes_avail);
 
       lines_found = record_line_starts (b);
       bytes_wanted = b->bytes_alloc * 2;
@@ -575,7 +526,7 @@ load_buffer (void)
 
 /* Return the line number of the first line that has not yet been retrieved. */
 
-static unsigned int
+static uintmax_t
 get_first_line_in_buffer (void)
 {
   if (head == NULL && !load_buffer ())
@@ -591,9 +542,20 @@ get_first_line_in_buffer (void)
 static struct cstring *
 remove_line (void)
 {
+  /* If non-NULL, this is the buffer for which the previous call
+     returned the final line.  So now, presuming that line has been
+     processed, we can free the buffer and reset this pointer.  */
+  static struct buffer_record *prev_buf = NULL;
+
   struct cstring *line;		/* Return value. */
   struct line *l;		/* For convenience. */
 
+  if (prev_buf)
+    {
+      free_buffer (prev_buf);
+      prev_buf = NULL;
+    }
+
   if (head == NULL && !load_buffer ())
     return NULL;
 
@@ -613,10 +575,11 @@ remove_line (void)
       head->curr_line = l->next;
       if (head->curr_line == NULL || head->curr_line->used == 0)
 	{
-	  /* Go on to the next data block. */
-	  struct buffer_record *b = head;
+	  /* Go on to the next data block.
+	     but first record the current one so we can free it
+	     once the line we're returning has been processed.  */
+	  prev_buf = head;
 	  head = head->next;
-	  free_buffer (b);
 	}
     }
 
@@ -627,7 +590,7 @@ remove_line (void)
    Return a pointer to the line, or NULL if it is not found in the file. */
 
 static struct cstring *
-find_line (unsigned int linenum)
+find_line (uintmax_t linenum)
 {
   struct buffer_record *b;
 
@@ -643,7 +606,7 @@ find_line (unsigned int linenum)
 	{
 	  /* The line is in this buffer. */
 	  struct line *l;
-	  unsigned int offset;	/* How far into the buffer the line is. */
+	  size_t offset;	/* How far into the buffer the line is. */
 
 	  l = b->line_start;
 	  offset = linenum - b->start_line;
@@ -661,12 +624,12 @@ find_line (unsigned int linenum)
     }
 }
 
-/* Return TRUE if at least one more line is available for input. */
+/* Return true if at least one more line is available for input. */
 
-static boolean
+static bool
 no_more_lines (void)
 {
-  return (find_line (current_line + 1) == NULL) ? TRUE : FALSE;
+  return find_line (current_line + 1) == NULL;
 }
 
 /* Set the name of the input file to NAME and open it. */
@@ -686,16 +649,16 @@ set_input_file (const char *name)
 
 /* Write all lines from the beginning of the buffer up to, but
    not including, line LAST_LINE, to the current output file.
-   If IGNORE is TRUE, do not output lines selected here.
+   If IGNORE is true, do not output lines selected here.
    ARGNUM is the index in ARGV of the current pattern. */
 
 static void
-write_to_file (unsigned int last_line, boolean ignore, int argnum)
+write_to_file (uintmax_t last_line, bool ignore, int argnum)
 {
   struct cstring *line;
-  unsigned int first_line;		/* First available input line. */
-  unsigned int lines;		/* Number of lines to output. */
-  unsigned int i;
+  uintmax_t first_line;		/* First available input line. */
+  uintmax_t lines;		/* Number of lines to output. */
+  uintmax_t i;
 
   first_line = get_first_line_in_buffer ();
 
@@ -734,15 +697,17 @@ dump_rest_of_file (void)
 /* Handle an attempt to read beyond EOF under the control of record P,
    on iteration REPETITION if nonzero. */
 
+static void handle_line_error (const struct control *, uintmax_t)
+     ATTRIBUTE_NORETURN;
 static void
-handle_line_error (const struct control *p, int repetition)
+handle_line_error (const struct control *p, uintmax_t repetition)
 {
   char buf[INT_BUFSIZE_BOUND (uintmax_t)];
 
   fprintf (stderr, _("%s: `%s': line number out of range"),
 	   program_name, umaxtostr (p->lines_required, buf));
   if (repetition)
-    fprintf (stderr, _(" on repetition %d\n"), repetition);
+    fprintf (stderr, _(" on repetition %s\n"), umaxtostr (repetition, buf));
   else
     fprintf (stderr, "\n");
 
@@ -755,9 +720,9 @@ handle_line_error (const struct control *p, int repetition)
    REPETITION is the repetition number. */
 
 static void
-process_line_count (const struct control *p, int repetition)
+process_line_count (const struct control *p, uintmax_t repetition)
 {
-  unsigned int linenum;
+  uintmax_t linenum;
   uintmax_t last_line_to_save = p->lines_required * (repetition + 1);
   struct cstring *line;
 
@@ -781,14 +746,18 @@ process_line_count (const struct control *p, int repetition)
     handle_line_error (p, repetition);
 }
 
+static void regexp_error (struct control *, uintmax_t, bool) ATTRIBUTE_NORETURN;
 static void
-regexp_error (struct control *p, int repetition, boolean ignore)
+regexp_error (struct control *p, uintmax_t repetition, bool ignore)
 {
   fprintf (stderr, _("%s: `%s': match not found"),
 	   program_name, global_argv[p->argnum]);
 
   if (repetition)
-    fprintf (stderr, _(" on repetition %d\n"), repetition);
+    {
+      char buf[INT_BUFSIZE_BOUND (uintmax_t)];
+      fprintf (stderr, _(" on repetition %s\n"), umaxtostr (repetition, buf));
+    }
   else
     fprintf (stderr, "\n");
 
@@ -801,16 +770,16 @@ regexp_error (struct control *p, int repetition, boolean ignore)
 }
 
 /* Read the input until a line matches the regexp in P, outputting
-   it unless P->IGNORE is TRUE.
+   it unless P->IGNORE is true.
    REPETITION is this repeat-count; 0 means the first time. */
 
 static void
-process_regexp (struct control *p, int repetition)
+process_regexp (struct control *p, uintmax_t repetition)
 {
   struct cstring *line;		/* From input file. */
-  unsigned int line_len;	/* To make "$" in regexps work. */
-  unsigned int break_line;	/* First line number of next file. */
-  boolean ignore = p->ignore;	/* If TRUE, skip this section. */
+  size_t line_len;		/* To make "$" in regexps work. */
+  uintmax_t break_line;		/* First line number of next file. */
+  bool ignore = p->ignore;	/* If true, skip this section. */
   int ret;
 
   if (!ignore)
@@ -842,7 +811,7 @@ process_regexp (struct control *p, int repetition)
 	  if (line->str[line_len - 1] == '\n')
 	    line_len--;
 	  ret = re_search (&p->re_compiled, line->str, line_len,
-			   0, line_len, (struct re_registers *) 0);
+			   0, line_len, NULL);
 	  if (ret == -2)
 	    {
 	      error (0, 0, _("error in regular expression search"));
@@ -882,7 +851,7 @@ process_regexp (struct control *p, int repetition)
 	  if (line->str[line_len - 1] == '\n')
 	    line_len--;
 	  ret = re_search (&p->re_compiled, line->str, line_len,
-			   0, line_len, (struct re_registers *) 0);
+			   0, line_len, NULL);
 	  if (ret == -2)
 	    {
 	      error (0, 0, _("error in regular expression search"));
@@ -910,10 +879,11 @@ process_regexp (struct control *p, int repetition)
 static void
 split_file (void)
 {
-  unsigned int i, j;
+  size_t i;
 
   for (i = 0; i < control_used; i++)
     {
+      uintmax_t j;
       if (controls[i].regexpr)
 	{
 	  for (j = 0; (controls[i].repeat_forever
@@ -942,7 +912,7 @@ make_filename (unsigned int num)
   if (suffix)
     sprintf (filename_space+strlen(prefix), suffix, num);
   else
-    sprintf (filename_space+strlen(prefix), "%0*d", digits, num);
+    sprintf (filename_space+strlen(prefix), "%0*u", digits, num);
   return filename_space;
 }
 
@@ -951,31 +921,47 @@ make_filename (unsigned int num)
 static void
 create_output_file (void)
 {
+  sigset_t oldset;
+  bool fopen_ok;
+  int fopen_errno;
+
   output_filename = make_filename (files_created);
+
+  /* Create the output file in a critical section, to avoid races.  */
+  sigprocmask (SIG_BLOCK, &caught_signals, &oldset);
   output_stream = fopen (output_filename, "w");
-  if (output_stream == NULL)
+  fopen_ok = (output_stream != NULL);
+  fopen_errno = errno;
+  files_created += fopen_ok;
+  sigprocmask (SIG_SETMASK, &oldset, NULL);
+
+  if (! fopen_ok)
     {
-      error (0, errno, "%s", output_filename);
+      error (0, fopen_errno, "%s", output_filename);
       cleanup_fatal ();
     }
-  files_created++;
   bytes_written = 0;
 }
 
-/* Delete all the files we have created. */
+/* If requested, delete all the files we have created.  This function
+   must be called only from critical sections.  */
 
 static void
 delete_all_files (void)
 {
   unsigned int i;
-  char *name;
+
+  if (! remove_files)
+    return;
 
   for (i = 0; i < files_created; i++)
     {
-      name = make_filename (i);
+      const char *name = make_filename (i);
       if (unlink (name))
 	error (0, errno, "%s", name);
     }
+
+  files_created = 0;
 }
 
 /* Close the current output file and print the count
@@ -986,24 +972,41 @@ close_output_file (void)
 {
   if (output_stream)
     {
-      if (ferror (output_stream) || fclose (output_stream) == EOF)
+      if (ferror (output_stream))
 	{
-	  error (0, errno, _("write error for `%s'"), output_filename);
+	  error (0, 0, _("write error for `%s'"), output_filename);
+	  output_stream = NULL;
+	  cleanup_fatal ();
+	}
+      if (fclose (output_stream) != 0)
+	{
+	  error (0, errno, "%s", output_filename);
 	  output_stream = NULL;
 	  cleanup_fatal ();
 	}
       if (bytes_written == 0 && elide_empty_files)
 	{
-	  if (unlink (output_filename))
-	    error (0, errno, "%s", output_filename);
-	  files_created--;
+	  sigset_t oldset;
+	  bool unlink_ok;
+	  int unlink_errno;
+
+	  /* Remove the output file in a critical section, to avoid races.  */
+	  sigprocmask (SIG_BLOCK, &caught_signals, &oldset);
+	  unlink_ok = (unlink (output_filename) == 0);
+	  unlink_errno = errno;
+	  files_created -= unlink_ok;
+	  sigprocmask (SIG_SETMASK, &oldset, NULL);
+
+	  if (! unlink_ok)
+	    error (0, unlink_errno, "%s", output_filename);
 	}
       else
 	{
-	  /* FIXME: if we write to stdout here, we have to close stdout
-	     and check for errors.  */
 	  if (!suppress_count)
-	    fprintf (stdout, "%d\n", bytes_written);
+	    {
+	      char buf[INT_BUFSIZE_BOUND (uintmax_t)];
+	      fprintf (stdout, "%s\n", umaxtostr (bytes_written, buf));
+	    }
 	}
       output_stream = NULL;
     }
@@ -1024,26 +1027,15 @@ save_line_to_file (const struct cstring *line)
 static struct control *
 new_control_record (void)
 {
-  static unsigned control_allocated = 0; /* Total space allocated. */
+  static size_t control_allocated = 0; /* Total space allocated. */
   struct control *p;
 
-  if (control_allocated == 0)
-    {
-      control_allocated = ALLOC_SIZE;
-      controls = (struct control *)
-	xmalloc (sizeof (struct control) * control_allocated);
-    }
-  else if (control_used == control_allocated)
-    {
-      control_allocated += ALLOC_SIZE;
-      controls = (struct control *)
-	xrealloc (controls,
-		  sizeof (struct control) * control_allocated);
-    }
+  if (control_used == control_allocated)
+    controls = x2nrealloc (controls, &control_allocated, sizeof *controls);
   p = &controls[control_used++];
   p->regexpr = NULL;
   p->repeat = 0;
-  p->repeat_forever = 0;
+  p->repeat_forever = false;
   p->lines_required = 0;
   p->offset = 0;
   return p;
@@ -1057,18 +1049,8 @@ new_control_record (void)
 static void
 check_for_offset (struct control *p, const char *str, const char *num)
 {
-  unsigned long val;
-
-  if (*num != '-' && *num != '+')
-    error (EXIT_FAILURE, 0, _("%s: `+' or `-' expected after delimeter"), str);
-
-  if (xstrtoul (num + 1, NULL, 10, &val, "") != LONGINT_OK
-      || val > UINT_MAX)
-    error (EXIT_FAILURE, 0, _("%s: integer expected after `%c'"), str, *num);
-  p->offset = (unsigned int) val;
-
-  if (*num == '-')
-    p->offset = -p->offset;
+  if (xstrtoimax (num, NULL, 10, &p->offset, "") != LONGINT_OK)
+    error (EXIT_FAILURE, 0, _("%s: integer expected after delimiter"), str);
 }
 
 /* Given that the first character of command line arg STR is '{',
@@ -1088,7 +1070,7 @@ parse_repeat_count (int argnum, struct control *p, char *str)
   *end = '\0';
 
   if (str+1 == end-1 && *(str+1) == '*')
-    p->repeat_forever = 1;
+    p->repeat_forever = true;
   else
     {
       if (xstrtoumax (str + 1, NULL, 10, &val, "") != LONGINT_OK)
@@ -1107,12 +1089,12 @@ parse_repeat_count (int argnum, struct control *p, char *str)
    STR should start with the regexp delimiter character.
    Return a new control record for the regular expression.
    ARGNUM is the ARGV index of STR.
-   Unless IGNORE is TRUE, mark these lines for output. */
+   Unless IGNORE is true, mark these lines for output. */
 
 static struct control *
-extract_regexp (int argnum, boolean ignore, char *str)
+extract_regexp (int argnum, bool ignore, char *str)
 {
-  int len;			/* Number of chars in this regexp. */
+  size_t len;			/* Number of bytes in this regexp. */
   char delim = *str;
   char *closing_delim;
   struct control *p;
@@ -1121,18 +1103,18 @@ extract_regexp (int argnum, boolean ignore, char *str)
   closing_delim = strrchr (str + 1, delim);
   if (closing_delim == NULL)
     error (EXIT_FAILURE, 0,
-	   _("%s: closing delimeter `%c' missing"), str, delim);
+	   _("%s: closing delimiter `%c' missing"), str, delim);
 
   len = closing_delim - str - 1;
   p = new_control_record ();
   p->argnum = argnum;
   p->ignore = ignore;
 
-  p->regexpr = xmalloc ((unsigned) (len + 1));
+  p->regexpr = xmalloc (len + 1);
   strncpy (p->regexpr, str + 1, len);
   p->re_compiled.allocated = len * 2;
   p->re_compiled.buffer = xmalloc (p->re_compiled.allocated);
-  p->re_compiled.fastmap = xmalloc (256);
+  p->re_compiled.fastmap = xmalloc (1 << CHAR_BIT);
   p->re_compiled.translate = 0;
   err = re_compile_pattern (p->regexpr, len, &p->re_compiled);
   if (err)
@@ -1202,10 +1184,10 @@ parse_patterns (int argc, int start, char **argv)
     }
 }
 
-static unsigned
+static unsigned int
 get_format_flags (char **format_ptr)
 {
-  unsigned count = 0;
+  unsigned int count = 0;
 
   for (; **format_ptr; (*format_ptr)++)
     {
@@ -1216,11 +1198,11 @@ get_format_flags (char **format_ptr)
 
 	case '+':
 	case ' ':
-	  count++;
+	  count |= 1;
 	  break;
 
 	case '#':
-	  count += 2;	/* Allow for 0x prefix preceeding an `x' conversion.  */
+	  count |= 2;	/* Allow for 0x prefix preceding an `x' conversion.  */
 	  break;
 
 	default:
@@ -1230,74 +1212,44 @@ get_format_flags (char **format_ptr)
   return count;
 }
 
-static unsigned
+static size_t
 get_format_width (char **format_ptr)
 {
-  unsigned count = 0;
-  char *start;
-  int ch_save;
+  unsigned long int val = 0;
 
-  start = *format_ptr;
-  for (; ISDIGIT (**format_ptr); (*format_ptr)++)
-    continue;
+  if (ISDIGIT (**format_ptr)
+      && (xstrtoul (*format_ptr, format_ptr, 10, &val, NULL) != LONGINT_OK
+	  || SIZE_MAX < val))
+    error (EXIT_FAILURE, 0, _("invalid format width"));
 
-  ch_save = **format_ptr;
-  **format_ptr = '\0';
-  /* In the case where no minimum field width is explicitly specified,
-     allow for enough octal digits to represent the value of LONG_MAX.  */
-  count = ((*format_ptr == start)
-	   ? bytes_to_octal_digits[sizeof (long)]
-	   /* FIXME: don't use atoi, it may silently overflow.
-	      Besides, we know the result is non-negative, so shouldn't
-	      need that cast.  */
-	   : (unsigned) atoi (start));
-  **format_ptr = ch_save;
-  return count;
+  /* Allow for enough octal digits to represent the value of UINT_MAX,
+     even if the field width is less than that.  */
+  return MAX (val, (sizeof (unsigned int) * CHAR_BIT + 2) / 3);
 }
 
-static unsigned
+static size_t
 get_format_prec (char **format_ptr)
 {
-  unsigned count = 0;
-  char *start;
-  int ch_save;
-  int is_negative;
-
   if (**format_ptr != '.')
     return 0;
   (*format_ptr)++;
 
-  if (**format_ptr == '-' || **format_ptr == '+')
-    {
-      is_negative = (**format_ptr == '-');
-      (*format_ptr)++;
-    }
+  if (! ISDIGIT (**format_ptr))
+    return 0;
   else
     {
-      is_negative = 0;
+      unsigned long int val;
+      if (xstrtoul (*format_ptr, format_ptr, 10, &val, NULL) != LONGINT_OK
+	  || SIZE_MAX < val)
+	error (EXIT_FAILURE, 0, _("invalid format precision"));
+      return val;
     }
-
-  start = *format_ptr;
-  for (; ISDIGIT (**format_ptr); (*format_ptr)++)
-    continue;
-
-  /* ANSI 4.9.6.1 says that if the precision is negative, it's as good as
-     not there. */
-  if (is_negative)
-    start = *format_ptr;
-
-  ch_save = **format_ptr;
-  **format_ptr = '\0';
-  count = (*format_ptr == start) ? 11 : atoi (start);
-  **format_ptr = ch_save;
-
-  return count;
 }
 
 static void
 get_format_conv_type (char **format_ptr)
 {
-  int ch = *((*format_ptr)++);
+  unsigned char ch = *(*format_ptr)++;
 
   switch (ch)
     {
@@ -1323,25 +1275,31 @@ get_format_conv_type (char **format_ptr)
     }
 }
 
-static unsigned
+static size_t
 max_out (char *format)
 {
-  unsigned out_count = 0;
-  unsigned percents = 0;
+  size_t out_count = 0;
+  bool percent = false;
 
-  for (; *format; )
+  while (*format)
     {
-      int ch = *format++;
-
-      if (ch != '%')
+      if (*format++ != '%')
         out_count++;
+      else if (*format == '%')
+	{
+	  format++;
+	  out_count++;
+	}
       else
 	{
-	  percents++;
+	  if (percent)
+	    error (EXIT_FAILURE, 0,
+		   _("too many %% conversion specifications in suffix"));
+	  percent = true;
 	  out_count += get_format_flags (&format);
 	  {
-	    int width = get_format_width (&format);
-	    int prec = get_format_prec (&format);
+	    size_t width = get_format_width (&format);
+	    size_t prec = get_format_prec (&format);
 
 	    out_count += MAX (width, prec);
 	  }
@@ -1349,12 +1307,9 @@ max_out (char *format)
 	}
     }
 
-  if (percents == 0)
+  if (! percent)
     error (EXIT_FAILURE, 0,
 	   _("missing %% conversion specification in suffix"));
-  else if (percents > 1)
-    error (EXIT_FAILURE, 0,
-	   _("too many %% conversion specifications in suffix"));
 
   return out_count;
 }
@@ -1363,10 +1318,7 @@ int
 main (int argc, char **argv)
 {
   int optc;
-  unsigned long val;
-#ifdef SA_NOCLDSTOP
-  struct sigaction oldact, newact;
-#endif
+  unsigned long int val;
 
   initialize_main (&argc, &argv);
   program_name = argv[0];
@@ -1379,50 +1331,13 @@ main (int argc, char **argv)
   global_argv = argv;
   controls = NULL;
   control_used = 0;
-  suppress_count = FALSE;
-  remove_files = TRUE;
+  suppress_count = false;
+  remove_files = true;
   prefix = DEFAULT_PREFIX;
 
-  /* Change the way xmalloc and xrealloc fail.  */
-  xalloc_fail_func = cleanup;
-
-#ifdef SA_NOCLDSTOP
-  newact.sa_handler = interrupt_handler;
-  sigemptyset (&newact.sa_mask);
-  newact.sa_flags = 0;
-
-  sigaction (SIGHUP, NULL, &oldact);
-  if (oldact.sa_handler != SIG_IGN)
-    sigaction (SIGHUP, &newact, NULL);
-
-  sigaction (SIGINT, NULL, &oldact);
-  if (oldact.sa_handler != SIG_IGN)
-    sigaction (SIGINT, &newact, NULL);
-
-  sigaction (SIGQUIT, NULL, &oldact);
-  if (oldact.sa_handler != SIG_IGN)
-    sigaction (SIGQUIT, &newact, NULL);
-
-  sigaction (SIGTERM, NULL, &oldact);
-  if (oldact.sa_handler != SIG_IGN)
-    sigaction (SIGTERM, &newact, NULL);
-#else
-  if (signal (SIGHUP, SIG_IGN) != SIG_IGN)
-    signal (SIGHUP, interrupt_handler);
-  if (signal (SIGINT, SIG_IGN) != SIG_IGN)
-    signal (SIGINT, interrupt_handler);
-  if (signal (SIGQUIT, SIG_IGN) != SIG_IGN)
-    signal (SIGQUIT, interrupt_handler);
-  if (signal (SIGTERM, SIG_IGN) != SIG_IGN)
-    signal (SIGTERM, interrupt_handler);
-#endif
-
   while ((optc = getopt_long (argc, argv, "f:b:kn:sqz", longopts, NULL)) != -1)
     switch (optc)
       {
-      case 0:
-	break;
-
       case 'f':
 	prefix = optarg;
 	break;
@@ -1432,23 +1347,23 @@ main (int argc, char **argv)
 	break;
 
       case 'k':
-	remove_files = FALSE;
+	remove_files = false;
 	break;
 
       case 'n':
 	if (xstrtoul (optarg, NULL, 10, &val, "") != LONGINT_OK
 	    || val > INT_MAX)
 	  error (EXIT_FAILURE, 0, _("%s: invalid number"), optarg);
-	digits = (int) val;
+	digits = val;
 	break;
 
       case 's':
       case 'q':
-	suppress_count = TRUE;
+	suppress_count = true;
 	break;
 
       case 'z':
-	elide_empty_files = TRUE;
+	elide_empty_files = true;
 	break;
 
       case_GETOPT_HELP_CHAR;
@@ -1461,7 +1376,10 @@ main (int argc, char **argv)
 
   if (argc - optind < 2)
     {
-      error (0, 0, _("too few arguments"));
+      if (argc <= optind)
+	error (0, 0, _("missing operand"));
+      else
+	error (0, 0, _("missing operand after %s"), quote (argv[argc - 1]));
       usage (EXIT_FAILURE);
     }
 
@@ -1474,6 +1392,36 @@ main (int argc, char **argv)
 
   parse_patterns (argc, optind, argv);
 
+  {
+    int i;
+    static int const sig[] = { SIGHUP, SIGINT, SIGQUIT, SIGTERM };
+    enum { nsigs = sizeof sig / sizeof sig[0] };
+
+#ifdef SA_NOCLDSTOP
+    struct sigaction act;
+
+    sigemptyset (&caught_signals);
+    for (i = 0; i < nsigs; i++)
+      {
+	sigaction (sig[i], NULL, &act);
+	if (act.sa_handler != SIG_IGN)
+	  sigaddset (&caught_signals, sig[i]);
+      }
+
+    act.sa_handler = interrupt_handler;
+    act.sa_mask = caught_signals;
+    act.sa_flags = 0;
+
+    for (i = 0; i < nsigs; i++)
+      if (sigismember (&caught_signals, sig[i]))
+	sigaction (sig[i], &act, NULL);
+#else
+    for (i = 0; i < nsigs; i++)
+      if (signal (sig[i], SIG_IGN) != SIG_IGN)
+	signal (sig[i], interrupt_handler);
+#endif
+  }
+
   split_file ();
 
   if (close (input_desc) < 0)
@@ -1488,7 +1436,7 @@ main (int argc, char **argv)
 void
 usage (int status)
 {
-  if (status != 0)
+  if (status != EXIT_SUCCESS)
     fprintf (stderr, _("Try `%s --help' for more information.\n"),
 	     program_name);
   else
@@ -1498,7 +1446,7 @@ Usage: %s [OPTION]... FILE PATTERN...\n\
 "),
 	      program_name);
       fputs (_("\
-Output pieces of FILE separated by PATTERN(s) to files `xx01', `xx02', ...,\n\
+Output pieces of FILE separated by PATTERN(s) to files `xx00', `xx01', ...,\n\
 and output byte counts of each piece to standard output.\n\
 \n\
 "), stdout);
@@ -1506,7 +1454,7 @@ and output byte counts of each piece to standard output.\n\
 Mandatory arguments to long options are mandatory for short options too.\n\
 "), stdout);
       fputs (_("\
-  -b, --suffix-format=FORMAT use sprintf FORMAT instead of %d\n\
+  -b, --suffix-format=FORMAT use sprintf FORMAT instead of %02d\n\
   -f, --prefix=PREFIX        use PREFIX instead of `xx'\n\
   -k, --keep-files           do not remove output files on errors\n\
 "), stdout);
@@ -1533,5 +1481,5 @@ A line OFFSET is a required `+' or `-' followed by a positive integer.\n\
 "), stdout);
       printf (_("\nReport bugs to <%s>.\n"), PACKAGE_BUGREPORT);
     }
-  exit (status == 0 ? EXIT_SUCCESS : EXIT_FAILURE);
+  exit (status);
 }
