@@ -30,6 +30,7 @@
 #include "error.h"
 #include "full-write.h"
 #include "getpagesize.h"
+#include "human.h"
 #include "inttostr.h"
 #include "long-options.h"
 #include "quote.h"
@@ -49,9 +50,9 @@
 # define S_TYPEISSHM(Stat_ptr) 0
 #endif
 
-#define ROUND_UP_OFFSET(X, M) ((M) - 1 - (((X) + (M) - 1) % (M)))
-#define PTR_ALIGN(Ptr, M) ((Ptr) \
-			   + ROUND_UP_OFFSET ((char *)(Ptr) - (char *)0, (M)))
+#if ! HAVE_FDATASYNC
+# define fdatasync(fd) (errno = ENOSYS, -1)
+#endif
 
 #define max(a, b) ((a) > (b) ? (a) : (b))
 #define output_char(c)				\
@@ -66,20 +67,43 @@
 /* Default input and output blocksize. */
 #define DEFAULT_BLOCKSIZE 512
 
+/* Maximum blocksize.  Keep it smaller than SIZE_MAX, so that we can
+   allocate buffers that size.  Keep it smaller than SSIZE_MAX, for
+   the benefit of system calls like "read".  And keep it smaller than
+   OFF_T_MAX, for the benefit of the large-offset seek code.  */
+#define MAX_BLOCKSIZE MIN (SIZE_MAX, MIN (SSIZE_MAX, OFF_T_MAX))
+
 /* Conversions bit masks. */
-#define C_ASCII 01
-#define C_EBCDIC 02
-#define C_IBM 04
-#define C_BLOCK 010
-#define C_UNBLOCK 020
-#define C_LCASE 040
-#define C_UCASE 0100
-#define C_SWAB 0200
-#define C_NOERROR 0400
-#define C_NOTRUNC 01000
-#define C_SYNC 02000
-/* Use separate input and output buffers, and combine partial input blocks. */
-#define C_TWOBUFS 04000
+enum
+  {
+    C_ASCII = 01,
+
+    C_EBCDIC = 02,
+    C_IBM = 04,
+    C_BLOCK = 010,
+    C_UNBLOCK = 020,
+    C_LCASE = 040,
+    C_UCASE = 0100,
+    C_SWAB = 0200,
+    C_NOERROR = 0400,
+    C_NOTRUNC = 01000,
+    C_SYNC = 02000,
+
+    /* Use separate input and output buffers, and combine partial
+       input blocks. */
+    C_TWOBUFS = 04000,
+
+    C_NOCREAT = 010000,
+    C_EXCL = 020000,
+    C_FDATASYNC = 040000,
+    C_FSYNC = 0100000
+  };
+
+/* Status bit masks.  */
+enum
+  {
+    STATUS_NOXFER = 01
+  };
 
 /* The name this program was run with. */
 char *program_name;
@@ -111,8 +135,15 @@ static uintmax_t max_records = (uintmax_t) -1;
 /* Bit vector of conversions to apply. */
 static int conversions_mask = 0;
 
+/* Open flags for the input and output files.  */
+static int input_flags = 0;
+static int output_flags = 0;
+
+/* Status flags for what is printed to stderr.  */
+static int status_flags = 0;
+
 /* If nonzero, filter characters through the translation table.  */
-static int translation_needed = 0;
+static bool translation_needed = false;
 
 /* Number of partial blocks written. */
 static uintmax_t w_partial = 0;
@@ -126,6 +157,25 @@ static uintmax_t r_partial = 0;
 /* Number of full blocks read. */
 static uintmax_t r_full = 0;
 
+/* Number of bytes written.  */
+static uintmax_t w_bytes = 0;
+
+/* Time that dd started.  */
+static struct timespec start_time;
+
+/* True if input is seekable.  */
+static bool input_seekable;
+
+/* Error number corresponding to initial attempt to lseek input.
+   If ESPIPE, do not issue any more diagnostics about it.  */
+static int input_seek_errno;
+
+/* File offset of the input, in bytes, along with a flag recording
+   whether it overflowed.  The offset is valid only if the input is
+   seekable and if the offset has not overflowed.  */
+static uintmax_t input_offset;
+static bool input_offset_overflow;
+
 /* Records truncated by conv=block. */
 static uintmax_t r_truncate = 0;
 
@@ -143,13 +193,18 @@ static size_t oc = 0;
 /* Index into current line, for `conv=block' and `conv=unblock'.  */
 static size_t col = 0;
 
-struct conversion
+/* A longest symbol in the struct symbol_values tables below.  */
+#define LONGEST_SYMBOL "fdatasync"
+
+/* A symbol and the corresponding integer value.  */
+struct symbol_value
 {
-  char *convname;
-  int conversion;
+  char symbol[sizeof LONGEST_SYMBOL];
+  int value;
 };
 
-static struct conversion conversions[] =
+/* Conversion symbols, for conv="...".  */
+static struct symbol_value const conversions[] =
 {
   {"ascii", C_ASCII | C_TWOBUFS},	/* EBCDIC to ASCII. */
   {"ebcdic", C_EBCDIC | C_TWOBUFS},	/* ASCII to EBCDIC. */
@@ -160,9 +215,33 @@ static struct conversion conversions[] =
   {"ucase", C_UCASE | C_TWOBUFS},	/* Translate lower to upper case. */
   {"swab", C_SWAB | C_TWOBUFS},	/* Swap bytes of input. */
   {"noerror", C_NOERROR},	/* Ignore i/o errors. */
+  {"nocreat", C_NOCREAT},	/* Do not create output file.  */
+  {"excl", C_EXCL},		/* Fail if the output file already exists.  */
   {"notrunc", C_NOTRUNC},	/* Do not truncate output file. */
   {"sync", C_SYNC},		/* Pad input records to ibs with NULs. */
-  {NULL, 0}
+  {"fdatasync", C_FDATASYNC},	/* Synchronize output data before finishing.  */
+  {"fsync", C_FSYNC},		/* Also synchronize output metadata.  */
+  {"", 0}
+};
+
+/* Flags, for iflag="..." and oflag="...".  */
+static struct symbol_value const flags[] =
+{
+  {"append",	O_APPEND},
+  {"direct",	O_DIRECT},
+  {"dsync",	O_DSYNC},
+  {"noctty",	O_NOCTTY},
+  {"nofollow",	O_NOFOLLOW},
+  {"nonblock",	O_NONBLOCK},
+  {"sync",	O_SYNC},
+  {"",		0}
+};
+
+/* Status, for status="...".  */
+static struct symbol_value const statuses[] =
+{
+  {"noxfer",	STATUS_NOXFER},
+  {"",		0}
 };
 
 /* Translation table formed by applying successive transformations. */
@@ -284,60 +363,94 @@ usage (int status)
 	     program_name);
   else
     {
-      printf (_("Usage: %s [OPTION]...\n"), program_name);
+      printf (_("\
+Usage: %s [OPERAND]...\n\
+  or:  %s OPTION\n\
+"),
+	      program_name, program_name);
       fputs (_("\
-Copy a file, converting and formatting according to the options.\n\
+Copy a file, converting and formatting according to the operands.\n\
 \n\
   bs=BYTES        force ibs=BYTES and obs=BYTES\n\
   cbs=BYTES       convert BYTES bytes at a time\n\
-  conv=KEYWORDS   convert the file as per the comma separated keyword list\n\
+  conv=CONVS      convert the file as per the comma separated symbol list\n\
   count=BLOCKS    copy only BLOCKS input blocks\n\
   ibs=BYTES       read BYTES bytes at a time\n\
 "), stdout);
       fputs (_("\
   if=FILE         read from FILE instead of stdin\n\
+  iflag=FLAGS     read as per the comma separated symbol list\n\
   obs=BYTES       write BYTES bytes at a time\n\
   of=FILE         write to FILE instead of stdout\n\
+  oflag=FLAGS     write as per the comma separated symbol list\n\
   seek=BLOCKS     skip BLOCKS obs-sized blocks at start of output\n\
   skip=BLOCKS     skip BLOCKS ibs-sized blocks at start of input\n\
+  status=noxfer   suppress transfer statistics\n\
 "), stdout);
-      fputs (HELP_OPTION_DESCRIPTION, stdout);
-      fputs (VERSION_OPTION_DESCRIPTION, stdout);
       fputs (_("\
 \n\
 BLOCKS and BYTES may be followed by the following multiplicative suffixes:\n\
 xM M, c 1, w 2, b 512, kB 1000, K 1024, MB 1000*1000, M 1024*1024,\n\
 GB 1000*1000*1000, G 1024*1024*1024, and so on for T, P, E, Z, Y.\n\
-Each KEYWORD may be:\n\
+\n\
+Each CONV symbol may be:\n\
 \n\
 "), stdout);
       fputs (_("\
   ascii     from EBCDIC to ASCII\n\
   ebcdic    from ASCII to EBCDIC\n\
-  ibm       from ASCII to alternated EBCDIC\n\
+  ibm       from ASCII to alternate EBCDIC\n\
   block     pad newline-terminated records with spaces to cbs-size\n\
   unblock   replace trailing spaces in cbs-size records with newline\n\
   lcase     change upper case to lower case\n\
 "), stdout);
       fputs (_("\
+  nocreat   do not create the output file\n\
+  excl      fail if the output file already exists\n\
   notrunc   do not truncate the output file\n\
   ucase     change lower case to upper case\n\
   swab      swap every pair of input bytes\n\
   noerror   continue after read errors\n\
   sync      pad every input block with NULs to ibs-size; when used\n\
               with block or unblock, pad with spaces rather than NULs\n\
+  fdatasync physically write output file data before finishing\n\
+  fsync     likewise, but also write metadata\n\
 "), stdout);
       fputs (_("\
 \n\
-Note that sending a SIGUSR1 signal to a running `dd' process makes it\n\
-print to standard error the number of records read and written so far,\n\
-then to resume copying.\n\
+Each FLAG symbol may be:\n\
+\n\
+  append    append mode (makes sense only for output)\n\
+"), stdout);
+      if (O_DIRECT)
+	fputs (_("  direct    use direct I/O for data\n"), stdout);
+      if (O_DSYNC)
+	fputs (_("  dsync     use synchronized I/O for data\n"), stdout);
+      if (O_SYNC)
+	fputs (_("  sync      likewise, but also for metadata\n"), stdout);
+      if (O_NONBLOCK)
+	fputs (_("  nonblock  use non-blocking I/O\n"), stdout);
+      if (O_NOFOLLOW)
+	fputs (_("  nofollow  do not follow symlinks\n"), stdout);
+      if (O_NOCTTY)
+	fputs (_("  noctty    do not assign controlling terminal from file\n"),
+	       stdout);
+      fputs (_("\
+\n\
+Sending a SIGUSR1 signal to a running `dd' process makes it\n\
+print I/O statistics to standard error, then to resume copying.\n\
 \n\
   $ dd if=/dev/zero of=/dev/null& pid=$!\n\
   $ kill -USR1 $pid; sleep 1; kill $pid\n\
-  10899206+0 records in\n\
-  10899206+0 records out\n\
+  18335302+0 records in\n\
+  18335302+0 records out\n\
+  9387674624 bytes (9.4 GB) copied, 34.6279 seconds, 271 MB/s\n\
+\n\
+Options are:\n\
+\n\
 "), stdout);
+      fputs (HELP_OPTION_DESCRIPTION, stdout);
+      fputs (VERSION_OPTION_DESCRIPTION, stdout);
       printf (_("\nReport bugs to <%s>.\n"), PACKAGE_BUGREPORT);
     }
   exit (status);
@@ -350,25 +463,32 @@ translate_charset (char const *new_trans)
 
   for (i = 0; i < 256; i++)
     trans_table[i] = new_trans[trans_table[i]];
-  translation_needed = 1;
+  translation_needed = true;
 }
 
-/* Return the number of 1 bits in `i'. */
+/* Return true if I has more than one bit set.  I must be nonnegative.  */
 
-static int
-bit_count (register int i)
+static inline bool
+multiple_bits_set (int i)
 {
-  register int set_bits;
-
-  for (set_bits = 0; i != 0; set_bits++)
-    i &= i - 1;
-  return set_bits;
+  return (i & (i - 1)) != 0;
 }
 
 static void
 print_stats (void)
 {
-  char buf[2][INT_BUFSIZE_BOUND (uintmax_t)];
+  char buf[2][MAX (INT_BUFSIZE_BOUND (uintmax_t), LONGEST_HUMAN_READABLE + 1)];
+  struct timespec now;
+  int human_opts =
+    (human_autoscale | human_round_to_nearest
+     | human_space_before_unit | human_SI | human_B);
+  uintmax_t start_sec = start_time.tv_sec;
+  enum { BILLION = 1000000000 };
+  double BILLIONe0 = BILLION;
+  double delta_s;
+  char const *bytes_per_second;
+
+  gettime (&now);
   fprintf (stderr, _("%s+%s records in\n"),
 	   umaxtostr (r_full, buf[0]), umaxtostr (r_partial, buf[1]));
   fprintf (stderr, _("%s+%s records out\n"),
@@ -381,18 +501,59 @@ print_stats (void)
 		? _("truncated record")
 		: _("truncated records")));
     }
+
+  if (status_flags & STATUS_NOXFER)
+    return;
+
+  /* Use integer arithmetic to compute the transfer rate if possible,
+     since that makes it easy to use SI abbreviations; otherwise, fall
+     back on floating-point without abbreviations.  */
+
+  fprintf (stderr,
+	   ngettext ("%s byte (%s) copied",
+		     "%s bytes (%s) copied", w_bytes == 1),
+	   umaxtostr (w_bytes, buf[0]),
+	   human_readable (w_bytes, buf[1], human_opts, 1, 1));
+
+  if ((start_time.tv_sec < now.tv_sec
+       || (start_time.tv_sec == now.tv_sec
+	   && start_time.tv_nsec < now.tv_nsec))
+      && now.tv_sec - start_sec < UINTMAX_MAX / BILLION)
+    {
+      uintmax_t delta_ns = (BILLION * (now.tv_sec - start_sec)
+			    + now.tv_nsec - start_time.tv_nsec);
+      delta_s = delta_ns / BILLIONe0;
+      bytes_per_second = human_readable (w_bytes, buf[1], human_opts,
+					 BILLION, delta_ns);
+    }
+  else
+    {
+      delta_s = now.tv_sec;
+      delta_s -= start_time.tv_sec;
+      delta_s += (now.tv_nsec - start_time.tv_nsec) / BILLIONe0;
+      if (0 < delta_s)
+	sprintf (buf[1], "%gB", w_bytes / delta_s);
+      else
+	sprintf (buf[1], "%s B", _("Infinity"));
+      bytes_per_second = buf[1];
+    }
+
+  fprintf (stderr,
+	   ngettext (", %g second, %s/s\n",
+		     ", %g seconds, %s/s\n", delta_s == 1),
+	   delta_s, bytes_per_second);
 }
 
 static void
 cleanup (void)
 {
-  print_stats ();
   if (close (STDIN_FILENO) < 0)
     error (EXIT_FAILURE, errno,
 	   _("closing input file %s"), quote (input_file));
   if (close (STDOUT_FILENO) < 0)
     error (EXIT_FAILURE, errno,
 	   _("closing output file %s"), quote (output_file));
+  print_stats ();
 }
 
 static inline void
@@ -474,6 +635,7 @@ static void
 write_output (void)
 {
   size_t nwritten = full_write (STDOUT_FILENO, obuf, output_blocksize);
+  w_bytes += nwritten;
   if (nwritten != output_blocksize)
     {
       error (0, errno, _("writing to %s"), quote (output_file));
@@ -486,42 +648,51 @@ write_output (void)
   oc = 0;
 }
 
-/* Interpret one "conv=..." option.
+/* Interpret one "conv=..." or similar operand STR according to the
+   symbols in TABLE, returning the flags specified.  If the operand
+   cannot be parsed, use ERROR_MSGID to generate a diagnostic.
    As a by product, this function replaces each `,' in STR with a NUL byte.  */
 
-static void
-parse_conversion (char *str)
+static int
+parse_symbols (char *str, struct symbol_value const *table,
+	       char const *error_msgid)
 {
-  char *new;
-  int i;
+  int value = 0;
 
   do
     {
-      new = strchr (str, ',');
+      struct symbol_value const *entry;
+      char *new = strchr (str, ',');
       if (new != NULL)
 	*new++ = '\0';
-      for (i = 0; conversions[i].convname != NULL; i++)
-	if (STREQ (conversions[i].convname, str))
-	  {
-	    conversions_mask |= conversions[i].conversion;
-	    break;
-	  }
-      if (conversions[i].convname == NULL)
+      for (entry = table; ; entry++)
 	{
-	  error (0, 0, _("invalid conversion: %s"), quote (str));
-	  usage (EXIT_FAILURE);
+	  if (! entry->symbol[0])
+	    {
+	      error (0, 0, _(error_msgid), quote (str));
+	      usage (EXIT_FAILURE);
+	    }
+	  if (STREQ (entry->symbol, str))
+	    {
+	      if (! entry->value)
+		error (EXIT_FAILURE, 0, _(error_msgid), quote (str));
+	      value |= entry->value;
+	      break;
+	    }
 	}
       str = new;
-  } while (new != NULL);
+    }
+  while (str);
+
+  return value;
 }
 
 /* Return the value of STR, interpreted as a non-negative decimal integer,
    optionally multiplied by various values.
-   Assign nonzero to *INVALID if STR does not represent a number in
-   this format. */
+   Set *INVALID if STR does not represent a number in this format.  */
 
 static uintmax_t
-parse_integer (const char *str, int *invalid)
+parse_integer (const char *str, bool *invalid)
 {
   uintmax_t n;
   char *suffix;
@@ -533,7 +704,7 @@ parse_integer (const char *str, int *invalid)
 
       if (multiplier != 0 && n * multiplier / multiplier != n)
 	{
-	  *invalid = 1;
+	  *invalid = true;
 	  return 0;
 	}
 
@@ -541,7 +712,7 @@ parse_integer (const char *str, int *invalid)
     }
   else if (e != LONGINT_OK)
     {
-      *invalid = 1;
+      *invalid = true;
       return 0;
     }
 
@@ -553,9 +724,6 @@ scanargs (int argc, char **argv)
 {
   int i;
 
-  --argc;
-  ++argv;
-
   for (i = optind; i < argc; i++)
     {
       char *name, *val;
@@ -564,7 +732,7 @@ scanargs (int argc, char **argv)
       val = strchr (name, '=');
       if (val == NULL)
 	{
-	  error (0, 0, _("unrecognized option %s"), quote (name));
+	  error (0, 0, _("unrecognized operand %s"), quote (name));
 	  usage (EXIT_FAILURE);
 	}
       *val++ = '\0';
@@ -574,34 +742,38 @@ scanargs (int argc, char **argv)
       else if (STREQ (name, "of"))
 	output_file = val;
       else if (STREQ (name, "conv"))
-	parse_conversion (val);
+	conversions_mask |= parse_symbols (val, conversions,
+					   N_("invalid conversion: %s"));
+      else if (STREQ (name, "iflag"))
+	input_flags |= parse_symbols (val, flags,
+				      N_("invalid input flag: %s"));
+      else if (STREQ (name, "oflag"))
+	output_flags |= parse_symbols (val, flags,
+				       N_("invalid output flag: %s"));
+      else if (STREQ (name, "status"))
+	status_flags |= parse_symbols (val, statuses,
+				       N_("invalid status flag: %s"));
       else
 	{
-	  int invalid = 0;
+	  bool invalid = false;
 	  uintmax_t n = parse_integer (val, &invalid);
 
 	  if (STREQ (name, "ibs"))
 	    {
-	      /* Ensure that each blocksize is <= SSIZE_MAX.  */
-	      invalid |= SSIZE_MAX < n;
+	      invalid |= ! (0 < n && n <= MAX_BLOCKSIZE);
 	      input_blocksize = n;
-	      invalid |= input_blocksize != n || input_blocksize == 0;
 	      conversions_mask |= C_TWOBUFS;
 	    }
 	  else if (STREQ (name, "obs"))
 	    {
-	      /* Ensure that each blocksize is <= SSIZE_MAX.  */
-	      invalid |= SSIZE_MAX < n;
+	      invalid |= ! (0 < n && n <= MAX_BLOCKSIZE);
 	      output_blocksize = n;
-	      invalid |= output_blocksize != n || output_blocksize == 0;
 	      conversions_mask |= C_TWOBUFS;
 	    }
 	  else if (STREQ (name, "bs"))
 	    {
-	      /* Ensure that each blocksize is <= SSIZE_MAX.  */
-	      invalid |= SSIZE_MAX < n;
+	      invalid |= ! (0 < n && n <= MAX_BLOCKSIZE);
 	      output_blocksize = input_blocksize = n;
-	      invalid |= output_blocksize != n || output_blocksize == 0;
 	    }
 	  else if (STREQ (name, "cbs"))
 	    {
@@ -617,7 +789,7 @@ scanargs (int argc, char **argv)
 	    max_records = n;
 	  else
 	    {
-	      error (0, 0, _("unrecognized option %s=%s"),
+	      error (0, 0, _("unrecognized operand %s=%s"),
 		     quote_n (0, name), quote_n (1, val));
 	      usage (EXIT_FAILURE);
 	    }
@@ -638,6 +810,18 @@ scanargs (int argc, char **argv)
     output_blocksize = DEFAULT_BLOCKSIZE;
   if (conversion_blocksize == 0)
     conversions_mask &= ~(C_BLOCK | C_UNBLOCK);
+
+  if (input_flags & (O_DSYNC | O_SYNC))
+    input_flags |= O_RSYNC;
+
+  if (multiple_bits_set (conversions_mask & (C_ASCII | C_EBCDIC | C_IBM)))
+    error (EXIT_FAILURE, 0, _("cannot combine any two of {ascii,ebcdic,ibm}"));
+  if (multiple_bits_set (conversions_mask & (C_BLOCK | C_UNBLOCK)))
+    error (EXIT_FAILURE, 0, _("cannot combine block and unblock"));
+  if (multiple_bits_set (conversions_mask & (C_LCASE | C_UCASE)))
+    error (EXIT_FAILURE, 0, _("cannot combine lcase and ucase"));
+  if (multiple_bits_set (conversions_mask & (C_EXCL | C_NOCREAT)))
+    error (EXIT_FAILURE, 0, _("cannot combine excl and nocreat"));
 }
 
 /* Fix up translation table. */
@@ -647,16 +831,6 @@ apply_translations (void)
 {
   int i;
 
-#define MX(a) (bit_count (conversions_mask & (a)))
-  if ((MX (C_ASCII | C_EBCDIC | C_IBM) > 1)
-      || (MX (C_BLOCK | C_UNBLOCK) > 1)
-      || (MX (C_LCASE | C_UCASE) > 1))
-    {
-      error (EXIT_FAILURE, 0, _("\
-	only one conv in {ascii,ebcdic,ibm}, {lcase,ucase}, {block,unblock}"));
-    }
-#undef MX
-
   if (conversions_mask & C_ASCII)
     translate_charset (ebcdic_to_ascii);
 
@@ -665,14 +839,14 @@ apply_translations (void)
       for (i = 0; i < 256; i++)
 	if (ISLOWER (trans_table[i]))
 	  trans_table[i] = TOUPPER (trans_table[i]);
-      translation_needed = 1;
+      translation_needed = true;
     }
   else if (conversions_mask & C_LCASE)
     {
       for (i = 0; i < 256; i++)
 	if (ISUPPER (trans_table[i]))
 	  trans_table[i] = TOLOWER (trans_table[i]);
-      translation_needed = 1;
+      translation_needed = true;
     }
 
   if (conversions_mask & C_EBCDIC)
@@ -699,12 +873,12 @@ translate_buffer (char *buf, size_t nread)
   size_t i;
 
   for (i = nread, cp = buf; i; i--, cp++)
-    *cp = trans_table[(unsigned char) *cp];
+    *cp = trans_table[to_uchar (*cp)];
 }
 
-/* If nonnzero, the last char from the previous call to `swab_buffer'
+/* If true, the last char from the previous call to `swab_buffer'
    is saved in `saved_char'.  */
-static int char_is_saved = 0;
+static bool char_is_saved = false;
 
 /* Odd char from previous call.  */
 static char saved_char;
@@ -718,21 +892,21 @@ swab_buffer (char *buf, size_t *nread)
 {
   char *bufstart = buf;
   register char *cp;
-  register int i;
+  size_t i;
 
   /* Is a char left from last time?  */
   if (char_is_saved)
     {
       *--bufstart = saved_char;
       (*nread)++;
-      char_is_saved = 0;
+      char_is_saved = false;
     }
 
   if (*nread & 1)
     {
       /* An odd number of chars are in the buffer.  */
       saved_char = bufstart[--*nread];
-      char_is_saved = 1;
+      char_is_saved = true;
     }
 
   /* Do the byte-swapping by moving every second character two
@@ -746,6 +920,17 @@ swab_buffer (char *buf, size_t *nread)
   return ++bufstart;
 }
 
+/* Add OFFSET to the input offset, setting the overflow flag if
+   necessary.  */
+
+static void
+advance_input_offset (uintmax_t offset)
+{
+  input_offset += offset;
+  if (input_offset < offset)
+    input_offset_overflow = true;
+}
+
 /* This is a wrapper for lseek.  It detects and warns about a kernel
    bug that makes lseek a no-op for tape devices, even though the kernel
    lseek return value suggests that the function succeeded.
@@ -774,14 +959,11 @@ skip_via_lseek (char const *filename, int fdesc, off_t offset, int whence)
 {
   struct mtget s1;
   struct mtget s2;
-  off_t new_position;
-  int got_original_tape_position;
-
-  got_original_tape_position = (ioctl (fdesc, MTIOCGET, &s1) == 0);
+  bool got_original_tape_position = (ioctl (fdesc, MTIOCGET, &s1) == 0);
   /* known bad device type */
   /* && s.mt_type == MT_ISSCSI2 */
 
-  new_position = lseek (fdesc, offset, whence);
+  off_t new_position = lseek (fdesc, offset, whence);
   if (0 <= new_position
       && got_original_tape_position
       && ioctl (fdesc, MTIOCGET, &s2) == 0
@@ -790,6 +972,7 @@ skip_via_lseek (char const *filename, int fdesc, off_t offset, int whence)
       error (0, 0, _("warning: working around lseek kernel bug for file (%s)\n\
   of mt_type=0x%0lx -- see <sys/mtio.h> for the list of types"),
 	     filename, s2.mt_type);
+      errno = 0;
       new_position = -1;
     }
 
@@ -802,7 +985,7 @@ skip_via_lseek (char const *filename, int fdesc, off_t offset, int whence)
 /* Throw away RECORDS blocks of BLOCKSIZE bytes on file descriptor FDESC,
    which is open with read permission for FILE.  Store up to BLOCKSIZE
    bytes of the data at a time in BUF, if necessary.  RECORDS must be
-   nonzero.  */
+   nonzero.  If fdesc is STDIN_FILENO, advance the input offset.  */
 
 static void
 skip (int fdesc, char const *file, uintmax_t records, size_t blocksize,
@@ -814,15 +997,32 @@ skip (int fdesc, char const *file, uintmax_t records, size_t blocksize,
      or if the the file offset is not representable as an off_t --
      fall back on using read.  */
 
-  if ((uintmax_t) offset / blocksize != records
-      || skip_via_lseek (file, fdesc, offset, SEEK_CUR) < 0)
+  errno = 0;
+  if ((uintmax_t) offset / blocksize == records
+      && 0 <= skip_via_lseek (file, fdesc, offset, SEEK_CUR))
     {
+      if (fdesc == STDIN_FILENO)
+	advance_input_offset (offset);
+    }
+  else
+    {
+      int lseek_errno = errno;
       while (records--)
 	{
 	  size_t nread = safe_read (fdesc, buf, blocksize);
 	  if (nread == SAFE_READ_ERROR)
 	    {
-	      error (0, errno, _("reading %s"), quote (file));
+	      if (fdesc == STDIN_FILENO)
+		{
+		  error (0, errno, _("reading %s"), quote (file));
+		  if (conversions_mask & C_NOERROR)
+		    {
+		      print_stats ();
+		      continue;
+		    }
+		}
+	      else
+		error (0, lseek_errno, _("%s: cannot seek"), quote (file));
 	      quit (EXIT_FAILURE);
 	    }
 	  /* POSIX doesn't say what to do when dd detects it has been
@@ -830,23 +1030,68 @@ skip (int fdesc, char const *file, uintmax_t records, size_t blocksize,
 	     FIXME: maybe give a warning.  */
 	  if (nread == 0)
 	    break;
+	  if (fdesc == STDIN_FILENO)
+	    advance_input_offset (nread);
 	}
     }
 }
 
+/* Advance the input by NBYTES if possible, after a read error.
+   The input file offset may or may not have advanced after the failed
+   read; adjust it to point just after the bad record regardless.
+   Return true if successful, or if the input is already known to not
+   be seekable.  */
+
+static bool
+advance_input_after_read_error (size_t nbytes)
+{
+  if (! input_seekable)
+    {
+      if (input_seek_errno == ESPIPE)
+	return true;
+      errno = input_seek_errno;
+    }
+  else
+    {
+      off_t offset;
+      advance_input_offset (nbytes);
+      input_offset_overflow |= (OFF_T_MAX < input_offset);
+      if (input_offset_overflow)
+	{
+	  error (0, 0, _("offset overflow while reading file %s"),
+		 quote (input_file));
+	  return false;
+	}
+      offset = lseek (STDIN_FILENO, 0, SEEK_CUR);
+      if (0 <= offset)
+	{
+	  off_t diff;
+	  if (offset == input_offset)
+	    return true;
+	  diff = input_offset - offset;
+	  if (! (0 <= diff && diff <= nbytes))
+	    error (0, 0, _("warning: screwy file offset after failed read"));
+	  if (0 <= skip_via_lseek (input_file, STDIN_FILENO, diff, SEEK_CUR))
+	    return true;
+	  if (errno == 0)
+	    error (0, 0, _("cannot work around kernel bug after all"));
+	}
+    }
+
+  error (0, errno, _("%s: cannot seek"), quote (input_file));
+  return false;
+}
+
 /* Copy NREAD bytes of BUF, with no conversions.  */
 
 static void
-copy_simple (char const *buf, int nread)
+copy_simple (char const *buf, size_t nread)
 {
-  int nfree;			/* Number of unused bytes in `obuf'.  */
   const char *start = buf;	/* First uncopied char in BUF.  */
 
   do
     {
-      nfree = output_blocksize - oc;
-      if (nfree > nread)
-	nfree = nread;
+      size_t nfree = MIN (nread, output_blocksize - oc);
 
       memcpy (obuf + oc, start, nfree);
 
@@ -856,7 +1101,7 @@ copy_simple (char const *buf, int nread)
       if (oc >= output_blocksize)
 	write_output ();
     }
-  while (nread > 0);
+  while (nread != 0);
 }
 
 /* Copy NREAD bytes of BUF, doing conv=block
@@ -900,7 +1145,7 @@ copy_with_unblock (char const *buf, size_t nread)
 {
   size_t i;
   char c;
-  static int pending_spaces = 0;
+  static size_t pending_spaces = 0;
 
   for (i = 0; i < nread; i++)
     {
@@ -928,6 +1173,22 @@ copy_with_unblock (char const *buf, size_t nread)
     }
 }
 
+/* Set the file descriptor flags for FD that correspond to the nonzero bits
+   in ADD_FLAGS.  The file's name is NAME.  */
+
+static void
+set_fd_flags (int fd, int add_flags, char const *name)
+{
+  if (add_flags)
+    {
+      int old_flags = fcntl (fd, F_GETFL);
+      int new_flags = old_flags | add_flags;
+      if (old_flags < 0
+	  || (new_flags != old_flags && fcntl (fd, F_SETFL, new_flags) == -1))
+	error (EXIT_FAILURE, errno, _("setting flags for %s"), quote (name));
+    }
+}
+
 /* The main loop.  */
 
 static int
@@ -937,6 +1198,11 @@ dd_copy (void)
   char *real_buf;		/* real buffer address before alignment */
   char *real_obuf;
   size_t nread;			/* Bytes read in the current block. */
+
+  /* If nonzero, then the previously read block was partial and
+     PARTREAD was its size.  */
+  size_t partread = 0;
+
   int exit_status = EXIT_SUCCESS;
   size_t page_size = getpagesize ();
   size_t n_bytes_read;
@@ -964,13 +1230,13 @@ dd_copy (void)
   ibuf = real_buf;
   ibuf += SWAB_ALIGN_OFFSET;	/* allow space for swab */
 
-  ibuf = PTR_ALIGN (ibuf, page_size);
+  ibuf = ptr_align (ibuf, page_size);
 
   if (conversions_mask & C_TWOBUFS)
     {
       /* Page-align the output buffer, too.  */
       real_obuf = xmalloc (output_blocksize + page_size - 1);
-      obuf = PTR_ALIGN (real_obuf, page_size);
+      obuf = ptr_align (real_obuf, page_size);
     }
   else
     {
@@ -982,19 +1248,10 @@ dd_copy (void)
     skip (STDIN_FILENO, input_file, skip_records, input_blocksize, ibuf);
 
   if (seek_records != 0)
-    {
-      /* FIXME: this loses for
-	 % ./dd if=dd seek=1 |:
-	 ./dd: standard output: Bad file descriptor
-	 0+0 records in
-	 0+0 records out
-	 */
-
-      skip (STDOUT_FILENO, output_file, seek_records, output_blocksize, obuf);
-    }
+    skip (STDOUT_FILENO, output_file, seek_records, output_blocksize, obuf);
 
   if (max_records == 0)
-    quit (exit_status);
+    return exit_status;
 
   while (1)
     {
@@ -1021,8 +1278,15 @@ dd_copy (void)
 	    {
 	      print_stats ();
 	      /* Seek past the bad block if possible. */
-	      lseek (STDIN_FILENO, (off_t) input_blocksize, SEEK_CUR);
-	      if (conversions_mask & C_SYNC)
+	      if (!advance_input_after_read_error (input_blocksize - partread))
+		{
+		  exit_status = EXIT_FAILURE;
+
+		  /* Suppress duplicate diagnostics.  */
+		  input_seekable = false;
+		  input_seek_errno = ESPIPE;
+		}
+	      if ((conversions_mask & C_SYNC) && !partread)
 		/* Replace the missing input with null bytes and
 		   proceed normally.  */
 		nread = 0;
@@ -1038,10 +1302,12 @@ dd_copy (void)
 	}
 
       n_bytes_read = nread;
+      advance_input_offset (nread);
 
       if (n_bytes_read < input_blocksize)
 	{
 	  r_partial++;
+	  partread = n_bytes_read;
 	  if (conversions_mask & C_SYNC)
 	    {
 	      if (!(conversions_mask & C_NOERROR))
@@ -1053,15 +1319,19 @@ dd_copy (void)
 	    }
 	}
       else
-	r_full++;
+	{
+	  r_full++;
+	  partread = 0;
+	}
 
       if (ibuf == obuf)		/* If not C_TWOBUFS. */
 	{
 	  size_t nwritten = full_write (STDOUT_FILENO, obuf, n_bytes_read);
+	  w_bytes += nwritten;
 	  if (nwritten != n_bytes_read)
 	    {
 	      error (0, errno, _("writing %s"), quote (output_file));
-	      quit (EXIT_FAILURE);
+	      return EXIT_FAILURE;
 	    }
 	  else if (n_bytes_read == input_blocksize)
 	    w_full++;
@@ -1117,12 +1387,13 @@ dd_copy (void)
   if (oc != 0)
     {
       size_t nwritten = full_write (STDOUT_FILENO, obuf, oc);
+      w_bytes += nwritten;
       if (nwritten != 0)
 	w_partial++;
       if (nwritten != oc)
 	{
 	  error (0, errno, _("writing %s"), quote (output_file));
-	  quit (EXIT_FAILURE);
+	  return EXIT_FAILURE;
 	}
     }
 
@@ -1130,6 +1401,24 @@ dd_copy (void)
   if (real_obuf)
     free (real_obuf);
 
+  if ((conversions_mask & C_FDATASYNC) && fdatasync (STDOUT_FILENO) != 0)
+    {
+      if (errno != ENOSYS && errno != EINVAL)
+	{
+	  error (0, errno, _("fdatasync failed for %s"), quote (output_file));
+	  exit_status = EXIT_FAILURE;
+	}
+      conversions_mask |= C_FSYNC;
+    }
+
+  if (conversions_mask & C_FSYNC)
+    while (fsync (STDOUT_FILENO) != 0)
+      if (errno != EINTR)
+	{
+	  error (0, errno, _("fsync failed for %s"), quote (output_file));
+	  return EXIT_FAILURE;
+	}
+
   return exit_status;
 }
 
@@ -1149,6 +1438,7 @@ main (int argc, char **argv)
 {
   int i;
   int exit_status;
+  off_t offset;
 
   initialize_main (&argc, &argv);
   program_name = argv[0];
@@ -1161,6 +1451,8 @@ main (int argc, char **argv)
 
   parse_long_options (argc, argv, PROGRAM_NAME, PACKAGE, VERSION,
 		      usage, AUTHORS, (char const *) NULL);
+  if (getopt_long (argc, argv, "", NULL, NULL) != -1)
+    usage (EXIT_FAILURE);
 
   /* Don't close stdout on exit from here on.  */
   closeout_func = NULL;
@@ -1174,19 +1466,34 @@ main (int argc, char **argv)
 
   apply_translations ();
 
-  if (input_file != NULL)
+  if (input_file == NULL)
     {
-      if (open_fd (STDIN_FILENO, input_file, O_RDONLY, 0) < 0)
-	error (EXIT_FAILURE, errno, _("opening %s"), quote (input_file));
+      input_file = _("standard input");
+      set_fd_flags (STDIN_FILENO, input_flags, input_file);
     }
   else
-    input_file = _("standard input");
+    {
+      if (open_fd (STDIN_FILENO, input_file, O_RDONLY | input_flags, 0) < 0)
+	error (EXIT_FAILURE, errno, _("opening %s"), quote (input_file));
+    }
 
-  if (output_file != NULL)
+  offset = lseek (STDIN_FILENO, 0, SEEK_CUR);
+  input_seekable = (0 <= offset);
+  input_offset = offset;
+  input_seek_errno = errno;
+
+  if (output_file == NULL)
+    {
+      output_file = _("standard output");
+      set_fd_flags (STDOUT_FILENO, output_flags, output_file);
+    }
+  else
     {
       mode_t perms = S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH;
       int opts
-	= (O_CREAT
+	= (output_flags
+	   | (conversions_mask & C_NOCREAT ? 0 : O_CREAT)
+	   | (conversions_mask & C_EXCL ? O_EXCL : 0)
 	   | (seek_records || (conversions_mask & C_NOTRUNC) ? 0 : O_TRUNC));
 
       /* Open the output file with *read* access only if we might
@@ -1210,8 +1517,8 @@ main (int argc, char **argv)
 		   quote (output_file));
 
 	  /* Complain only when ftruncate fails on a regular file, a
-	     directory, or a shared memory object, as the 2000-08
-	     POSIX draft specifies ftruncate's behavior only for these
+	     directory, or a shared memory object, as
+	     POSIX 1003.1-2004 specifies ftruncate's behavior only for these
 	     file types.  For example, do not complain when Linux 2.4
 	     ftruncate fails on /dev/fd0.  */
 	  if (ftruncate (STDOUT_FILENO, o) != 0
@@ -1227,16 +1534,14 @@ main (int argc, char **argv)
 	}
 #endif
     }
-  else
-    {
-      output_file = _("standard output");
-    }
 
   install_handler (SIGINT, interrupt_handler);
   install_handler (SIGQUIT, interrupt_handler);
   install_handler (SIGPIPE, interrupt_handler);
   install_handler (SIGINFO, siginfo_handler);
 
+  gettime (&start_time);
+
   exit_status = dd_copy ();
 
   quit (exit_status);
