@@ -136,10 +136,10 @@ A mere - implies -i.  If no COMMAND, print the resulting environment.\n\
 }
 
 int
-main (register int argc, register char **argv, char **envp)
+main (register int argc, register char **argv)
 {
   int optc;
-  int ignore_environment = 0;
+  bool ignore_environment = false;
 
   initialize_main (&argc, &argv);
   program_name = argv[0];
@@ -154,10 +154,8 @@ main (register int argc, register char **argv, char **envp)
     {
       switch (optc)
 	{
-	case 0:
-	  break;
 	case 'i':
-	  ignore_environment = 1;
+	  ignore_environment = true;
 	  break;
 	case 'u':
 	  break;
@@ -168,8 +166,8 @@ main (register int argc, register char **argv, char **envp)
 	}
     }
 
-  if (optind < argc && !strcmp (argv[optind], "-"))
-    ignore_environment = 1;
+  if (optind < argc && STREQ (argv[optind], "-"))
+    ignore_environment = true;
 
   if (ignore_environment)
     {
@@ -182,7 +180,7 @@ main (register int argc, register char **argv, char **envp)
     if (optc == 'u')
       putenv (optarg);		/* Requires GNU putenv. */
 
-  if (optind < argc && !strcmp (argv[optind], "-"))
+  if (optind < argc && STREQ (argv[optind], "-"))
     ++optind;
 
   while (optind < argc && strchr (argv[optind], '='))
