@@ -1,5 +1,5 @@
 /* expr -- evaluate expressions.
-   Copyright (C) 86, 1991-1997, 1999-2003 Free Software Foundation, Inc.
+   Copyright (C) 86, 1991-1997, 1999-2004 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -36,7 +36,6 @@
 #include <regex.h>
 #include "long-options.h"
 #include "error.h"
-#include "exitfail.h"
 #include "inttostr.h"
 #include "quotearg.h"
 
@@ -45,10 +44,6 @@
 
 #define AUTHORS "Mike Parker"
 
-#undef NEW
-#define NEW(Type) XMALLOC (Type, 1)
-#define OLD(x) free (x)
-
 /* Exit statuses.  */
 enum
   {
@@ -58,7 +53,7 @@ enum
     EXPR_INVALID = 2,
 
     /* Some other error occurred.  */
-    EXPR_ERROR
+    EXPR_FAILURE
   };
 
 /* The kinds of value we can have.  */
@@ -87,7 +82,7 @@ static char **args;
 /* The name this program was run with. */
 char *program_name;
 
-static VALUE *eval (void);
+static VALUE *eval (bool);
 static bool nomoreargs (void);
 static bool null (VALUE *v);
 static void printv (VALUE *v);
@@ -95,7 +90,7 @@ static void printv (VALUE *v);
 void
 usage (int status)
 {
-  if (status != 0)
+  if (status != EXIT_SUCCESS)
     fprintf (stderr, _("Try `%s --help' for more information.\n"),
 	     program_name);
   else
@@ -158,6 +153,11 @@ Beware that many operators need to be escaped or quoted for shells.\n\
 Comparisons are arithmetic if both ARGs are numbers, else lexicographical.\n\
 Pattern matches return the string matched between \\( and \\) or null; if\n\
 \\( and \\) are not used, they return the number of characters matched or 0.\n\
+"), stdout);
+      fputs (_("\
+\n\
+Exit status is 0 if EXPRESSION is neither null nor 0, 1 if EXPRESSION is null\n\
+or 0, 2 if EXPRESSION is syntactically invalid, and 3 if an error occurred.\n\
 "), stdout);
       printf (_("\nReport bugs to <%s>.\n"), PACKAGE_BUGREPORT);
     }
@@ -182,13 +182,11 @@ main (int argc, char **argv)
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
-  /* Change the way library functions fail.  */
-  exit_failure = EXPR_ERROR;
-
+  initialize_exit_failure (EXPR_FAILURE);
   atexit (close_stdout);
 
   parse_long_options (argc, argv, PROGRAM_NAME, GNU_PACKAGE, VERSION,
-		      AUTHORS, usage);
+		      usage, AUTHORS, (char const *) NULL);
   /* The above handles --help and --version.
      Since there is no other invocation of getopt, handle `--' here.  */
   if (argc > 1 && STREQ (argv[1], "--"))
@@ -199,13 +197,13 @@ main (int argc, char **argv)
 
   if (argc <= 1)
     {
-      error (0, 0, _("too few arguments"));
+      error (0, 0, _("missing operand"));
       usage (EXPR_INVALID);
     }
 
   args = argv + 1;
 
-  v = eval ();
+  v = eval (true);
   if (!nomoreargs ())
     syntax_error ();
   printv (v);
@@ -218,9 +216,7 @@ main (int argc, char **argv)
 static VALUE *
 int_value (intmax_t i)
 {
-  VALUE *v;
-
-  v = NEW (VALUE);
+  VALUE *v = xmalloc (sizeof *v);
   v->type = integer;
   v->u.i = i;
   return v;
@@ -231,9 +227,7 @@ int_value (intmax_t i)
 static VALUE *
 str_value (char *s)
 {
-  VALUE *v;
-
-  v = NEW (VALUE);
+  VALUE *v = xmalloc (sizeof *v);
   v->type = string;
   v->u.s = xstrdup (s);
   return v;
@@ -246,7 +240,7 @@ freev (VALUE *v)
 {
   if (v->type == string)
     free (v->u.s);
-  OLD (v);
+  free (v);
 }
 
 /* Print VALUE V.  */
@@ -282,7 +276,22 @@ null (VALUE *v)
     case integer:
       return v->u.i == 0;
     case string:
-      return v->u.s[0] == '\0' || strcmp (v->u.s, "0") == 0;
+      {
+	char const *cp = v->u.s;
+	if (*cp == '\0')
+	  return true;
+
+	cp += (*cp == '-');
+
+	do
+	  {
+	    if (*cp != '0')
+	      return false;
+	  }
+	while (*++cp);
+
+	return true;
+      }
     default:
       abort ();
     }
@@ -346,17 +355,17 @@ toarith (VALUE *v)
     }
 }
 
-/* Return nonzero and advance if the next token matches STR exactly.
+/* Return true and advance if the next token matches STR exactly.
    STR must not be NULL.  */
 
 static bool
 nextarg (char const *str)
 {
   if (*args == NULL)
-    return 0;
+    return false;
   else
     {
-      bool r = strcmp (*args, str) == 0;
+      bool r = STREQ (*args, str);
       args += r;
       return r;
     }
@@ -422,7 +431,7 @@ of the basic regular expression is not portable; it is being ignored"),
   re_syntax_options = RE_SYNTAX_POSIX_BASIC;
   errmsg = re_compile_pattern (pv->u.s, len, &re_buffer);
   if (errmsg)
-    error (EXPR_ERROR, 0, "%s", errmsg);
+    error (EXPR_FAILURE, 0, "%s", errmsg);
 
   matchlen = re_match (&re_buffer, sv->u.s, strlen (sv->u.s), 0, &re_regs);
   if (0 <= matchlen)
@@ -451,7 +460,7 @@ of the basic regular expression is not portable; it is being ignored"),
 /* Handle bare operands and ( expr ) syntax.  */
 
 static VALUE *
-eval7 (void)
+eval7 (bool evaluate)
 {
   VALUE *v;
 
@@ -463,7 +472,7 @@ eval7 (void)
 
   if (nextarg ("("))
     {
-      v = eval ();
+      v = eval (evaluate);
       if (!nextarg (")"))
 	syntax_error ();
       return v;
@@ -478,7 +487,7 @@ eval7 (void)
 /* Handle match, substr, index, and length keywords, and quoting "+".  */
 
 static VALUE *
-eval6 (void)
+eval6 (bool evaluate)
 {
   VALUE *l;
   VALUE *r;
@@ -497,7 +506,7 @@ eval6 (void)
     }
   else if (nextarg ("length"))
     {
-      r = eval6 ();
+      r = eval6 (evaluate);
       tostring (r);
       v = int_value (strlen (r->u.s));
       freev (r);
@@ -505,17 +514,22 @@ eval6 (void)
     }
   else if (nextarg ("match"))
     {
-      l = eval6 ();
-      r = eval6 ();
-      v = docolon (l, r);
-      freev (l);
+      l = eval6 (evaluate);
+      r = eval6 (evaluate);
+      if (evaluate)
+	{
+	  v = docolon (l, r);
+	  freev (l);
+	}
+      else
+	v = l;
       freev (r);
       return v;
     }
   else if (nextarg ("index"))
     {
-      l = eval6 ();
-      r = eval6 ();
+      l = eval6 (evaluate);
+      r = eval6 (evaluate);
       tostring (l);
       tostring (r);
       v = int_value (strcspn (l->u.s, r->u.s) + 1);
@@ -527,9 +541,9 @@ eval6 (void)
     }
   else if (nextarg ("substr"))
     {
-      l = eval6 ();
-      i1 = eval6 ();
-      i2 = eval6 ();
+      l = eval6 (evaluate);
+      i1 = eval6 (evaluate);
+      i2 = eval6 (evaluate);
       tostring (l);
       if (!toarith (i1) || !toarith (i2)
 	  || strlen (l->u.s) < i1->u.i
@@ -537,7 +551,7 @@ eval6 (void)
 	v = str_value ("");
       else
 	{
-	  v = NEW (VALUE);
+	  v = xmalloc (sizeof *v);
 	  v->type = string;
 	  v->u.s = strncpy (xmalloc (i2->u.i + 1),
 			    l->u.s + i1->u.i - 1, i2->u.i);
@@ -549,14 +563,14 @@ eval6 (void)
       return v;
     }
   else
-    return eval7 ();
+    return eval7 (evaluate);
 }
 
 /* Handle : operator (pattern matching).
    Calls docolon to do the real work.  */
 
 static VALUE *
-eval5 (void)
+eval5 (bool evaluate)
 {
   VALUE *l;
   VALUE *r;
@@ -565,16 +579,19 @@ eval5 (void)
 #ifdef EVAL_TRACE
   trace ("eval5");
 #endif
-  l = eval6 ();
+  l = eval6 (evaluate);
   while (1)
     {
       if (nextarg (":"))
 	{
-	  r = eval6 ();
-	  v = docolon (l, r);
-	  freev (l);
+	  r = eval6 (evaluate);
+	  if (evaluate)
+	    {
+	      v = docolon (l, r);
+	      freev (l);
+	      l = v;
+	    }
 	  freev (r);
-	  l = v;
 	}
       else
 	return l;
@@ -584,17 +601,17 @@ eval5 (void)
 /* Handle *, /, % operators.  */
 
 static VALUE *
-eval4 (void)
+eval4 (bool evaluate)
 {
   VALUE *l;
   VALUE *r;
   enum { multiply, divide, mod } fxn;
-  intmax_t val;
+  intmax_t val = 0;
 
 #ifdef EVAL_TRACE
   trace ("eval4");
 #endif
-  l = eval5 ();
+  l = eval5 (evaluate);
   while (1)
     {
       if (nextarg ("*"))
@@ -605,16 +622,19 @@ eval4 (void)
 	fxn = mod;
       else
 	return l;
-      r = eval5 ();
-      if (!toarith (l) || !toarith (r))
-	error (EXPR_ERROR, 0, _("non-numeric argument"));
-      if (fxn == multiply)
-	val = l->u.i * r->u.i;
-      else
+      r = eval5 (evaluate);
+      if (evaluate)
 	{
-	  if (r->u.i == 0)
-	    error (EXPR_ERROR, 0, _("division by zero"));
-	  val = fxn == divide ? l->u.i / r->u.i : l->u.i % r->u.i;
+	  if (!toarith (l) || !toarith (r))
+	    error (EXPR_FAILURE, 0, _("non-numeric argument"));
+	  if (fxn == multiply)
+	    val = l->u.i * r->u.i;
+	  else
+	    {
+	      if (r->u.i == 0)
+		error (EXPR_FAILURE, 0, _("division by zero"));
+	      val = fxn == divide ? l->u.i / r->u.i : l->u.i % r->u.i;
+	    }
 	}
       freev (l);
       freev (r);
@@ -625,17 +645,17 @@ eval4 (void)
 /* Handle +, - operators.  */
 
 static VALUE *
-eval3 (void)
+eval3 (bool evaluate)
 {
   VALUE *l;
   VALUE *r;
   enum { plus, minus } fxn;
-  intmax_t val;
+  intmax_t val = 0;
 
 #ifdef EVAL_TRACE
   trace ("eval3");
 #endif
-  l = eval4 ();
+  l = eval4 (evaluate);
   while (1)
     {
       if (nextarg ("+"))
@@ -644,10 +664,13 @@ eval3 (void)
 	fxn = minus;
       else
 	return l;
-      r = eval4 ();
-      if (!toarith (l) || !toarith (r))
-	error (EXPR_ERROR, 0, _("non-numeric argument"));
-      val = fxn == plus ? l->u.i + r->u.i : l->u.i - r->u.i;
+      r = eval4 (evaluate);
+      if (evaluate)
+	{
+	  if (!toarith (l) || !toarith (r))
+	    error (EXPR_FAILURE, 0, _("non-numeric argument"));
+	  val = fxn == plus ? l->u.i + r->u.i : l->u.i - r->u.i;
+	}
       freev (l);
       freev (r);
       l = int_value (val);
@@ -657,7 +680,7 @@ eval3 (void)
 /* Handle comparisons.  */
 
 static VALUE *
-eval2 (void)
+eval2 (bool evaluate)
 {
   VALUE *l;
   VALUE *r;
@@ -674,7 +697,7 @@ eval2 (void)
 #ifdef EVAL_TRACE
   trace ("eval2");
 #endif
-  l = eval3 ();
+  l = eval3 (evaluate);
   while (1)
     {
       if (nextarg ("<"))
@@ -691,7 +714,7 @@ eval2 (void)
 	fxn = greater_than;
       else
 	return l;
-      r = eval3 ();
+      r = eval3 (evaluate);
       tostring (l);
       tostring (r);
 
@@ -709,11 +732,11 @@ eval2 (void)
 	  lval = l->u.i;
 	  rval = r->u.i;
 	}
-      else if (collation_errno)
+      else if (collation_errno && evaluate)
 	{
 	  error (0, collation_errno, _("string comparison failed"));
 	  error (0, 0, _("Set LC_ALL='C' to work around the problem."));
-	  error (EXPR_ERROR, 0,
+	  error (EXPR_FAILURE, 0,
 		 _("The strings compared were %s and %s."),
 		 quotearg_n_style (0, locale_quoting_style, collation_arg1),
 		 quotearg_n_style (1, locale_quoting_style, r->u.s));
@@ -739,7 +762,7 @@ eval2 (void)
 /* Handle &.  */
 
 static VALUE *
-eval1 (void)
+eval1 (bool evaluate)
 {
   VALUE *l;
   VALUE *r;
@@ -747,12 +770,12 @@ eval1 (void)
 #ifdef EVAL_TRACE
   trace ("eval1");
 #endif
-  l = eval2 ();
+  l = eval2 (evaluate);
   while (1)
     {
       if (nextarg ("&"))
 	{
-	  r = eval2 ();
+	  r = eval2 (evaluate & ~ null (l));
 	  if (null (l) || null (r))
 	    {
 	      freev (l);
@@ -770,7 +793,7 @@ eval1 (void)
 /* Handle |.  */
 
 static VALUE *
-eval (void)
+eval (bool evaluate)
 {
   VALUE *l;
   VALUE *r;
@@ -778,16 +801,21 @@ eval (void)
 #ifdef EVAL_TRACE
   trace ("eval");
 #endif
-  l = eval1 ();
+  l = eval1 (evaluate);
   while (1)
     {
       if (nextarg ("|"))
 	{
-	  r = eval1 ();
+	  r = eval1 (evaluate & null (l));
 	  if (null (l))
 	    {
 	      freev (l);
 	      l = r;
+	      if (null (l))
+		{
+		  freev (l);
+		  l = int_value (0);
+		}
 	    }
 	  else
 	    freev (r);
