@@ -1,5 +1,5 @@
 /* join - join lines of two files on a common field
-   Copyright (C) 91, 1995-2003 Free Software Foundation, Inc.
+   Copyright (C) 91, 1995-2004 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -150,7 +150,7 @@ get_option (int argc, char **argv)
 void
 usage (int status)
 {
-  if (status != 0)
+  if (status != EXIT_SUCCESS)
     fprintf (stderr, _("Try `%s --help' for more information.\n"),
 	     program_name);
   else
@@ -194,7 +194,7 @@ Important: FILE1 and FILE2 must be sorted on the join fields.\n\
 "), stdout);
       printf (_("\nReport bugs to <%s>.\n"), PACKAGE_BUGREPORT);
     }
-  exit (status == 0 ? EXIT_SUCCESS : EXIT_FAILURE);
+  exit (status);
 }
 
 /* Return true if C is a blank (a default input field separator).  */
@@ -378,7 +378,7 @@ keycmp (struct line const *line1, struct line const *line2)
     }
   else
     {
-      if (HAVE_SETLOCALE && hard_LC_COLLATE)
+      if (hard_LC_COLLATE)
 	return xmemcoll (beg1, len1, beg2, len2);
       diff = memcmp (beg1, beg2, MIN (len1, len2));
     }
@@ -635,14 +635,14 @@ add_field (int file, size_t field)
    (the zero-based field number).  If it cannot be converted, give a
    diagnostic and exit.  */
 
-size_t
+static size_t
 string_to_join_field (char const *str, char const *err_msg_fmt)
 {
   size_t result;
   uintmax_t val;
 
   strtol_error s_err = xstrtoumax (str, NULL, 10, &val, "");
-  if (s_err == LONGINT_OVERFLOW || SIZE_MAX < val)
+  if (s_err == LONGINT_OVERFLOW || (s_err == LONGINT_OK && SIZE_MAX < val))
     {
       error (EXIT_FAILURE, 0,
 	     _("value %s is so large that it is not representable"),
@@ -731,12 +731,12 @@ add_field_list (char *str)
 /* Add NAME to the array of input file NAMES; currently there are
    *NFILES names in the list.  */
 
-void
+static void
 add_file_name (char const *name, char const *names[2], int *nfiles)
 {
   if (*nfiles == 2)
     {
-      error (0, 0, _("too many non-option arguments"));
+      error (0, 0, _("extra operand %s"), quote (name));
       usage (EXIT_FAILURE);
     }
   names[(*nfiles)++] = name;
@@ -768,9 +768,6 @@ main (int argc, char **argv)
 
       switch (optc)
 	{
-	case 0:
-	  break;
-
 	case 'v':
 	    print_pairables = false;
 	    /* Fall through.  */
@@ -848,7 +845,10 @@ main (int argc, char **argv)
 
   if (nfiles != 2)
     {
-      error (0, 0, _("too few non-option arguments"));
+      if (nfiles == 0)
+	error (0, 0, _("missing operand"));
+      else
+	error (0, 0, _("missing operand after %s"), quote (argv[argc - 1]));
       usage (EXIT_FAILURE);
     }
 
