@@ -1,5 +1,5 @@
 /* printenv -- print all or part of environment
-   Copyright (C) 1989-1997, 1999-2002 Free Software Foundation, Inc.
+   Copyright (C) 1989-1997, 1999-2004 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -24,6 +24,7 @@
    Exit status:
    0 if all variables specified were found
    1 if not
+   2 if some other error occurred
 
    David MacKenzie and Richard Mlynarik */
 
@@ -33,29 +34,26 @@
 #include <getopt.h>
 
 #include "system.h"
-#include "closeout.h"
 #include "error.h"
 #include "long-options.h"
 
+/* Exit status for syntax errors, etc.  */
+enum { PRINTENV_FAILURE = 2 };
+
 /* The official name of this program (e.g., no `g' prefix).  */
 #define PROGRAM_NAME "printenv"
 
-#define AUTHORS N_ ("David MacKenzie and Richard Mlynarik")
+#define AUTHORS "David MacKenzie", "Richard Mlynarik"
 
 /* The name this program was run with. */
 char *program_name;
 
-static struct option const long_options[] =
-{
-  {0, 0, 0, 0}
-};
-
 extern char **environ;
 
 void
 usage (int status)
 {
-  if (status != 0)
+  if (status != EXIT_SUCCESS)
     fprintf (stderr, _("Try `%s --help' for more information.\n"),
 	     program_name);
   else
@@ -80,43 +78,36 @@ main (int argc, char **argv)
   char **env;
   char *ep, *ap;
   int i;
-  int matches = 0;
-  int c;
-  int exit_status;
+  bool ok;
 
+  initialize_main (&argc, &argv);
   program_name = argv[0];
   setlocale (LC_ALL, "");
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
-  close_stdout_set_status (2);
+  initialize_exit_failure (PRINTENV_FAILURE);
   atexit (close_stdout);
 
   parse_long_options (argc, argv, PROGRAM_NAME, GNU_PACKAGE, VERSION,
-		      AUTHORS, usage);
+		      usage, AUTHORS, (char const *) NULL);
+  if (getopt_long (argc, argv, "+", NULL, NULL) != -1)
+    usage (PRINTENV_FAILURE);
 
-  while ((c = getopt_long (argc, argv, "", long_options, NULL)) != -1)
-    {
-      switch (c)
-	{
-	case 0:
-	  break;
-
-	default:
-	  usage (EXIT_FAILURE);
-	}
-    }
-
-  if (optind == argc)
+  if (optind >= argc)
     {
       for (env = environ; *env != NULL; ++env)
 	puts (*env);
-      exit_status = 0;
+      ok = true;
     }
   else
     {
+      int matches = 0;
+
       for (i = optind; i < argc; ++i)
 	{
+	  bool matched = false;
+
 	  for (env = environ; *env; ++env)
 	    {
 	      ep = *env;
@@ -126,14 +117,17 @@ main (int argc, char **argv)
 		  if (*ep == '=' && *ap == '\0')
 		    {
 		      puts (ep + 1);
-		      ++matches;
+		      matched = true;
 		      break;
 		    }
 		}
 	    }
+
+	  matches += matched;
 	}
-      exit_status = (matches != argc - optind);
+
+      ok = (matches == argc - optind);
     }
 
-  exit (exit_status);
+  exit (ok ? EXIT_SUCCESS : EXIT_FAILURE);
 }
