@@ -1,5 +1,5 @@
 /* printf - format and print data
-   Copyright (C) 1990-2003, Free Software Foundation, Inc.
+   Copyright (C) 1990-2004 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -52,34 +52,33 @@
 #include <getopt.h>
 
 #include "system.h"
+#include "c-strtod.h"
 #include "long-options.h"
 #include "error.h"
 #include "unicodeio.h"
 
+#if ! (HAVE_DECL_STRTOIMAX || defined strtoimax)
+intmax_t strtoimax ();
+#endif
+#if ! (HAVE_DECL_STRTOUMAX || defined strtoumax)
+uintmax_t strtoumax ();
+#endif
+
 /* The official name of this program (e.g., no `g' prefix).  */
 #define PROGRAM_NAME "printf"
 
 #define AUTHORS "David MacKenzie"
 
-#ifndef STDC_HEADERS
-double strtod ();
-long int strtol ();
-unsigned long int strtoul ();
-#endif
-
 #define isodigit(c) ((c) >= '0' && (c) <= '7')
 #define hextobin(c) ((c) >= 'a' && (c) <= 'f' ? (c) - 'a' + 10 : \
 		     (c) >= 'A' && (c) <= 'F' ? (c) - 'A' + 10 : (c) - '0')
 #define octtobin(c) ((c) - '0')
 
-/* A value for field_width or precision that indicates it was not specified.  */
-#define UNSPECIFIED INT_MIN
-
 /* The value to return to the calling program.  */
 static int exit_status;
 
-/* Non-zero if the POSIXLY_CORRECT environment variable is set.  */
-static int posixly_correct;
+/* True if the POSIXLY_CORRECT environment variable is set.  */
+static bool posixly_correct;
 
 /* This message appears in N_() here rather than just in _() below because
    the sole use would have been in a #define.  */
@@ -92,7 +91,7 @@ char *program_name;
 void
 usage (int status)
 {
-  if (status != 0)
+  if (status != EXIT_SUCCESS)
     fprintf (stderr, _("Try `%s --help' for more information.\n"),
 	     program_name);
   else
@@ -129,10 +128,9 @@ FORMAT controls the output as in C printf.  Interpreted sequences are:\n\
   \\v      vertical tab\n\
 "), stdout);
       fputs (_("\
-  \\xNN    byte with hexadecimal value NN (1 to 2 digits)\n\
-\n\
-  \\uNNNN  character with hexadecimal value NNNN (4 digits)\n\
-  \\UNNNNNNNN  character with hexadecimal value NNNNNNNN (8 digits)\n\
+  \\xHH    byte with hexadecimal value HH (1 to 2 digits)\n\
+  \\uHHHH  Unicode (ISO/IEC 10646) character with hex value HHHH (4 digits)\n\
+  \\UHHHHHHHH  Unicode character with hex value HHHHHHHH (8 digits)\n\
 "), stdout);
       fputs (_("\
   %%      a single %\n\
@@ -153,7 +151,7 @@ verify (const char *s, const char *end)
   if (errno)
     {
       error (0, errno, "%s", s);
-      exit_status = 1;
+      exit_status = EXIT_FAILURE;
     }
   else if (*end)
     {
@@ -161,21 +159,21 @@ verify (const char *s, const char *end)
 	error (0, 0, _("%s: expected a numeric value"), s);
       else
 	error (0, 0, _("%s: value not completely converted"), s);
-      exit_status = 1;
+      exit_status = EXIT_FAILURE;
     }
 }
 
 #define STRTOX(TYPE, FUNC_NAME, LIB_FUNC_EXPR)				 \
 static TYPE								 \
-FUNC_NAME (s)								 \
-     const char *s;							 \
+FUNC_NAME (char const *s)						 \
 {									 \
   char *end;								 \
   TYPE val;								 \
 									 \
   if (*s == '\"' || *s == '\'')						 \
     {									 \
-      val = *(unsigned char *) ++s;					 \
+      unsigned char ch = *++s;						 \
+      val = ch;								 \
       /* If POSIXLY_CORRECT is not set, then give a warning that there	 \
 	 are characters following the character constant and that GNU	 \
 	 printf is ignoring those characters.  If POSIXLY_CORRECT *is*	 \
@@ -186,46 +184,46 @@ FUNC_NAME (s)								 \
   else									 \
     {									 \
       errno = 0;							 \
-      val = LIB_FUNC_EXPR;						 \
+      val = (LIB_FUNC_EXPR);						 \
       verify (s, end);							 \
     }									 \
   return val;								 \
 }									 \
 
-STRTOX (unsigned long int, xstrtoul, (strtoul (s, &end, 0)))
-STRTOX (long int,          xstrtol,  (strtol  (s, &end, 0)))
-STRTOX (double,            xstrtod,  (strtod  (s, &end)))
+STRTOX (intmax_t,    vstrtoimax, strtoimax (s, &end, 0))
+STRTOX (uintmax_t,   vstrtoumax, strtoumax (s, &end, 0))
+STRTOX (long double, vstrtold,   c_strtold (s, &end))
 
 /* Output a single-character \ escape.  */
 
 static void
-print_esc_char (int c)
+print_esc_char (char c)
 {
   switch (c)
     {
     case 'a':			/* Alert. */
-      putchar (7);
+      putchar ('\a');
       break;
     case 'b':			/* Backspace. */
-      putchar (8);
+      putchar ('\b');
       break;
     case 'c':			/* Cancel the rest of the output. */
       exit (EXIT_SUCCESS);
       break;
     case 'f':			/* Form feed. */
-      putchar (12);
+      putchar ('\f');
       break;
     case 'n':			/* New line. */
-      putchar (10);
+      putchar ('\n');
       break;
     case 'r':			/* Carriage return. */
-      putchar (13);
+      putchar ('\r');
       break;
     case 't':			/* Horizontal tab. */
-      putchar (9);
+      putchar ('\t');
       break;
     case 'v':			/* Vertical tab. */
-      putchar (11);
+      putchar ('\v');
       break;
     default:
       putchar (c);
@@ -246,7 +244,7 @@ print_esc (const char *escstart, bool octal_0)
   int esc_value = 0;		/* Value of \nnn escape. */
   int esc_length;		/* Length of \nnn escape. */
 
-  if (!posixly_correct && *p == 'x')
+  if (*p == 'x')
     {
       /* A hexadecimal \xhh escape sequence must have 1 or 2 hex. digits.  */
       for (esc_length = 0, ++p;
@@ -270,7 +268,7 @@ print_esc (const char *escstart, bool octal_0)
     }
   else if (*p && strchr ("\"\\abcfnrtv", *p))
     print_esc_char (*p++);
-  else if (!posixly_correct && (*p == 'u' || *p == 'U'))
+  else if (*p == 'u' || *p == 'U')
     {
       char esc_char = *p;
       unsigned int uni_value;
@@ -322,97 +320,147 @@ print_esc_string (const char *str)
       putchar (*str);
 }
 
-/* Output a % directive.  START is the start of the directive,
-   LENGTH is its length, and ARGUMENT is its argument.
-   If FIELD_WIDTH or PRECISION is UNSPECIFIED, they are args for
-   '*' values in those fields. */
+/* Evaluate a printf conversion specification.  START is the start of
+   the directive, LENGTH is its length, and CONVERSION specifies the
+   type of conversion.  LENGTH does not include any length modifier or
+   the conversion specifier itself.  FIELD_WIDTH and PRECISION are the
+   field width and precision for '*' values, if HAVE_FIELD_WIDTH and
+   HAVE_PRECISION are true, respectively.  ARGUMENT is the argument to
+   be formatted.  */
 
 static void
-print_direc (const char *start, size_t length, int field_width,
-	     int precision, const char *argument)
+print_direc (const char *start, size_t length, char conversion,
+	     bool have_field_width, int field_width,
+	     bool have_precision, int precision,
+	     char const *argument)
 {
   char *p;		/* Null-terminated copy of % directive. */
 
-  p = xmalloc ((unsigned) (length + 1));
-  strncpy (p, start, length);
-  p[length] = 0;
+  /* Create a null-terminated copy of the % directive, with an
+     intmax_t-wide length modifier substituted for any existing
+     integer length modifier.  */
+  {
+    char *q;
+    char const *length_modifier;
+    size_t length_modifier_len;
 
-  switch (p[length - 1])
+    switch (conversion)
+      {
+      case 'd': case 'i': case 'o': case 'u': case 'x': case 'X':
+	length_modifier = PRIdMAX;
+	length_modifier_len = sizeof PRIdMAX - 2;
+	break;
+
+      case 'a': case 'e': case 'f': case 'g':
+      case 'A': case 'E': case 'F': case 'G':
+	length_modifier = "L";
+	length_modifier_len = 1;
+	break;
+
+      default:
+	length_modifier = start;  /* Any valid pointer will do.  */
+	length_modifier_len = 0;
+	break;
+      }
+
+    p = xmalloc (length + length_modifier_len + 2);
+    q = mempcpy (p, start, length);
+    q = mempcpy (q, length_modifier, length_modifier_len);
+    *q++ = conversion;
+    *q = '\0';
+  }
+
+  switch (conversion)
     {
     case 'd':
     case 'i':
-      if (field_width == UNSPECIFIED)
-	{
-	  if (precision == UNSPECIFIED)
-	    printf (p, xstrtol (argument));
-	  else
-	    printf (p, precision, xstrtol (argument));
-	}
-      else
-	{
-	  if (precision == UNSPECIFIED)
-	    printf (p, field_width, xstrtol (argument));
-	  else
-	    printf (p, field_width, precision, xstrtol (argument));
-	}
+      {
+	intmax_t arg = vstrtoimax (argument);
+	if (!have_field_width)
+	  {
+	    if (!have_precision)
+	      printf (p, arg);
+	    else
+	      printf (p, precision, arg);
+	  }
+	else
+	  {
+	    if (!have_precision)
+	      printf (p, field_width, arg);
+	    else
+	      printf (p, field_width, precision, arg);
+	  }
+      }
       break;
 
     case 'o':
     case 'u':
     case 'x':
     case 'X':
-      if (field_width == UNSPECIFIED)
-	{
-	  if (precision == UNSPECIFIED)
-	    printf (p, xstrtoul (argument));
-	  else
-	    printf (p, precision, xstrtoul (argument));
-	}
-      else
-	{
-	  if (precision == UNSPECIFIED)
-	    printf (p, field_width, xstrtoul (argument));
-	  else
-	    printf (p, field_width, precision, xstrtoul (argument));
-	}
+      {
+	uintmax_t arg = vstrtoumax (argument);
+	if (!have_field_width)
+	  {
+	    if (!have_precision)
+	      printf (p, arg);
+	    else
+	      printf (p, precision, arg);
+	  }
+	else
+	  {
+	    if (!have_precision)
+	      printf (p, field_width, arg);
+	    else
+	      printf (p, field_width, precision, arg);
+	  }
+      }
       break;
 
-    case 'f':
+    case 'a':
+    case 'A':
     case 'e':
     case 'E':
+    case 'f':
+    case 'F':
     case 'g':
     case 'G':
-      if (field_width == UNSPECIFIED)
-	{
-	  if (precision == UNSPECIFIED)
-	    printf (p, xstrtod (argument));
-	  else
-	    printf (p, precision, xstrtod (argument));
-	}
-      else
-	{
-	  if (precision == UNSPECIFIED)
-	    printf (p, field_width, xstrtod (argument));
-	  else
-	    printf (p, field_width, precision, xstrtod (argument));
-	}
+      {
+	long double arg = vstrtold (argument);
+	if (!have_field_width)
+	  {
+	    if (!have_precision)
+	      printf (p, arg);
+	    else
+	      printf (p, precision, arg);
+	  }
+	else
+	  {
+	    if (!have_precision)
+	      printf (p, field_width, arg);
+	    else
+	      printf (p, field_width, precision, arg);
+	  }
+      }
       break;
 
     case 'c':
-      printf (p, *argument);
+      if (!have_field_width)
+	printf (p, *argument);
+      else
+	printf (p, field_width, *argument);
       break;
 
     case 's':
-      if (field_width == UNSPECIFIED)
+      if (!have_field_width)
 	{
-	  if (precision == UNSPECIFIED)
+	  if (!have_precision)
 	    printf (p, argument);
 	  else
 	    printf (p, precision, argument);
 	}
       else
 	{
-	  if (precision == UNSPECIFIED)
+	  if (!have_precision)
 	    printf (p, field_width, argument);
 	  else
 	    printf (p, field_width, precision, argument);
@@ -434,8 +482,11 @@ print_formatted (const char *format, int argc, char **argv)
   const char *f;		/* Pointer into `format'.  */
   const char *direc_start;	/* Start of % directive.  */
   size_t direc_length;		/* Length of % directive.  */
-  int field_width;		/* Arg to first '*', or UNSPECIFIED if none. */
-  int precision;		/* Arg to second '*', or UNSPECIFIED if none. */
+  bool have_field_width;	/* True if FIELD_WIDTH is valid.  */
+  int field_width = 0;		/* Arg to first '*'.  */
+  bool have_precision;		/* True if PRECISION is valid.  */
+  int precision = 0;		/* Arg to second '*'.  */
+  char ok[UCHAR_MAX + 1];	/* ok['x'] is true if %x is allowed.  */
 
   for (f = format; *f; ++f)
     {
@@ -444,7 +495,7 @@ print_formatted (const char *format, int argc, char **argv)
 	case '%':
 	  direc_start = f++;
 	  direc_length = 1;
-	  field_width = precision = UNSPECIFIED;
+	  have_field_width = have_precision = false;
 	  if (*f == '%')
 	    {
 	      putchar ('%');
@@ -452,6 +503,8 @@ print_formatted (const char *format, int argc, char **argv)
 	    }
 	  if (*f == 'b')
 	    {
+	      /* FIXME: Field width and precision are not supported
+		 for %b, even though POSIX requires it.  */
 	      if (argc > 0)
 		{
 		  print_esc_string (*argv);
@@ -460,19 +513,42 @@ print_formatted (const char *format, int argc, char **argv)
 		}
 	      break;
 	    }
-	  while (*f == ' ' || *f == '#' || *f == '+' || *f == '-')
-	    {
-	      ++f;
-	      ++direc_length;
-	    }
+
+	  memset (ok, 0, sizeof ok);
+	  ok['a'] = ok['A'] = ok['c'] = ok['d'] = ok['e'] = ok['E'] =
+	    ok['f'] = ok['F'] = ok['g'] = ok['G'] = ok['i'] = ok['o'] =
+	    ok['s'] = ok['u'] = ok['x'] = ok['X'] = 1;
+
+	  for (;; f++, direc_length++)
+	    switch (*f)
+	      {
+	      case '\'':
+		ok['a'] = ok['A'] = ok['c'] = ok['e'] = ok['E'] =
+		  ok['o'] = ok['s'] = ok['x'] = ok['X'] = 0;
+		break;
+	      case '-': case '+': case ' ':
+		break;
+	      case '#':
+		ok['c'] = ok['d'] = ok['i'] = ok['s'] = ok['u'] = 0;
+		break;
+	      case '0':
+		ok['c'] = ok['s'] = 0;
+		break;
+	      default:
+		goto no_more_flag_characters;
+	      }
+	no_more_flag_characters:;
+
 	  if (*f == '*')
 	    {
 	      ++f;
 	      ++direc_length;
 	      if (argc > 0)
 		{
-		  field_width = xstrtoul (*argv);
-		  if (field_width == UNSPECIFIED)
+		  intmax_t width = vstrtoimax (*argv);
+		  if (INT_MIN <= width && width <= INT_MAX)
+		    field_width = width;
+		  else
 		    error (EXIT_FAILURE, 0, _("invalid field width: %s"),
 			   *argv);
 		  ++argv;
@@ -480,6 +556,7 @@ print_formatted (const char *format, int argc, char **argv)
 		}
 	      else
 		field_width = 0;
+	      have_field_width = true;
 	    }
 	  else
 	    while (ISDIGIT (*f))
@@ -491,21 +568,32 @@ print_formatted (const char *format, int argc, char **argv)
 	    {
 	      ++f;
 	      ++direc_length;
+	      ok['c'] = 0;
 	      if (*f == '*')
 		{
 		  ++f;
 		  ++direc_length;
 		  if (argc > 0)
 		    {
-		      precision = xstrtoul (*argv);
-		      if (precision == UNSPECIFIED)
+		      intmax_t prec = vstrtoimax (*argv);
+		      if (prec < 0)
+			{
+			  /* A negative precision is taken as if the
+			     precision were omitted, so -1 is safe
+			     here even if prec < INT_MIN.  */
+			  precision = -1;
+			}
+		      else if (INT_MAX < prec)
 			error (EXIT_FAILURE, 0, _("invalid precision: %s"),
 			       *argv);
+		      else
+			precision = prec;
 		      ++argv;
 		      --argc;
 		    }
 		  else
 		    precision = 0;
+		  have_precision = true;
 		}
 	      else
 		while (ISDIGIT (*f))
@@ -514,24 +602,23 @@ print_formatted (const char *format, int argc, char **argv)
 		    ++direc_length;
 		  }
 	    }
-	  if (*f == 'l' || *f == 'L' || *f == 'h')
-	    {
-	      ++f;
-	      ++direc_length;
-	    }
-	  if (! (*f && strchr ("diouxXfeEgGcs", *f)))
-	    error (EXIT_FAILURE, 0, _("%%%c: invalid directive"), *f);
-	  ++direc_length;
-	  if (argc > 0)
-	    {
-	      print_direc (direc_start, direc_length, field_width,
-			   precision, *argv);
-	      ++argv;
-	      --argc;
-	    }
-	  else
-	    print_direc (direc_start, direc_length, field_width,
-			 precision, "");
+
+	  while (*f == 'l' || *f == 'L' || *f == 'h'
+		 || *f == 'j' || *f == 't' || *f == 'z')
+	    ++f;
+
+	  {
+	    unsigned char conversion = *f;
+	    if (! ok[conversion])
+	      error (EXIT_FAILURE, 0,
+		     _("%.*s: invalid conversion specification"),
+		     (int) (f + 1 - direc_start), direc_start);
+	  }
+
+	  print_direc (direc_start, direc_length, *f,
+		       have_field_width, field_width,
+		       have_precision, precision,
+		       (argc <= 0 ? "" : (argc--, *argv++)));
 	  break;
 
 	case '\\':
@@ -560,13 +647,12 @@ main (int argc, char **argv)
 
   atexit (close_stdout);
 
-  exit_status = 0;
+  exit_status = EXIT_SUCCESS;
 
-  /* Don't recognize --help or --version if POSIXLY_CORRECT is set.  */
   posixly_correct = (getenv ("POSIXLY_CORRECT") != NULL);
-  if (!posixly_correct)
-    parse_long_options (argc, argv, PROGRAM_NAME, GNU_PACKAGE, VERSION,
-			AUTHORS, usage);
+
+  parse_long_options (argc, argv, PROGRAM_NAME, GNU_PACKAGE, VERSION,
+		      usage, AUTHORS, (char const *) NULL);
 
   /* The above handles --help and --version.
      Since there is no other invocation of getopt, handle `--' here.  */
@@ -578,8 +664,8 @@ main (int argc, char **argv)
 
   if (argc <= 1)
     {
-      fprintf (stderr, _("Usage: %s format [argument...]\n"), program_name);
-      exit (EXIT_FAILURE);
+      error (0, 0, _("missing operand"));
+      usage (EXIT_FAILURE);
     }
 
   format = argv[1];
