@@ -1,4 +1,4 @@
-/* stat.c -- display file or filesystem status
+/* stat.c -- display file or file system status
    Copyright (C) 2001, 2002, 2003, 2004 Free Software Foundation.
 
    This program is free software; you can redistribute it and/or modify
@@ -59,18 +59,25 @@
 
 #if HAVE_STRUCT_STATVFS_F_BASETYPE
 # define STRUCT_STATVFS struct statvfs
-# define HAVE_STRUCT_STATXFS_F_TYPE HAVE_STRUCT_STATVFS
+# define HAVE_STRUCT_STATXFS_F_TYPE HAVE_STRUCT_STATVFS_F_TYPE
 # if HAVE_STRUCT_STATVFS_F_NAMEMAX
 #  define SB_F_NAMEMAX(S) ((uintmax_t) ((S)->f_namemax))
 # endif
+# if STAT_STATVFS
+#  define STATFS statvfs
+# endif
 #else
 # define STRUCT_STATVFS struct statfs
-# define HAVE_STRUCT_STATXFS_F_TYPE HAVE_STRUCT_STATFS
+# define HAVE_STRUCT_STATXFS_F_TYPE HAVE_STRUCT_STATFS_F_TYPE
 # if HAVE_STRUCT_STATFS_F_NAMELEN
 #  define SB_F_NAMEMAX(S) ((uintmax_t) ((S)->f_namelen))
 # endif
 #endif
 
+#ifndef STATFS
+# define STATFS statfs
+#endif
+
 #ifndef SB_F_NAMEMAX
 /* NetBSD 1.5.2 has neither f_namemax nor f_namelen.  */
 # define SB_F_NAMEMAX(S) "*"
@@ -93,31 +100,27 @@
 static struct option const long_options[] = {
   {"link", no_argument, 0, 'l'}, /* deprecated.  FIXME: remove in 2003 */
   {"dereference", no_argument, 0, 'L'},
+  {"file-system", no_argument, 0, 'f'},
+  {"filesystem", no_argument, 0, 'f'}, /* obsolete and undocumented alias */
   {"format", required_argument, 0, 'c'},
-  {"filesystem", no_argument, 0, 'f'},
   {"terse", no_argument, 0, 't'},
   {GETOPT_HELP_OPTION_DECL},
   {GETOPT_VERSION_OPTION_DECL},
   {NULL, 0, NULL, 0}
 };
 
-/* Nonzero means we should exit with EXIT_FAILURE upon completion.  */
-static int G_fail;
-
 char *program_name;
 
 /* Return the type of the specified file system.
    Some systems have statfvs.f_basetype[FSTYPSZ]. (AIX, HP-UX, and Solaris)
    Others have statfs.f_fstypename[MFSNAMELEN]. (NetBSD 1.5.2)
    Still others have neither and have to get by with f_type (Linux).  */
-static char *
+static char const *
 human_fstype (STRUCT_STATVFS const *statfsbuf)
 {
 #ifdef STATXFS_FILE_SYSTEM_TYPE_MEMBER_NAME
-  /* Cast away the `const' attribute.  */
-  return (char *) statfsbuf->STATXFS_FILE_SYSTEM_TYPE_MEMBER_NAME;
+  return statfsbuf->STATXFS_FILE_SYSTEM_TYPE_MEMBER_NAME;
 #else
-  char const *type;
   switch (statfsbuf->f_type)
     {
 # if defined __linux__
@@ -128,192 +131,136 @@ human_fstype (STRUCT_STATVFS const *statfsbuf)
 	 combined to produce the #define directives in fs.h.  */
 
     case S_MAGIC_AFFS: /* 0xADFF */
-      type = "affs";
-      break;
+      return "affs";
     case S_MAGIC_DEVPTS: /* 0x1CD1 */
-      type = "devpts";
-      break;
+      return "devpts";
     case S_MAGIC_EXT: /* 0x137D */
-      type = "ext";
-      break;
+      return "ext";
     case S_MAGIC_EXT2_OLD: /* 0xEF51 */
-      type = "ext2";
-      break;
+      return "ext2";
     case S_MAGIC_EXT2: /* 0xEF53 */
-      type = "ext2/ext3";
-      break;
+      return "ext2/ext3";
     case S_MAGIC_HPFS: /* 0xF995E849 */
-      type = "hpfs";
-      break;
+      return "hpfs";
     case S_MAGIC_ISOFS: /* 0x9660 */
-      type = "isofs";
-      break;
+      return "isofs";
     case S_MAGIC_ISOFS_WIN: /* 0x4000 */
-      type = "isofs";
-      break;
+      return "isofs";
     case S_MAGIC_ISOFS_R_WIN: /* 0x4004 */
-      type = "isofs";
-      break;
+      return "isofs";
     case S_MAGIC_MINIX: /* 0x137F */
-      type = "minix";
-      break;
+      return "minix";
     case S_MAGIC_MINIX_30: /* 0x138F */
-      type = "minix (30 char.)";
-      break;
+      return "minix (30 char.)";
     case S_MAGIC_MINIX_V2: /* 0x2468 */
-      type = "minix v2";
-      break;
+      return "minix v2";
     case S_MAGIC_MINIX_V2_30: /* 0x2478 */
-      type = "minix v2 (30 char.)";
-      break;
+      return "minix v2 (30 char.)";
     case S_MAGIC_MSDOS: /* 0x4d44 */
-      type = "msdos";
-      break;
+      return "msdos";
     case S_MAGIC_FAT: /* 0x4006 */
-      type = "fat";
-      break;
+      return "fat";
     case S_MAGIC_NCP: /* 0x564c */
-      type = "novell";
-      break;
+      return "novell";
     case S_MAGIC_NFS: /* 0x6969 */
-      type = "nfs";
-      break;
+      return "nfs";
     case S_MAGIC_PROC: /* 0x9fa0 */
-      type = "proc";
-      break;
+      return "proc";
     case S_MAGIC_SMB: /* 0x517B */
-      type = "smb";
-      break;
+      return "smb";
     case S_MAGIC_XENIX: /* 0x012FF7B4 */
-      type = "xenix";
-      break;
+      return "xenix";
     case S_MAGIC_SYSV4: /* 0x012FF7B5 */
-      type = "sysv4";
-      break;
+      return "sysv4";
     case S_MAGIC_SYSV2: /* 0x012FF7B6 */
-      type = "sysv2";
-      break;
+      return "sysv2";
     case S_MAGIC_COH: /* 0x012FF7B7 */
-      type = "coh";
-      break;
+      return "coh";
     case S_MAGIC_UFS: /* 0x00011954 */
-      type = "ufs";
-      break;
+      return "ufs";
     case S_MAGIC_XIAFS: /* 0x012FD16D */
-      type = "xia";
-      break;
+      return "xia";
     case S_MAGIC_NTFS: /* 0x5346544e */
-      type = "ntfs";
-      break;
+      return "ntfs";
     case S_MAGIC_TMPFS: /* 0x1021994 */
-      type = "tmpfs";
-      break;
+      return "tmpfs";
     case S_MAGIC_REISERFS: /* 0x52654973 */
-      type = "reiserfs";
-      break;
+      return "reiserfs";
     case S_MAGIC_CRAMFS: /* 0x28cd3d45 */
-      type = "cramfs";
-      break;
+      return "cramfs";
     case S_MAGIC_ROMFS: /* 0x7275 */
-      type = "romfs";
-      break;
+      return "romfs";
+    case S_MAGIC_RAMFS: /* 0x858458f6 */
+      return "ramfs";
+    case S_MAGIC_SQUASHFS: /* 0x73717368 */
+      return "squashfs";
+    case S_MAGIC_SYSFS: /* 0x62656572 */
+      return "sysfs";
 # elif __GNU__
     case FSTYPE_UFS:
-      type = "ufs";
-      break;
+      return "ufs";
     case FSTYPE_NFS:
-      type = "nfs";
-      break;
+      return "nfs";
     case FSTYPE_GFS:
-      type = "gfs";
-      break;
+      return "gfs";
     case FSTYPE_LFS:
-      type = "lfs";
-      break;
+      return "lfs";
     case FSTYPE_SYSV:
-      type = "sysv";
-      break;
+      return "sysv";
     case FSTYPE_FTP:
-      type = "ftp";
-      break;
+      return "ftp";
     case FSTYPE_TAR:
-      type = "tar";
-      break;
+      return "tar";
     case FSTYPE_AR:
-      type = "ar";
-      break;
+      return "ar";
     case FSTYPE_CPIO:
-      type = "cpio";
-      break;
+      return "cpio";
     case FSTYPE_MSLOSS:
-      type = "msloss";
-      break;
+      return "msloss";
     case FSTYPE_CPM:
-      type = "cpm";
-      break;
+      return "cpm";
     case FSTYPE_HFS:
-      type = "hfs";
-      break;
+      return "hfs";
     case FSTYPE_DTFS:
-      type = "dtfs";
-      break;
+      return "dtfs";
     case FSTYPE_GRFS:
-      type = "grfs";
-      break;
+      return "grfs";
     case FSTYPE_TERM:
-      type = "term";
-      break;
+      return "term";
     case FSTYPE_DEV:
-      type = "dev";
-      break;
+      return "dev";
     case FSTYPE_PROC:
-      type = "proc";
-      break;
+      return "proc";
     case FSTYPE_IFSOCK:
-      type = "ifsock";
-      break;
+      return "ifsock";
     case FSTYPE_AFS:
-      type = "afs";
-      break;
+      return "afs";
     case FSTYPE_DFS:
-      type = "dfs";
-      break;
+      return "dfs";
     case FSTYPE_PROC9:
-      type = "proc9";
-      break;
+      return "proc9";
     case FSTYPE_SOCKET:
-      type = "socket";
-      break;
+      return "socket";
     case FSTYPE_MISC:
-      type = "misc";
-      break;
+      return "misc";
     case FSTYPE_EXT2FS:
-      type = "ext2/ext3";
-      break;
+      return "ext2/ext3";
     case FSTYPE_HTTP:
-      type = "http";
-      break;
+      return "http";
     case FSTYPE_MEMFS:
-      type = "memfs";
-      break;
+      return "memfs";
     case FSTYPE_ISO9660:
-      type = "iso9660";
-      break;
+      return "iso9660";
 # endif
     default:
-      type = NULL;
-      break;
+      {
+	unsigned long int type = statfsbuf->f_type;
+	static char buf[sizeof "UNKNOWN (0x%lx)" - 3
+			+ (sizeof type * CHAR_BIT + 3) / 4];
+	sprintf (buf, "UNKNOWN (0x%lx)", type);
+	return buf;
+      }
     }
-
-  if (type)
-    return (char *) type;
-
-  {
-    static char buf[sizeof "UNKNOWN (0x%lx)" - 3
-		    + 2 * sizeof (statfsbuf->f_type)];
-    sprintf (buf, "UNKNOWN (0x%lx)", (unsigned long) statfsbuf->f_type);
-    return buf;
-  }
 #endif
 }
 
@@ -374,7 +321,8 @@ print_statfs (char *pformat, char m, char const *filename,
     case 't':
 #if HAVE_STRUCT_STATXFS_F_TYPE
       strcat (pformat, "lx");
-      printf (pformat, (long int) (statfsbuf->f_type));  /* no equiv. */
+      printf (pformat,
+	      (unsigned long int) (statfsbuf->f_type));  /* no equiv. */
 #else
       fputc ('*', stdout);
 #endif
@@ -396,8 +344,8 @@ print_statfs (char *pformat, char m, char const *filename,
       printf (pformat, (intmax_t) (statfsbuf->f_bavail));
       break;
     case 's':
-      strcat (pformat, "ld");
-      printf (pformat, (long int) (statfsbuf->f_bsize));
+      strcat (pformat, "lu");
+      printf (pformat, (unsigned long int) (statfsbuf->f_bsize));
       break;
     case 'c':
       strcat (pformat, PRIdMAX);
@@ -433,7 +381,7 @@ print_stat (char *pformat, char m, char const *filename, void const *data)
       strcat (pformat, "s");
       if (S_ISLNK (statbuf->st_mode))
 	{
-	  char *linkname = xreadlink (filename);
+	  char *linkname = xreadlink (filename, statbuf->st_size);
 	  if (linkname == NULL)
 	    {
 	      error (0, errno, _("cannot read symbolic link %s"),
@@ -451,40 +399,41 @@ print_stat (char *pformat, char m, char const *filename, void const *data)
 	}
       break;
     case 'd':
-      strcat (pformat, "d");
-      printf (pformat, (int) statbuf->st_dev);
+      strcat (pformat, PRIuMAX);
+      printf (pformat, (uintmax_t) statbuf->st_dev);
       break;
     case 'D':
-      strcat (pformat, "x");
-      printf (pformat, (int) statbuf->st_dev);
+      strcat (pformat, PRIxMAX);
+      printf (pformat, (uintmax_t) statbuf->st_dev);
       break;
     case 'i':
-      strcat (pformat, "d");
-      printf (pformat, (int) statbuf->st_ino);
+      strcat (pformat, PRIuMAX);
+      printf (pformat, (uintmax_t) statbuf->st_ino);
       break;
     case 'a':
-      strcat (pformat, "o");
-      printf (pformat, statbuf->st_mode & 07777);
+      strcat (pformat, "lo");
+      printf (pformat,
+	      (unsigned long int) (statbuf->st_mode & CHMOD_MODE_BITS));
       break;
     case 'A':
       strcat (pformat, "s");
       printf (pformat, human_access (statbuf));
       break;
     case 'f':
-      strcat (pformat, "x");
-      printf (pformat, statbuf->st_mode);
+      strcat (pformat, "lx");
+      printf (pformat, (unsigned long int) statbuf->st_mode);
       break;
     case 'F':
       strcat (pformat, "s");
       printf (pformat, file_type (statbuf));
       break;
     case 'h':
-      strcat (pformat, "d");
-      printf (pformat, (int) statbuf->st_nlink);
+      strcat (pformat, "lu");
+      printf (pformat, (unsigned long int) statbuf->st_nlink);
       break;
     case 'u':
-      strcat (pformat, "d");
-      printf (pformat, statbuf->st_uid);
+      strcat (pformat, "lu");
+      printf (pformat, (unsigned long int) statbuf->st_uid);
       break;
     case 'U':
       strcat (pformat, "s");
@@ -493,8 +442,8 @@ print_stat (char *pformat, char m, char const *filename, void const *data)
       printf (pformat, (pw_ent != 0L) ? pw_ent->pw_name : "UNKNOWN");
       break;
     case 'g':
-      strcat (pformat, "d");
-      printf (pformat, statbuf->st_gid);
+      strcat (pformat, "lu");
+      printf (pformat, (unsigned long int) statbuf->st_gid);
       break;
     case 'G':
       strcat (pformat, "s");
@@ -503,28 +452,28 @@ print_stat (char *pformat, char m, char const *filename, void const *data)
       printf (pformat, (gw_ent != 0L) ? gw_ent->gr_name : "UNKNOWN");
       break;
     case 't':
-      strcat (pformat, "x");
-      printf (pformat, major (statbuf->st_rdev));
+      strcat (pformat, "lx");
+      printf (pformat, (unsigned long int) major (statbuf->st_rdev));
       break;
     case 'T':
-      strcat (pformat, "x");
-      printf (pformat, minor (statbuf->st_rdev));
+      strcat (pformat, "lx");
+      printf (pformat, (unsigned long int) minor (statbuf->st_rdev));
       break;
     case 's':
       strcat (pformat, PRIuMAX);
       printf (pformat, (uintmax_t) (statbuf->st_size));
       break;
     case 'B':
-      strcat (pformat, "u");
-      printf (pformat, (unsigned int) ST_NBLOCKSIZE);
+      strcat (pformat, "lu");
+      printf (pformat, (unsigned long int) ST_NBLOCKSIZE);
       break;
     case 'b':
-      strcat (pformat, "u");
-      printf (pformat, (unsigned int) ST_NBLOCKS (*statbuf));
+      strcat (pformat, PRIuMAX);
+      printf (pformat, (uintmax_t) ST_NBLOCKS (*statbuf));
       break;
     case 'o':
-      strcat (pformat, "d");
-      printf (pformat, (int) statbuf->st_blksize);
+      strcat (pformat, "lu");
+      printf (pformat, (unsigned long int) statbuf->st_blksize);
       break;
     case 'x':
       strcat (pformat, "s");
@@ -532,8 +481,8 @@ print_stat (char *pformat, char m, char const *filename, void const *data)
 				   TIMESPEC_NS (statbuf->st_atim)));
       break;
     case 'X':
-      strcat (pformat, "d");
-      printf (pformat, (int) statbuf->st_atime);
+      strcat (pformat, TYPE_SIGNED (time_t) ? "ld" : "lu");
+      printf (pformat, (unsigned long int) statbuf->st_atime);
       break;
     case 'y':
       strcat (pformat, "s");
@@ -541,8 +490,8 @@ print_stat (char *pformat, char m, char const *filename, void const *data)
 				   TIMESPEC_NS (statbuf->st_mtim)));
       break;
     case 'Y':
-      strcat (pformat, "d");
-      printf (pformat, (int) statbuf->st_mtime);
+      strcat (pformat, TYPE_SIGNED (time_t) ? "ld" : "lu");
+      printf (pformat, (unsigned long int) statbuf->st_mtime);
       break;
     case 'z':
       strcat (pformat, "s");
@@ -550,8 +499,8 @@ print_stat (char *pformat, char m, char const *filename, void const *data)
 				   TIMESPEC_NS (statbuf->st_ctim)));
       break;
     case 'Z':
-      strcat (pformat, "d");
-      printf (pformat, (int) statbuf->st_ctime);
+      strcat (pformat, TYPE_SIGNED (time_t) ? "ld" : "lu");
+      printf (pformat, (unsigned long int) statbuf->st_ctime);
       break;
     default:
       strcat (pformat, "c");
@@ -610,63 +559,59 @@ print_it (char const *masterformat, char const *filename,
     }
   free (format);
   free (dest);
-  fputc ('\n', stdout);
 }
 
-/* stat the filesystem and print what we find */
-static void
-do_statfs (char const *filename, int terse, char const *format)
+/* Stat the file system and print what we find.  */
+static bool
+do_statfs (char const *filename, bool terse, char const *format)
 {
   STRUCT_STATVFS statfsbuf;
-  int i = statfs (filename, &statfsbuf);
 
-  if (i == -1)
+  if (STATFS (filename, &statfsbuf) != 0)
     {
       error (0, errno, _("cannot read file system information for %s"),
 	     quote (filename));
-      return;
+      return false;
     }
 
   if (format == NULL)
     {
       format = (terse
-		? "%n %i %l %t %b %f %a %s %c %d"
+		? "%n %i %l %t %b %f %a %s %c %d\n"
 		: "  File: \"%n\"\n"
 		"    ID: %-8i Namelen: %-7l Type: %T\n"
 		"Blocks: Total: %-10b Free: %-10f Available: %-10a Size: %s\n"
-		"Inodes: Total: %-10c Free: %-10d");
+		"Inodes: Total: %-10c Free: %-10d\n");
     }
 
   print_it (format, filename, print_statfs, &statfsbuf);
+  return true;
 }
 
 /* stat the file and print what we find */
-static void
-do_stat (char const *filename, int follow_links, int terse,
+static bool
+do_stat (char const *filename, bool follow_links, bool terse,
 	 char const *format)
 {
   struct stat statbuf;
-  int i = ((follow_links == 1)
-	   ? stat (filename, &statbuf)
-	   : lstat (filename, &statbuf));
 
-  if (i == -1)
+  if ((follow_links ? stat : lstat) (filename, &statbuf) != 0)
     {
       error (0, errno, _("cannot stat %s"), quote (filename));
-      return;
+      return false;
     }
 
   if (format == NULL)
     {
-      if (terse != 0)
+      if (terse)
 	{
-	  format = "%n %s %b %f %u %g %D %i %h %t %T %X %Y %Z %o";
+	  format = "%n %s %b %f %u %g %D %i %h %t %T %X %Y %Z %o\n";
 	}
       else
 	{
-	  /* tmp hack to match orignal output until conditional implemented */
-	  i = statbuf.st_mode & S_IFMT;
-	  if (i == S_IFCHR || i == S_IFBLK)
+	  /* Temporary hack to match original output until conditional
+	     implemented.  */
+	  if (S_ISBLK (statbuf.st_mode) || S_ISCHR (statbuf.st_mode))
 	    {
 	      format =
 		"  File: %N\n"
@@ -674,7 +619,7 @@ do_stat (char const *filename, int follow_links, int terse,
 		"Device: %Dh/%dd\tInode: %-10i  Links: %-5h"
 		" Device type: %t,%T\n"
 		"Access: (%04a/%10.10A)  Uid: (%5u/%8U)   Gid: (%5g/%8G)\n"
-		"Access: %x\n" "Modify: %y\n" "Change: %z";
+		"Access: %x\n" "Modify: %y\n" "Change: %z\n";
 	    }
 	  else
 	    {
@@ -683,11 +628,12 @@ do_stat (char const *filename, int follow_links, int terse,
 		"  Size: %-10s\tBlocks: %-10b IO Block: %-6o %F\n"
 		"Device: %Dh/%dd\tInode: %-10i  Links: %h\n"
 		"Access: (%04a/%10.10A)  Uid: (%5u/%8U)   Gid: (%5g/%8G)\n"
-		"Access: %x\n" "Modify: %y\n" "Change: %z";
+		"Access: %x\n" "Modify: %y\n" "Change: %z\n";
 	    }
 	}
     }
   print_it (format, filename, print_stat, &statbuf);
+  return true;
 }
 
 void
@@ -700,9 +646,9 @@ usage (int status)
     {
       printf (_("Usage: %s [OPTION] FILE...\n"), program_name);
       fputs (_("\
-Display file or filesystem status.\n\
+Display file or file system status.\n\
 \n\
-  -f, --filesystem      display filesystem status instead of file status\n\
+  -f, --file-system     display file system status instead of file status\n\
   -c  --format=FORMAT   use the specified FORMAT instead of the default\n\
   -L, --dereference     follow links\n\
   -t, --terse           print the information in terse form\n\
@@ -711,7 +657,7 @@ Display file or filesystem status.\n\
       fputs (VERSION_OPTION_DESCRIPTION, stdout);
 
       fputs (_("\n\
-The valid format sequences for files (without --filesystem):\n\
+The valid format sequences for files (without --file-system):\n\
 \n\
   %A   Access rights in human readable form\n\
   %a   Access rights in octal\n\
@@ -775,10 +721,11 @@ main (int argc, char *argv[])
 {
   int c;
   int i;
-  int follow_links = 0;
-  int fs = 0;
-  int terse = 0;
+  bool follow_links = false;
+  bool fs = false;
+  bool terse = false;
   char *format = NULL;
+  bool ok = true;
 
   initialize_main (&argc, &argv);
   program_name = argv[0];
@@ -800,15 +747,15 @@ main (int argc, char *argv[])
 	  error (0, 0, _("Warning: `-l' is deprecated; use `-L' instead"));
 	  /* fall through */
 	case 'L':
-	  follow_links = 1;
+	  follow_links = true;
 	  break;
 
 	case 'f':
-	  fs = 1;
+	  fs = true;
 	  break;
 
 	case 't':
-	  terse = 1;
+	  terse = true;
 	  break;
 
 	case_GETOPT_HELP_CHAR;
@@ -822,17 +769,14 @@ main (int argc, char *argv[])
 
   if (argc == optind)
     {
-      error (0, 0, _("too few arguments"));
+      error (0, 0, _("missing operand"));
       usage (EXIT_FAILURE);
     }
 
   for (i = optind; i < argc; i++)
-    {
-      if (fs == 0)
-	do_stat (argv[i], follow_links, terse, format);
-      else
-	do_statfs (argv[i], terse, format);
-    }
+    ok &= (fs
+	   ? do_statfs (argv[i], terse, format)
+	   : do_stat (argv[i], follow_links, terse, format));
 
-  exit (G_fail ? EXIT_FAILURE : EXIT_SUCCESS);
+  exit (ok ? EXIT_SUCCESS : EXIT_FAILURE);
 }
