@@ -1,5 +1,5 @@
 /* stty -- change and print terminal line settings
-   Copyright (C) 1990-2003 Free Software Foundation, Inc.
+   Copyright (C) 1990-2004 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -53,13 +53,11 @@
 #endif
 #include <getopt.h>
 #include <stdarg.h>
-#define VA_START(args, lastarg) va_start(args, lastarg)
 
 #include "system.h"
-#include "long-options.h"
 #include "error.h"
+#include "vasprintf.h"
 #include "xstrtol.h"
-#include "closeout.h"
 
 /* The official name of this program (e.g., no `g' prefix).  */
 #define PROGRAM_NAME "stty"
@@ -67,7 +65,7 @@
 #define AUTHORS "David MacKenzie"
 
 #ifndef _POSIX_VDISABLE
-# define _POSIX_VDISABLE ((unsigned char) 0)
+# define _POSIX_VDISABLE 0
 #endif
 
 #define Control(c) ((c) & 0x1f)
@@ -233,6 +231,9 @@ static struct mode_info mode_info[] =
 #ifdef IMAXBEL
   {"imaxbel", input, SANE_SET | REV, IMAXBEL, 0},
 #endif
+#ifdef IUTF8
+  {"iutf8", input, SANE_UNSET | REV, IUTF8, 0},
+#endif
 
   {"opost", output, SANE_SET | REV, OPOST, 0},
 #ifdef OLCUC
@@ -350,8 +351,8 @@ static struct mode_info mode_info[] =
 struct control_info
   {
     const char *name;		/* Name given on command line.  */
-    unsigned char saneval;	/* Value to set for `stty sane'.  */
-    int offset;			/* Offset in c_cc.  */
+    cc_t saneval;		/* Value to set for `stty sane'.  */
+    size_t offset;		/* Offset in c_cc.  */
   };
 
 /* Control characters. */
@@ -402,13 +403,13 @@ static struct control_info control_info[] =
   {NULL, 0, 0}
 };
 
-static const char *visible (unsigned int ch);
-static unsigned long baud_to_value (speed_t speed);
-static int recover_mode (char *arg, struct termios *mode);
+static char const *visible (cc_t ch);
+static unsigned long int baud_to_value (speed_t speed);
+static bool recover_mode (char const *arg, struct termios *mode);
 static int screen_columns (void);
-static int set_mode (struct mode_info *info, int reversed,
-			     struct termios *mode);
-static long integer_arg (const char *s);
+static bool set_mode (struct mode_info *info, bool reversed,
+		      struct termios *mode);
+static unsigned long int integer_arg (const char *s, unsigned long int max);
 static speed_t string_to_baud (const char *arg);
 static tcflag_t *mode_type_flag (enum mode_type type, struct termios *mode);
 static void display_all (struct termios *mode, int fd, const char *device_name);
@@ -417,8 +418,8 @@ static void display_recoverable (struct termios *mode);
 static void display_settings (enum output_type output_type,
 			      struct termios *mode, int fd,
 			      const char *device_name);
-static void display_speed (struct termios *mode, int fancy);
-static void display_window_size (int fancy, int fd, const char *device_name);
+static void display_speed (struct termios *mode, bool fancy);
+static void display_window_size (bool fancy, int fd, const char *device_name);
 static void sane_mode (struct termios *mode);
 static void set_control_char (struct control_info *info,
 			      const char *arg,
@@ -439,6 +440,8 @@ static struct option longopts[] =
   {"all", no_argument, NULL, 'a'},
   {"save", no_argument, NULL, 'g'},
   {"file", required_argument, NULL, 'F'},
+  {GETOPT_HELP_OPTION_DECL},
+  {GETOPT_VERSION_OPTION_DECL},
   {NULL, 0, NULL, 0}
 };
 
@@ -452,37 +455,43 @@ static void wrapf (const char *message, ...)
    Wrap to next line first if it won't fit.
    Print a space first unless MESSAGE will start a new line. */
 
-/* VARARGS */
 static void
 wrapf (const char *message,...)
 {
   va_list args;
-  char buf[1024];		/* Plenty long for our needs. */
+  char *buf;
   int buflen;
 
-  VA_START (args, message);
-  vsprintf (buf, message, args);
+  va_start (args, message);
+  buflen = vasprintf (&buf, message, args);
   va_end (args);
-  buflen = strlen (buf);
-  if (0 < current_col
-      && max_col <= current_col + (0 < current_col) + buflen)
-    {
-      putchar ('\n');
-      current_col = 0;
-    }
+
+  if (buflen < 0)
+    xalloc_die ();
+
   if (0 < current_col)
     {
-      putchar (' ');
-      current_col++;
+      if (max_col - current_col < buflen)
+	{
+	  putchar ('\n');
+	  current_col = 0;
+	}
+      else
+	{
+	  putchar (' ');
+	  current_col++;
+	}
     }
+
   fputs (buf, stdout);
+  free (buf);
   current_col += buflen;
 }
 
 void
 usage (int status)
 {
-  if (status != 0)
+  if (status != EXIT_SUCCESS)
     fprintf (stderr, _("Try `%s --help' for more information.\n"),
 	     program_name);
   else
@@ -580,6 +589,9 @@ Input settings:\n\
    [-]inlcr      translate newline to carriage return\n\
    [-]inpck      enable input parity checking\n\
    [-]istrip     clear high (8th) bit of input characters\n\
+"), stdout);
+      fputs (_("\
+ * [-]iutf8      assume input characters are UTF-8 encoded\n\
 "), stdout);
       fputs (_("\
  * [-]iuclc      translate uppercase characters to lowercase\n\
@@ -683,7 +695,7 @@ Combination settings:\n\
    -raw          same as cooked\n\
 "), stdout);
       fputs (_("\
-   sane          same as cread -ignbrk brkint -inlcr -igncr icrnl\n\
+   sane          same as cread -ignbrk brkint -inlcr -igncr icrnl -iutf8\n\
                  -ixoff -iuclc -ixany imaxbel opost -olcuc -ocrnl onlcr\n\
                  -onocr -onlret -ofill -ofdel nl0 cr0 tab0 bs0 vt0 ff0\n\
                  isig icanon iexten echo echoe echok -echonl -noflsh\n\
@@ -702,45 +714,23 @@ settings, CHAR is taken literally, or coded as in ^c, 0x37, 0177 or\n\
   exit (status);
 }
 
-/* Return 1 if the string contains only valid options.  */
-static int
-valid_options (char *opt, const char *valid_opts,
-	       const char *valid_arg_opts)
-{
-  char ch;
-
-  if (*opt++ != '-' || *opt == 0)
-    return 0;
-
-  while ((ch = *opt))
-    {
-      opt++;
-      if (strchr (valid_opts, ch))
-	continue;
-      if (strchr (valid_arg_opts, ch))
-	return 1;
-      return 0;
-    }
-  return 1;
-}
-
 int
 main (int argc, char **argv)
 {
   struct termios mode;
   enum output_type output_type;
   int optc;
-  int require_set_attr;
-  int speed_was_set;
-  int verbose_output;
-  int recoverable_output;
+  int argi = 0;
+  int opti = 1;
+  bool require_set_attr;
+  bool speed_was_set;
+  bool verbose_output;
+  bool recoverable_output;
   int k;
-  int noargs = 1;
+  bool noargs = true;
   char *file_name = NULL;
   int fd;
   const char *device_name;
-  const char *posixly_correct = getenv ("POSIXLY_CORRECT");
-  int invalid_long_option = 0;
 
   initialize_main (&argc, &argv);
   program_name = argv[0];
@@ -750,134 +740,75 @@ main (int argc, char **argv)
 
   atexit (close_stdout);
 
-  parse_long_options (argc, argv, PROGRAM_NAME, GNU_PACKAGE, VERSION,
-		      AUTHORS, usage);
-
   output_type = changed;
-  verbose_output = 0;
-  recoverable_output = 0;
+  verbose_output = false;
+  recoverable_output = false;
 
   /* Don't print error messages for unrecognized options.  */
   opterr = 0;
 
-  while ((optc = getopt_long (argc, argv, "agF:", longopts, NULL)) != -1)
+  /* If any new options are ever added to stty, the short options MUST
+     NOT allow any ambiguity with the stty settings.  For example, the
+     stty setting "-gagFork" would not be feasible, since it will be
+     parsed as "-g -a -g -F ork".  If you change anything about how
+     stty parses options, be sure it still works with combinations of
+     short and long options, --, POSIXLY_CORRECT, etc.  */
+
+  while ((optc = getopt_long (argc - argi, argv + argi, "-agF:",
+			      longopts, NULL))
+	 != -1)
     {
-      int unrecognized_option = 0;
       switch (optc)
 	{
 	case 'a':
-	  verbose_output = 1;
+	  verbose_output = true;
 	  output_type = all;
 	  break;
 
 	case 'g':
-	  recoverable_output = 1;
+	  recoverable_output = true;
 	  output_type = recoverable;
 	  break;
 
 	case 'F':
 	  if (file_name)
-	    error (2, 0, _("only one device may be specified"));
+	    error (EXIT_FAILURE, 0, _("only one device may be specified"));
 	  file_name = optarg;
 	  break;
 
+	case_GETOPT_HELP_CHAR;
+
+	case_GETOPT_VERSION_CHAR (PROGRAM_NAME, AUTHORS);
+
 	default:
-	  unrecognized_option = 1;
+	  noargs = false;
+
+	  /* Skip the argument containing this unrecognized option;
+	     the 2nd pass will analyze it.  */
+	  argi += opti;
+
+	  /* Restart getopt_long from the first unskipped argument.  */
+	  opti = 1;
+	  optind = 0;
+
 	  break;
 	}
 
-      if (unrecognized_option)
-	break;
-    }
-
-  /* FIXME: what is this?!? */
-  if (invalid_long_option)
-    usage (EXIT_FAILURE);
-
-  /* Clear out the options that have been parsed.  This is really
-     gross, but it's needed because stty SETTINGS look like options to
-     getopt(), so we need to work around things in a really horrible
-     way.  If any new options are ever added to stty, the short option
-     MUST NOT be a letter which is the first letter of one of the
-     possible stty settings.  If you change anything about how stty
-     parses options, be sure it still works with combinations of
-     short and long options, --, POSIXLY_CORRECT, etc.  */
-  for (k = 1; k < argc; k++)
-    {
-      size_t len;
-      char *eq;
-
-      if (argv[k] == NULL)
-	continue;
-
-      /* Handle --, and reset noargs if there are arguments following it.  */
-      if (STREQ (argv[k], "--"))
-	{
-	  argv[k] = NULL;
-	  if (k < argc - 1)
-	    noargs = 0;
-	  break;
-	}
-
-      /* Handle "--file device" */
-      len = strlen (argv[k]);
-      if (len >= 3 && strstr ("--file", argv[k]))
-	{
-	  argv[k] = NULL;
-	  argv[k + 1] = NULL;
-	  continue;
-	}
-
-      /* Handle "--all" and "--save".  */
-      if (len >= 3
-	  && (strstr ("--all", argv[k])
-	      || strstr ("--save", argv[k])))
-	{
-	  argv[k] = NULL;
-	  continue;
-	}
-
-      /* Handle "--file=device".  */
-      eq = strchr (argv[k], '=');
-      if (eq && eq - argv[k] >= 3)
-	{
-	  *eq = '\0';
-	  if (strstr ("--file", argv[k]))
-	    {
-	      argv[k] = NULL;
-	      continue;
-	    }
-	  /* Put the equals sign back for the error message.  */
-	  *eq = '=';
-	}
-
-      /* Handle "-a", "-ag", "-aF/dev/foo", "-aF /dev/foo", etc.  */
-      if (valid_options (argv[k], "ag", "F"))
-	{
-	  /* FIXME: this loses when the device name ends in `F'.
-	     e.g. `stty -F/dev/BARF nl' would clobber the `nl' argument.  */
-	  if (argv[k][strlen (argv[k]) - 1] == 'F')
-	    argv[k + 1] = NULL;
-	  argv[k] = NULL;
-        }
-      /* Everything else must be a normal, non-option argument.  */
-      else
-	{
-	  noargs = 0;
-	  if (posixly_correct)
-	    break;
-        }
+      /* Clear fully-parsed arguments, so they don't confuse the 2nd pass.  */
+      while (opti < optind)
+	argv[argi + opti++] = NULL;
     }
 
   /* Specifying both -a and -g gets an error.  */
-  if (verbose_output && recoverable_output)
-    error (2, 0,
-	   _("the options for verbose and stty-readable output styles are\n\
-mutually exclusive"));
+  if (verbose_output & recoverable_output)
+    error (EXIT_FAILURE, 0,
+	   _("the options for verbose and stty-readable output styles are\n"
+	     "mutually exclusive"));
 
   /* Specifying any other arguments with -a or -g gets an error.  */
-  if (!noargs && (verbose_output || recoverable_output))
-    error (2, 0, _("when specifying an output style, modes may not be set"));
+  if (!noargs & (verbose_output | recoverable_output))
+    error (EXIT_FAILURE, 0,
+	   _("when specifying an output style, modes may not be set"));
 
   /* FIXME: it'd be better not to open the file until we've verified
      that all arguments are valid.  Otherwise, we could end up doing
@@ -898,7 +829,7 @@ mutually exclusive"));
     }
   else
     {
-      fd = 0;
+      fd = STDIN_FILENO;
       device_name = _("standard input");
     }
 
@@ -908,7 +839,7 @@ mutually exclusive"));
   if (tcgetattr (fd, &mode))
     error (EXIT_FAILURE, errno, "%s", device_name);
 
-  if (verbose_output || recoverable_output || noargs)
+  if (verbose_output | recoverable_output | noargs)
     {
       max_col = screen_columns ();
       current_col = 0;
@@ -916,151 +847,150 @@ mutually exclusive"));
       exit (EXIT_SUCCESS);
     }
 
-  speed_was_set = 0;
-  require_set_attr = 0;
-  k = 1;
-  while (k < argc)
+  speed_was_set = false;
+  require_set_attr = false;
+  for (k = 1; k < argc; k++)
     {
-      int match_found = 0;
-      int reversed = 0;
+      char const *arg = argv[k];
+      bool match_found = false;
+      bool reversed = false;
       int i;
 
-      if (argv[k] == 0)
-	{
-	  k++;
-	  continue;
-	}
+      if (! arg)
+	continue;
 
-      if (argv[k][0] == '-')
+      if (arg[0] == '-')
 	{
-	  ++argv[k];
-	  reversed = 1;
+	  ++arg;
+	  reversed = true;
 	}
       for (i = 0; mode_info[i].name != NULL; ++i)
 	{
-	  if (STREQ (argv[k], mode_info[i].name))
+	  if (STREQ (arg, mode_info[i].name))
 	    {
 	      match_found = set_mode (&mode_info[i], reversed, &mode);
-	      require_set_attr = 1;
+	      require_set_attr = true;
 	      break;
 	    }
 	}
-      if (match_found == 0 && reversed)
+      if (!match_found & reversed)
 	{
-	  error (0, 0, _("invalid argument `%s'"), --argv[k]);
+	  error (0, 0, _("invalid argument `%s'"), arg - 1);
 	  usage (EXIT_FAILURE);
 	}
-      if (match_found == 0)
+      if (!match_found)
 	{
 	  for (i = 0; control_info[i].name != NULL; ++i)
 	    {
-	      if (STREQ (argv[k], control_info[i].name))
+	      if (STREQ (arg, control_info[i].name))
 		{
 		  if (k == argc - 1)
 		    {
-		      error (0, 0, _("missing argument to `%s'"), argv[k]);
+		      error (0, 0, _("missing argument to `%s'"), arg);
 		      usage (EXIT_FAILURE);
 		    }
-		  match_found = 1;
+		  match_found = true;
 		  ++k;
 		  set_control_char (&control_info[i], argv[k], &mode);
-		  require_set_attr = 1;
+		  require_set_attr = true;
 		  break;
 		}
 	    }
 	}
-      if (match_found == 0)
+      if (!match_found)
 	{
-	  if (STREQ (argv[k], "ispeed"))
+	  if (STREQ (arg, "ispeed"))
 	    {
 	      if (k == argc - 1)
 		{
-		  error (0, 0, _("missing argument to `%s'"), argv[k]);
+		  error (0, 0, _("missing argument to `%s'"), arg);
 		  usage (EXIT_FAILURE);
 		}
 	      ++k;
 	      set_speed (input_speed, argv[k], &mode);
-	      speed_was_set = 1;
-	      require_set_attr = 1;
+	      speed_was_set = true;
+	      require_set_attr = true;
 	    }
-	  else if (STREQ (argv[k], "ospeed"))
+	  else if (STREQ (arg, "ospeed"))
 	    {
 	      if (k == argc - 1)
 		{
-		  error (0, 0, _("missing argument to `%s'"), argv[k]);
+		  error (0, 0, _("missing argument to `%s'"), arg);
 		  usage (EXIT_FAILURE);
 		}
 	      ++k;
 	      set_speed (output_speed, argv[k], &mode);
-	      speed_was_set = 1;
-	      require_set_attr = 1;
+	      speed_was_set = true;
+	      require_set_attr = true;
 	    }
 #ifdef TIOCGWINSZ
-	  else if (STREQ (argv[k], "rows"))
+	  else if (STREQ (arg, "rows"))
 	    {
 	      if (k == argc - 1)
 		{
-		  error (0, 0, _("missing argument to `%s'"), argv[k]);
+		  error (0, 0, _("missing argument to `%s'"), arg);
 		  usage (EXIT_FAILURE);
 		}
 	      ++k;
-	      set_window_size ((int) integer_arg (argv[k]), -1,
+	      set_window_size (integer_arg (argv[k], INT_MAX), -1,
 			       fd, device_name);
 	    }
-	  else if (STREQ (argv[k], "cols")
-		   || STREQ (argv[k], "columns"))
+	  else if (STREQ (arg, "cols")
+		   || STREQ (arg, "columns"))
 	    {
 	      if (k == argc - 1)
 		{
-		  error (0, 0, _("missing argument to `%s'"), argv[k]);
+		  error (0, 0, _("missing argument to `%s'"), arg);
 		  usage (EXIT_FAILURE);
 		}
 	      ++k;
-	      set_window_size (-1, (int) integer_arg (argv[k]),
+	      set_window_size (-1, integer_arg (argv[k], INT_MAX),
 			       fd, device_name);
 	    }
-	  else if (STREQ (argv[k], "size"))
+	  else if (STREQ (arg, "size"))
 	    {
 	      max_col = screen_columns ();
 	      current_col = 0;
-	      display_window_size (0, fd, device_name);
+	      display_window_size (false, fd, device_name);
 	    }
 #endif
 #ifdef HAVE_C_LINE
-	  else if (STREQ (argv[k], "line"))
+	  else if (STREQ (arg, "line"))
 	    {
+	      unsigned long int value;
 	      if (k == argc - 1)
 		{
-		  error (0, 0, _("missing argument to `%s'"), argv[k]);
+		  error (0, 0, _("missing argument to `%s'"), arg);
 		  usage (EXIT_FAILURE);
 		}
 	      ++k;
-	      mode.c_line = integer_arg (argv[k]);
-	      require_set_attr = 1;
+	      mode.c_line = value = integer_arg (argv[k], ULONG_MAX);
+	      if (mode.c_line != value)
+		error (0, 0, _("invalid line discipline `%s'"), argv[k]);
+	      require_set_attr = true;
 	    }
 #endif
-	  else if (STREQ (argv[k], "speed"))
+	  else if (STREQ (arg, "speed"))
 	    {
 	      max_col = screen_columns ();
-	      display_speed (&mode, 0);
+	      display_speed (&mode, false);
 	    }
-	  else if (string_to_baud (argv[k]) != (speed_t) -1)
+	  else if (string_to_baud (arg) != (speed_t) -1)
 	    {
-	      set_speed (both_speeds, argv[k], &mode);
-	      speed_was_set = 1;
-	      require_set_attr = 1;
+	      set_speed (both_speeds, arg, &mode);
+	      speed_was_set = true;
+	      require_set_attr = true;
 	    }
 	  else
 	    {
-	      if (recover_mode (argv[k], &mode) == 0)
+	      if (! recover_mode (arg, &mode))
 		{
-		  error (0, 0, _("invalid argument `%s'"), argv[k]);
+		  error (0, 0, _("invalid argument `%s'"), arg);
 		  usage (EXIT_FAILURE);
 		}
-	      require_set_attr = 1;
+	      require_set_attr = true;
 	    }
 	}
-      k++;
     }
 
   if (require_set_attr)
@@ -1126,15 +1056,16 @@ mutually exclusive"));
   exit (EXIT_SUCCESS);
 }
 
-/* Return 0 if not applied because not reversible; otherwise return 1.  */
+/* Return false if not applied because not reversible; otherwise
+   return true.  */
 
-static int
-set_mode (struct mode_info *info, int reversed, struct termios *mode)
+static bool
+set_mode (struct mode_info *info, bool reversed, struct termios *mode)
 {
   tcflag_t *bitsp;
 
   if (reversed && (info->flags & REV) == 0)
-    return 0;
+    return false;
 
   bitsp = mode_type_flag (info->type, mode);
 
@@ -1331,19 +1262,19 @@ set_mode (struct mode_info *info, int reversed, struct termios *mode)
   else
     *bitsp = (*bitsp & ~info->mask) | info->bits;
 
-  return 1;
+  return true;
 }
 
 static void
 set_control_char (struct control_info *info, const char *arg,
 		  struct termios *mode)
 {
-  unsigned char value;
+  unsigned long int value;
 
   if (STREQ (info->name, "min") || STREQ (info->name, "time"))
-    value = integer_arg (arg);
+    value = integer_arg (arg, TYPE_MAXIMUM (cc_t));
   else if (arg[0] == '\0' || arg[1] == '\0')
-    value = arg[0];
+    value = to_uchar (arg[0]);
   else if (STREQ (arg, "^-") || STREQ (arg, "undef"))
     value = _POSIX_VDISABLE;
   else if (arg[0] == '^' && arg[1] != '\0')	/* Ignore any trailing junk. */
@@ -1351,10 +1282,10 @@ set_control_char (struct control_info *info, const char *arg,
       if (arg[1] == '?')
 	value = 127;
       else
-	value = arg[1] & ~0140;	/* Non-letters get weird results. */
+	value = to_uchar (arg[1]) & ~0140; /* Non-letters get weird results. */
     }
   else
-    value = integer_arg (arg);
+    value = integer_arg (arg, TYPE_MAXIMUM (cc_t));
   mode->c_cc[info->offset] = value;
 }
 
@@ -1442,7 +1373,7 @@ set_window_size (int rows, int cols, int fd, const char *device_name)
 }
 
 static void
-display_window_size (int fancy, int fd, const char *device_name)
+display_window_size (bool fancy, int fd, const char *device_name)
 {
   struct winsize win;
 
@@ -1482,11 +1413,11 @@ screen_columns (void)
   {
     /* Use $COLUMNS if it's in [1..INT_MAX-1].  */
     char *col_string = getenv ("COLUMNS");
-    long n_columns;
+    long int n_columns;
     if (!(col_string != NULL
 	  && xstrtol (col_string, NULL, 0, &n_columns, "") == LONGINT_OK
 	  && 0 < n_columns
-	  && n_columns < INT_MAX))
+	  && n_columns <= INT_MAX))
       n_columns = 80;
     return n_columns;
   }
@@ -1541,19 +1472,19 @@ static void
 display_changed (struct termios *mode)
 {
   int i;
-  int empty_line;
+  bool empty_line;
   tcflag_t *bitsp;
   unsigned long mask;
   enum mode_type prev_type = control;
 
-  display_speed (mode, 1);
+  display_speed (mode, true);
 #ifdef HAVE_C_LINE
   wrapf ("line = %d;", mode->c_line);
 #endif
   putchar ('\n');
   current_col = 0;
 
-  empty_line = 1;
+  empty_line = true;
   for (i = 0; !STREQ (control_info[i].name, "min"); ++i)
     {
       if (mode->c_cc[control_info[i].offset] == control_info[i].saneval)
@@ -1571,31 +1502,32 @@ display_changed (struct termios *mode)
 	continue;
 #endif
 
-      empty_line = 0;
+      empty_line = false;
       wrapf ("%s = %s;", control_info[i].name,
 	     visible (mode->c_cc[control_info[i].offset]));
     }
   if ((mode->c_lflag & ICANON) == 0)
     {
-      wrapf ("min = %d; time = %d;\n", (int) mode->c_cc[VMIN],
-	     (int) mode->c_cc[VTIME]);
+      wrapf ("min = %lu; time = %lu;\n",
+	     (unsigned long int) mode->c_cc[VMIN],
+	     (unsigned long int) mode->c_cc[VTIME]);
     }
-  else if (empty_line == 0)
+  else if (!empty_line)
     putchar ('\n');
   current_col = 0;
 
-  empty_line = 1;
+  empty_line = true;
   for (i = 0; mode_info[i].name != NULL; ++i)
     {
       if (mode_info[i].flags & OMIT)
 	continue;
       if (mode_info[i].type != prev_type)
 	{
-	  if (empty_line == 0)
+	  if (!empty_line)
 	    {
 	      putchar ('\n');
 	      current_col = 0;
-	      empty_line = 1;
+	      empty_line = true;
 	    }
 	  prev_type = mode_info[i].type;
 	}
@@ -1607,16 +1539,16 @@ display_changed (struct termios *mode)
 	  if (mode_info[i].flags & SANE_UNSET)
 	    {
 	      wrapf ("%s", mode_info[i].name);
-	      empty_line = 0;
+	      empty_line = false;
 	    }
 	}
       else if ((mode_info[i].flags & (SANE_SET | REV)) == (SANE_SET | REV))
 	{
 	  wrapf ("-%s", mode_info[i].name);
-	  empty_line = 0;
+	  empty_line = false;
 	}
     }
-  if (empty_line == 0)
+  if (!empty_line)
     putchar ('\n');
   current_col = 0;
 }
@@ -1629,9 +1561,9 @@ display_all (struct termios *mode, int fd, const char *device_name)
   unsigned long mask;
   enum mode_type prev_type = control;
 
-  display_speed (mode, 1);
+  display_speed (mode, true);
 #ifdef TIOCGWINSZ
-  display_window_size (1, fd, device_name);
+  display_window_size (true, fd, device_name);
 #endif
 #ifdef HAVE_C_LINE
   wrapf ("line = %d;", mode->c_line);
@@ -1659,7 +1591,9 @@ display_all (struct termios *mode, int fd, const char *device_name)
 #if VEOF == VMIN
   if ((mode->c_lflag & ICANON) == 0)
 #endif
-    wrapf ("min = %d; time = %d;", mode->c_cc[VMIN], mode->c_cc[VTIME]);
+    wrapf ("min = %lu; time = %lu;",
+	   (unsigned long int) mode->c_cc[VMIN],
+	   (unsigned long int) mode->c_cc[VTIME]);
   if (current_col != 0)
     putchar ('\n');
   current_col = 0;
@@ -1687,7 +1621,7 @@ display_all (struct termios *mode, int fd, const char *device_name)
 }
 
 static void
-display_speed (struct termios *mode, int fancy)
+display_speed (struct termios *mode, bool fancy)
 {
   if (cfgetispeed (mode) == 0 || cfgetispeed (mode) == cfgetospeed (mode))
     wrapf (fancy ? "speed %lu baud;" : "%lu\n",
@@ -1703,53 +1637,63 @@ display_speed (struct termios *mode, int fancy)
 static void
 display_recoverable (struct termios *mode)
 {
-  int i;
+  size_t i;
 
   printf ("%lx:%lx:%lx:%lx",
-	  (unsigned long) mode->c_iflag, (unsigned long) mode->c_oflag,
-	  (unsigned long) mode->c_cflag, (unsigned long) mode->c_lflag);
+	  (unsigned long int) mode->c_iflag,
+	  (unsigned long int) mode->c_oflag,
+	  (unsigned long int) mode->c_cflag,
+	  (unsigned long int) mode->c_lflag);
   for (i = 0; i < NCCS; ++i)
-    printf (":%x", (unsigned int) mode->c_cc[i]);
+    printf (":%lx", (unsigned long int) mode->c_cc[i]);
   putchar ('\n');
 }
 
-static int
-recover_mode (char *arg, struct termios *mode)
+static bool
+recover_mode (char const *arg, struct termios *mode)
 {
-  int i, n;
-  unsigned int chr;
-  unsigned long iflag, oflag, cflag, lflag;
+  size_t i;
+  int n;
+  unsigned long int chr;
+  unsigned long int iflag, oflag, cflag, lflag;
 
   /* Scan into temporaries since it is too much trouble to figure out
      the right format for `tcflag_t'.  */
   if (sscanf (arg, "%lx:%lx:%lx:%lx%n",
 	      &iflag, &oflag, &cflag, &lflag, &n) != 4)
-    return 0;
+    return false;
   mode->c_iflag = iflag;
   mode->c_oflag = oflag;
   mode->c_cflag = cflag;
   mode->c_lflag = lflag;
+  if (mode->c_iflag != iflag
+      || mode->c_oflag != oflag
+      || mode->c_cflag != cflag
+      || mode->c_lflag != lflag)
+    return false;
   arg += n;
   for (i = 0; i < NCCS; ++i)
     {
-      if (sscanf (arg, ":%x%n", &chr, &n) != 1)
-	return 0;
+      if (sscanf (arg, ":%lx%n", &chr, &n) != 1)
+	return false;
       mode->c_cc[i] = chr;
+      if (mode->c_cc[i] != chr)
+	return false;
       arg += n;
     }
 
   /* Fail if there are too many fields.  */
   if (*arg != '\0')
-    return 0;
+    return false;
 
-  return 1;
+  return true;
 }
 
 struct speed_map
 {
   const char *string;		/* ASCII representation. */
   speed_t speed;		/* Internal form. */
-  unsigned long value;		/* Numeric value. */
+  unsigned long int value;	/* Numeric value. */
 };
 
 struct speed_map speeds[] =
@@ -1832,7 +1776,7 @@ string_to_baud (const char *arg)
   return (speed_t) -1;
 }
 
-static unsigned long
+static unsigned long int
 baud_to_value (speed_t speed)
 {
   int i;
@@ -1877,7 +1821,7 @@ sane_mode (struct termios *mode)
 /* Adapted from `cat' by Torbjorn Granlund.  */
 
 static const char *
-visible (unsigned int ch)
+visible (cc_t ch)
 {
   static char buf[10];
   char *bpout = buf;
@@ -1925,14 +1869,15 @@ visible (unsigned int ch)
 }
 
 /* Parse string S as an integer, using decimal radix by default,
-   but allowing octal and hex numbers as in C.  */
-/* From `od' by Richard Stallman.  */
+   but allowing octal and hex numbers as in C.  Reject values
+   larger than MAXVAL.  */
 
-static long
-integer_arg (const char *s)
+static unsigned long int
+integer_arg (const char *s, unsigned long int maxval)
 {
-  long value;
-  if (xstrtol (s, NULL, 0, &value, "bB") != LONGINT_OK)
+  unsigned long int value;
+  if (xstrtoul (s, NULL, 0, &value, "bB") != LONGINT_OK
+      || maxval < value)
     {
       error (0, 0, _("invalid integer argument `%s'"), s);
       usage (EXIT_FAILURE);
