@@ -1,5 +1,5 @@
 /* uniq -- remove duplicate lines from a sorted file
-   Copyright (C) 86, 91, 1995-2003, Free Software Foundation, Inc.
+   Copyright (C) 86, 91, 1995-2004 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -29,6 +29,7 @@
 #include "error.h"
 #include "hard-locale.h"
 #include "posixver.h"
+#include "quote.h"
 #include "xmemcoll.h"
 #include "xstrtol.h"
 #include "memcasecmp.h"
@@ -51,8 +52,8 @@
 /* The name this program was run with. */
 char *program_name;
 
-/* Nonzero if the LC_COLLATE locale is hard.  */
-static int hard_LC_COLLATE;
+/* True if the LC_COLLATE locale is hard.  */
+static bool hard_LC_COLLATE;
 
 /* Number of fields to skip on each line when doing comparisons. */
 static size_t skip_fields;
@@ -80,8 +81,8 @@ static bool output_unique;
 static bool output_first_repeated;
 static bool output_later_repeated;
 
-/* If nonzero, ignore case when comparing.  */
-static int ignore_case;
+/* If true, ignore case when comparing.  */
+static bool ignore_case;
 
 enum delimit_method
 {
@@ -126,7 +127,7 @@ static struct option const longopts[] =
 void
 usage (int status)
 {
-  if (status != 0)
+  if (status != EXIT_SUCCESS)
     fprintf (stderr, _("Try `%s --help' for more information.\n"),
 	     program_name);
   else
@@ -168,7 +169,7 @@ Fields are skipped before chars.\n\
 "), stdout);
       printf (_("\nReport bugs to <%s>.\n"), PACKAGE_BUGREPORT);
     }
-  exit (status == 0 ? EXIT_SUCCESS : EXIT_FAILURE);
+  exit (status);
 }
 
 /* Convert OPT to size_t, reporting an error using MSGID if it does
@@ -209,12 +210,12 @@ find_field (const struct linebuffer *line)
   return lp + i;
 }
 
-/* Return zero if two strings OLD and NEW match, nonzero if not.
+/* Return false if two strings OLD and NEW match, true if not.
    OLD and NEW point not to the beginnings of the lines
    but rather to the beginnings of the fields to compare.
    OLDLEN and NEWLEN are their lengths. */
 
-static int
+static bool
 different (char *old, char *new, size_t oldlen, size_t newlen)
 {
   if (check_chars < oldlen)
@@ -227,8 +228,8 @@ different (char *old, char *new, size_t oldlen, size_t newlen)
       /* FIXME: This should invoke strcoll somehow.  */
       return oldlen != newlen || memcasecmp (old, new, oldlen);
     }
-  else if (HAVE_SETLOCALE && hard_LC_COLLATE)
-    return xmemcoll (old, oldlen, new, newlen);
+  else if (hard_LC_COLLATE)
+    return xmemcoll (old, oldlen, new, newlen) != 0;
   else
     return oldlen != newlen || memcmp (old, new, oldlen);
 }
@@ -241,7 +242,7 @@ different (char *old, char *new, size_t oldlen, size_t newlen)
 
 static void
 writeline (struct linebuffer const *line, FILE *stream,
-	   bool match, int linecount)
+	   bool match, uintmax_t linecount)
 {
   if (! (linecount == 0 ? output_unique
 	 : !match ? output_first_repeated
@@ -249,7 +250,7 @@ writeline (struct linebuffer const *line, FILE *stream,
     return;
 
   if (countmode == count_occurrences)
-    fprintf (stream, "%7d ", linecount + 1);
+    fprintf (stream, "%7" PRIuMAX " ", linecount + 1);
 
   fwrite (line->buffer, sizeof (char), line->length, stream);
 }
@@ -321,8 +322,8 @@ check_file (const char *infile, const char *outfile)
     {
       char *prevfield;
       size_t prevlen;
-      int match_count = 0;
-      int first_delimiter = 1;
+      uintmax_t match_count = 0;
+      bool first_delimiter = true;
 
       if (readlinebuffer (prevline, istream) == 0)
 	goto closefiles;
@@ -343,16 +344,21 @@ check_file (const char *infile, const char *outfile)
 	  thisfield = find_field (thisline);
 	  thislen = thisline->length - 1 - (thisfield - thisline->buffer);
 	  match = !different (thisfield, prevfield, thislen, prevlen);
+	  match_count += match;
 
-	  if (match)
-	    ++match_count;
+	  if (match_count == UINTMAX_MAX)
+	    {
+	      if (count_occurrences)
+		error (EXIT_FAILURE, 0, _("too many repeated lines"));
+	      match_count--;
+	    }
 
           if (delimit_groups != DM_NONE)
 	    {
 	      if (!match)
 		{
 		  if (match_count) /* a previous match */
-		    first_delimiter = 0; /* Only used when DM_SEPARATE */
+		    first_delimiter = false; /* Only used when DM_SEPARATE */
 		}
 	      else if (match_count == 1)
 		{
@@ -381,12 +387,15 @@ check_file (const char *infile, const char *outfile)
   if (ferror (istream) || fclose (istream) == EOF)
     error (EXIT_FAILURE, errno, _("error reading %s"), infile);
 
-  if (ferror (ostream))
-    error (EXIT_FAILURE, 0, _("error writing %s"), outfile);
-  /* Close ostream only if it's not stdout -- the latter is closed
-     via the atexit-invoked close_stdout.  */
-  if (ostream != stdout && fclose (ostream) != 0)
-    error (EXIT_FAILURE, errno, _("error writing %s"), outfile);
+  /* Check for errors and close ostream only if it's not stdout --
+     stdout is handled via the atexit-invoked close_stdout function.  */
+  if (ostream != stdout)
+    {
+      if (ferror (ostream))
+	error (EXIT_FAILURE, 0, _("error writing %s"), outfile);
+      if (ostream != stdout && fclose (ostream) != 0)
+	error (EXIT_FAILURE, errno, _("error writing %s"), outfile);
+    }
 
   free (lb1.buffer);
   free (lb2.buffer);
@@ -435,7 +444,7 @@ main (int argc, char **argv)
 	    break;
 	  if (nfiles == 2)
 	    {
-	      error (0, 0, _("extra operand `%s'"), argv[optind]);
+	      error (0, 0, _("extra operand %s"), quote (argv[optind]));
 	      usage (EXIT_FAILURE);
 	    }
 	  file[nfiles++] = argv[optind++];
@@ -452,7 +461,7 @@ main (int argc, char **argv)
 	      skip_chars = size;
 	    else if (nfiles == 2)
 	      {
-		error (0, 0, _("extra operand `%s'"), optarg);
+		error (0, 0, _("extra operand %s"), quote (optarg));
 		usage (EXIT_FAILURE);
 	      }
 	    else
@@ -505,7 +514,7 @@ main (int argc, char **argv)
 	  break;
 
 	case 'i':
-	  ignore_case = 1;
+	  ignore_case = true;
 	  break;
 
 	case 's':		/* Like '+#'. */
@@ -534,7 +543,7 @@ main (int argc, char **argv)
   if (obsolete_skip_fields && 200112 <= posix2_version ())
     {
       error (0, 0, _("`-%lu' option is obsolete; use `-f %lu'"),
-	     (unsigned long) skip_fields, (unsigned long) skip_fields);
+	     (unsigned long int) skip_fields, (unsigned long int) skip_fields);
       usage (EXIT_FAILURE);
     }
 
