@@ -1,5 +1,5 @@
 /* GNU's uptime.
-   Copyright (C) 1992-2002 Free Software Foundation, Inc.
+   Copyright (C) 1992-2002, 2004 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -28,33 +28,30 @@
 # include <sys/sysctl.h>
 #endif
 
+#include "c-strtod.h"
 #include "error.h"
 #include "long-options.h"
+#include "quote.h"
 #include "readutmp.h"
 
 /* The official name of this program (e.g., no `g' prefix).  */
 #define PROGRAM_NAME "uptime"
 
-#define AUTHORS N_ ("Joseph Arceneaux, David MacKenzie, and Kaveh Ghazi")
+#define AUTHORS "Joseph Arceneaux", "David MacKenzie", "Kaveh Ghazi"
 
 int getloadavg ();
 
 /* The name this program was run with. */
 char *program_name;
 
-static struct option const longopts[] =
-{
-  {NULL, 0, NULL, 0}
-};
-
 static void
-print_uptime (int n, const STRUCT_UTMP *this)
+print_uptime (size_t n, const STRUCT_UTMP *this)
 {
-  register int entries = 0;
+  size_t entries = 0;
   time_t boot_time = 0;
   time_t time_now;
   time_t uptime = 0;
-  int updays;
+  long int updays;
   int uphours;
   int upmins;
   struct tm *tmn;
@@ -68,16 +65,14 @@ print_uptime (int n, const STRUCT_UTMP *this)
   if (fp != NULL)
     {
       char buf[BUFSIZ];
-      int res;
       char *b = fgets (buf, BUFSIZ, fp);
       if (b == buf)
 	{
-	  /* The following sscanf must use the C locale.  */
-	  setlocale (LC_NUMERIC, "C");
-	  res = sscanf (buf, "%lf", &upsecs);
-	  setlocale (LC_NUMERIC, "");
-	  if (res == 1)
-	    uptime = (time_t) upsecs;
+	  char *end_ptr;
+	  upsecs = c_strtod (buf, &end_ptr);
+	  if (buf != end_ptr)
+	    uptime = (0 <= upsecs && upsecs < TYPE_MAXIMUM (time_t)
+		      ? upsecs : -1);
 	}
 
       fclose (fp);
@@ -100,23 +95,9 @@ print_uptime (int n, const STRUCT_UTMP *this)
      ones, also in the process possibly gleaning boottime. */
   while (n--)
     {
-      if (UT_USER (this) [0]
-#ifdef USER_PROCESS
-	  && this->ut_type == USER_PROCESS
-#endif
-	  )
-	{
-	  ++entries;
-	}
-      /* If BOOT_MSG is defined, we can get boottime from utmp.  This avoids
-	 possibly needing special privs to read /dev/kmem. */
-#ifdef BOOT_MSG
-# if HAVE_PROC_UPTIME
-      if (uptime == 0)
-# endif /* HAVE_PROC_UPTIME */
-	if (STREQ (this->ut_line, BOOT_MSG))
-	  boot_time = UT_TIME_MEMBER (this);
-#endif /* BOOT_MSG */
+      entries += IS_USER_PROCESS (this);
+      if (UT_TYPE_BOOT_TIME (this))
+	boot_time = UT_TIME_MEMBER (this);
       ++this;
     }
   time_now = time (0);
@@ -132,15 +113,24 @@ print_uptime (int n, const STRUCT_UTMP *this)
   uphours = (uptime - (updays * 86400)) / 3600;
   upmins = (uptime - (updays * 86400) - (uphours * 3600)) / 60;
   tmn = localtime (&time_now);
-  printf (_(" %2d:%02d%s  up "), ((tmn->tm_hour % 12) == 0
-				  ? 12 : tmn->tm_hour % 12),
-	  /* FIXME: use strftime, not am, pm.  Uli reports that
-	     the german translation is meaningless.  */
-	  tmn->tm_min, (tmn->tm_hour < 12 ? _("am") : _("pm")));
-  if (updays > 0)
-    printf (ngettext("%d day", "%d days", updays), updays);
-  printf (" %2d:%02d,  ", uphours, upmins);
-  printf (ngettext ("%d user", "%d users", entries), entries);
+  if (tmn)
+    printf (_(" %2d:%02d%s  up "),
+	    ((tmn->tm_hour % 12) == 0 ? 12 : tmn->tm_hour % 12),
+	    /* FIXME: use strftime, not am, pm.  Uli reports that
+	       the german translation is meaningless.  */
+	    tmn->tm_min, (tmn->tm_hour < 12 ? _("am") : _("pm")));
+  else
+    printf (_(" ??:????  up "));
+  if (uptime == (time_t) -1)
+    printf (_("???? days ??:??,  "));
+  else
+    {
+      if (0 < updays)
+	printf (ngettext ("%ld day", "%ld days", updays), updays);
+      printf (" %2d:%02d,  ", uphours, upmins);
+    }
+  printf (ngettext ("%lu user", "%lu users", entries),
+	  (unsigned long int) entries);
 
 #if defined (HAVE_GETLOADAVG) || defined (C_GETLOADAVG)
   loads = getloadavg (avg, 3);
@@ -169,11 +159,10 @@ print_uptime (int n, const STRUCT_UTMP *this)
 static void
 uptime (const char *filename)
 {
-  int n_users;
+  size_t n_users;
   STRUCT_UTMP *utmp_buf;
-  int fail = read_utmp (filename, &n_users, &utmp_buf);
 
-  if (fail)
+  if (read_utmp (filename, &n_users, &utmp_buf) != 0)
     error (EXIT_FAILURE, errno, "%s", filename);
 
   print_uptime (n_users, utmp_buf);
@@ -182,7 +171,7 @@ uptime (const char *filename)
 void
 usage (int status)
 {
-  if (status != 0)
+  if (status != EXIT_SUCCESS)
     fprintf (stderr, _("Try `%s --help' for more information.\n"),
 	     program_name);
   else
@@ -206,7 +195,6 @@ If FILE is not specified, use %s.  %s as FILE is common.\n\
 int
 main (int argc, char **argv)
 {
-  int optc, longind;
   initialize_main (&argc, &argv);
   program_name = argv[0];
   setlocale (LC_ALL, "");
@@ -216,19 +204,9 @@ main (int argc, char **argv)
   atexit (close_stdout);
 
   parse_long_options (argc, argv, PROGRAM_NAME, GNU_PACKAGE, VERSION,
-		      AUTHORS, usage);
-
-  while ((optc = getopt_long (argc, argv, "", longopts, &longind)) != -1)
-    {
-      switch (optc)
-	{
-	case 0:
-	  break;
-
-	default:
-	  usage (EXIT_FAILURE);
-	}
-    }
+		      usage, AUTHORS, (char const *) NULL);
+  if (getopt_long (argc, argv, "", NULL, NULL) != -1)
+    usage (EXIT_FAILURE);
 
   switch (argc - optind)
     {
@@ -241,7 +219,7 @@ main (int argc, char **argv)
       break;
 
     default:			/* lose */
-      error (0, 0, _("too many arguments"));
+      error (0, 0, _("extra operand %s"), quote (argv[optind + 1]));
       usage (EXIT_FAILURE);
     }
 
