@@ -144,6 +144,20 @@ sub test_vector
 
      # Ensure that we can parse MONTHNAME-DAY-YEAR.
      ['moname-d-y', '--iso -d May-23-2003', {}, '2003-05-23', 0],
+
+     ['epoch', '--iso=sec -d @31536000', {},
+      '1971-01-01T00:00:00+0000', 0],
+
+     ['ns-10', '--iso=ns -d "1969-12-31 13:00:00.00000001-1100"', {},
+      '1970-01-01T00:00:00,000000010+0000', 0],
+
+     ['ns-max32', '--iso=ns -d "2038-01-19 03:14:07.999999999"', {},
+      '2038-01-19T03:14:07,999999999+0000', 0],
+
+     ['ns-relative',
+      "--iso=ns -d'1970-01-01 00:00:00.1234567 UTC +961062237.987654321 sec'",
+      {}, '2000-06-15T09:43:58,111111021+0000', 0],
+
      # FIXME: add a lot more...
      );
 
