@@ -3,7 +3,7 @@ package Test;
 require 5.002;
 use strict;
 
-# Tell head to accept old-style options like `-1'.
+# Tell pr to accept old-style options like operand-less `-S'.
 $Test::env_default = ['_POSIX2_VERSION=199209'];
 
 my @tv = (
@@ -363,6 +363,10 @@ sub test_vector
       my $sep = ($flags ? ' ' : '');
       $flags = "$common_option_prefix$sep$flags";
       push (@new_tv, [$test_name, $flags, $in, $exp, $ret]);
+
+      (my $new_flags = $flags) =~ s/(^| )-(\d+)( |$)/$1--columns=$2$3/g;
+      $new_flags ne $flags
+	and push (@new_tv, ["$test_name.C", $new_flags, $in, $exp, $ret]);
     }
 
   return @new_tv;
